#include "stdafx.h"

#include "AnimData.h"
#include "AnimKey.h"
#include "AnimTrack.h"
#include "AnimMgr.h"


CAnimMgr::~CAnimMgr()
{
    for(AnimListTracks::iterator theIterator = m_ListTracks.begin(); theIterator != m_ListTracks.end(); theIterator++)
        {
        CAnimMgrTrack Track = *theIterator;

        delete Track.m_lpTrack;
        }

    m_ListTracks.clear();
}


void    CAnimMgr::vPushTrack(CAnimTrack* lpTrack)
{
    lpTrack->vPackKeys();

    m_ListTracks.push_back(CAnimMgrTrack(lpTrack,m_dwNumTracks++));
}


void    CAnimMgr::vSetNextTrack(DWORD dwTrackIndex, DWORD dwNextTrackIndex, DWORD dwNextTrackFrame)
{
    ASSERT(dwTrackIndex >= 0  &&  dwTrackIndex < m_dwNumTracks);

    if(dwNextTrackIndex != NO_CHANGE) m_ListTracks[dwTrackIndex].m_dwNextTrackIndex = dwNextTrackIndex;
    if(dwNextTrackFrame != NO_CHANGE) m_ListTracks[dwTrackIndex].m_dwNextTrackFrame = dwNextTrackFrame;
}


void    CAnimMgr::vSetCurrentTrack(DWORD dwTrackIndex,FLOAT fFrame)
{
    ASSERT(dwTrackIndex >= 0  &&  dwTrackIndex < m_dwNumTracks);

    m_dwCurrentTrackIndex = dwTrackIndex;

    CAnimMgrTrack* lpTrack = &m_ListTracks[dwTrackIndex];

    m_ListTracks[dwTrackIndex].m_lpTrack->vSetOverallSpeed(m_fOverallSpeed);
    if(m_boChainTracks)
        {
        m_ListTracks[dwTrackIndex].m_lpTrack->vSetNextAnim(m_ListTracks[lpTrack->m_dwNextTrackIndex].m_lpTrack,lpTrack->m_dwNextTrackFrame);
        }
    else
        {
        m_ListTracks[dwTrackIndex].m_lpTrack->vSetNextAnim(lpTrack->m_lpTrack,0);
        }
    m_ListTracks[dwTrackIndex].m_lpTrack->boSetCurrentFrame(fFrame);
}


void    CAnimMgr::vGetNextAnimInfos(DWORD& dwNextTrackIndex,DWORD& dwNextTrackFrame,DWORD dwTrackIndex /*= -1*/)
{
    dwNextTrackIndex = dwNextTrackFrame = 0;

    if(dwTrackIndex == -1)
        {
        if(m_dwCurrentTrackIndex == -1) return;

        dwNextTrackIndex = m_ListTracks[m_dwCurrentTrackIndex].m_dwNextTrackIndex;
        dwNextTrackFrame = m_ListTracks[m_dwCurrentTrackIndex].m_dwNextTrackFrame;
        }
    else
        {
        ASSERT(dwTrackIndex < m_dwNumTracks);

        dwNextTrackIndex = m_ListTracks[dwTrackIndex].m_dwNextTrackIndex;
        dwNextTrackFrame = m_ListTracks[dwTrackIndex].m_dwNextTrackFrame;
        }
}


bool    CAnimMgr::boRunForward(FLOAT fElapsedTime)
{
    if(m_dwCurrentTrackIndex == -1) return true;

    if(m_ListTracks[m_dwCurrentTrackIndex].m_lpTrack->boRunForward(fElapsedTime)) return true;

    // We need to link to the next track
    vLinkToNextTrack();

    return false;
}


void    CAnimMgr::vSetOverallSpeed(FLOAT fOverallSpeed)
{
    ASSERT(fOverallSpeed >= (FLOAT)0.0);

    m_fOverallSpeed = fOverallSpeed;

    if(m_dwCurrentTrackIndex != -1) m_ListTracks[m_dwCurrentTrackIndex].m_lpTrack->vSetOverallSpeed(m_fOverallSpeed);
}


bool    CAnimMgr::boSetCurrentFrame(FLOAT fFrame)
{
    if(m_dwCurrentTrackIndex == -1) return true;

    if(m_ListTracks[m_dwCurrentTrackIndex].m_lpTrack->boSetCurrentFrame(fFrame)) return true;

    vLinkToNextTrack();

    return false;
}


void    CAnimMgr::vLinkToNextTrack()
{
    DWORD   dwNextTrack = m_ListTracks[m_dwCurrentTrackIndex].m_dwNextTrackIndex;
    DWORD   dwNextFrame = m_ListTracks[m_dwCurrentTrackIndex].m_dwNextTrackFrame;

    FLOAT   fNextFrame = (FLOAT)dwNextFrame + m_ListTracks[m_dwCurrentTrackIndex].m_lpTrack->fGetInterpolationFactor();

    vSetCurrentTrack(dwNextTrack,fNextFrame);
}


void    CAnimMgr::vSetChainTracks(bool boChain)
{
    m_boChainTracks = boChain;

    if(m_dwCurrentTrackIndex == -1) return;

    CAnimMgrTrack* lpTrack = &m_ListTracks[m_dwCurrentTrackIndex];

    if(boChain)
        {
        m_ListTracks[m_dwCurrentTrackIndex].m_lpTrack->vSetNextAnim(m_ListTracks[lpTrack->m_dwNextTrackIndex].m_lpTrack,lpTrack->m_dwNextTrackFrame);
        }
    else
        {
        m_ListTracks[m_dwCurrentTrackIndex].m_lpTrack->vSetNextAnim(lpTrack->m_lpTrack,0);
        }
}


void    CAnimMgr::vStealTrack( DWORD dwTrackDest, CAnimMgr& AnimMgr, DWORD dwTrackSrc )
{
    _ASSERTE( (dwTrackDest != -1  &&  dwTrackDest < m_dwNumTracks)  ||  (dwTrackDest == -1) );
    _ASSERTE( dwTrackSrc != -1  &&  dwTrackSrc  < AnimMgr.m_dwNumTracks );

    if( dwTrackDest == -1 )
        {
        // Add the stolen track to the current list of the current anim mgr
        vPushTrack( AnimMgr.lpGetTrack( dwTrackSrc ) );
        }
    else
        {
        // Replace the dwTrackDest track in the current anim mgr
        m_ListTracks[ dwTrackDest ] = AnimMgr.m_ListTracks[ dwTrackSrc ];
        }

    AnimMgr.m_dwCurrentTrackIndex = 0;
    AnimMgr.m_dwNumTracks--;
    AnimMgr.m_ListTracks.erase( AnimMgr.m_ListTracks.begin() + dwTrackSrc );
    AnimMgr.m_ListTracks.resize( AnimMgr.m_dwNumTracks );
}
