// EditMesh.cpp : implementation file
//

#include "stdafx.h"
#include "TRViewer.h"
#include "MainFrm.h"
#include "EditMesh.h"
#include "HandleMesh.h"
#include "ToolboxMesh.h"


#ifdef _DEBUG
//!#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

#define OPENGLWND_LEFT_OFFSET       8
#define OPENGLWND_TOP_OFFSET        4
#define OPENGLWND_RIGHT_OFFSET      4
#define OPENGLWND_BOTTOM_OFFSET     4

/////////////////////////////////////////////////////////////////////////////
// CEditMesh

IMPLEMENT_DYNCREATE(CEditMesh, CFormView)

CEditMesh::CEditMesh()
	: CFormView(CEditMesh::IDD)
{
	//{{AFX_DATA_INIT(CEditMesh)
	//}}AFX_DATA_INIT
    m_lpDataHandler     = NULL;
    m_OpenGLWnd         = NULL;
    m_OpenGLWndCreated  = false;
    m_lpToolbox         = reinterpret_cast<CToolboxMesh*> (((CMainFrame*)(AfxGetApp()->m_pMainWnd))->lpGetToolbox(TOOLBOX_MESH));
}

CEditMesh::~CEditMesh()
{
    vRelease();
}

void CEditMesh::vRelease()
{
    m_lpToolbox->vSetDataTypeHandler(NULL,NULL);

    if(m_OpenGLWndCreated)
        {
        m_OpenGLWnd->DestroyWindow();
        delete m_OpenGLWnd;
        }

    m_OpenGLWnd = NULL;
    m_OpenGLWndCreated = false;
    m_lpDataHandler = NULL;
}


void CEditMesh::DoDataExchange(CDataExchange* pDX)
{
	CFormView::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CEditMesh)
	//}}AFX_DATA_MAP
}


BEGIN_MESSAGE_MAP(CEditMesh, CFormView)
	//{{AFX_MSG_MAP(CEditMesh)
	ON_WM_SIZE()
	ON_WM_LBUTTONDOWN()
	ON_WM_LBUTTONUP()
	ON_WM_MOUSEMOVE()
	ON_WM_RBUTTONDOWN()
	ON_WM_RBUTTONUP()
	ON_WM_MOUSEWHEEL()
	ON_WM_LBUTTONDBLCLK()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CEditMesh diagnostics

#ifdef _DEBUG
void CEditMesh::AssertValid() const
{
	CFormView::AssertValid();
}

void CEditMesh::Dump(CDumpContext& dc) const
{
	CFormView::Dump(dc);
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CEditMesh message handlers

void CEditMesh::OnInitialUpdate() 
{
	CFormView::OnInitialUpdate();

    SetScrollSizes(MM_TEXT,CSize(0,0));
}


void CEditMesh::vSetDataTypeHandler(CHandleMesh* lpHandler)
{
    m_lpDataHandler = lpHandler;

    if(lpHandler  &&  !m_OpenGLWndCreated) vCreateOpenGLWnd();

    if(m_OpenGLWndCreated)
        {
        DWORD dwNumMesh;

        CTrLstMsh* lpTrMsh = m_lpDataHandler->GetMesh(dwNumMesh);

        m_OpenGLWnd->vSetGfxAddr(lpTrMsh,m_lpDataHandler->boIsStaticMesh() ? TrObj_Type_StaticMesh : TrObj_Type_Mesh,false,dwNumMesh);
        }

    m_lpToolbox->vSetDataTypeHandler(lpHandler,this);

    InvalidateRect(NULL);
}


void CEditMesh::OnDraw(CDC* pDC) 
{
}


void CEditMesh::vCreateOpenGLWnd()
{
    RECT rect;

    GetClientRect(&rect);

    rect.left   = OPENGLWND_LEFT_OFFSET;
    rect.top    = OPENGLWND_TOP_OFFSET;
    rect.bottom -= OPENGLWND_BOTTOM_OFFSET;
    rect.right  -= OPENGLWND_RIGHT_OFFSET;

    m_OpenGLWnd = new COpenGLWnd;

    m_OpenGLWnd->Create("static","",WS_CHILD | WS_VISIBLE | WS_CLIPCHILDREN | WS_CLIPSIBLINGS,rect,this,12300);
    m_OpenGLWnd->ShowWindow(SW_SHOW);

    m_OpenGLWndCreated = true;
}


void CEditMesh::OnSize(UINT nType, int cx, int cy) 
{
	CFormView::OnSize(nType, cx, cy);
	
    if(m_OpenGLWndCreated) m_OpenGLWnd->SetWindowPos(NULL,OPENGLWND_LEFT_OFFSET,OPENGLWND_TOP_OFFSET,cx-OPENGLWND_LEFT_OFFSET-OPENGLWND_RIGHT_OFFSET,cy-OPENGLWND_BOTTOM_OFFSET-OPENGLWND_TOP_OFFSET,0);
}


BOOL CEditMesh::boOnIdle()
{
    if(m_OpenGLWndCreated)
        {
        CMainFrame* lpMainFrame = (CMainFrame*)(AfxGetApp()->m_pMainWnd);

        if(lpMainFrame->boIsManualCamera())
            {
            m_OpenGLWnd->vCheckKeys();
            vCheckKeys(m_OpenGLWnd->fGetFrameRate());
            }
        vRedrawOpenGLWnd();
        }

    return TRUE;
}


void CEditMesh::vCheckKeys(float fFrameRate)
{
    bool boShifted  = GetAsyncKeyState(VK_SHIFT)   & 0x8001 ? true : false;
    bool boCtrled   = GetAsyncKeyState(VK_CONTROL) & 0x8001 ? true : false;

    if(fFrameRate == 0.0f) fFrameRate = 1.0f;

    bit16 sStep = 70.0f / fFrameRate; if(!sStep) sStep = 1;

    DWORD   dwNumMesh;

    CTrLstMsh*  lpMsh       = m_lpDataHandler ? m_lpDataHandler->GetMesh(dwNumMesh) : NULL;
    bool        boStatMesh  = m_lpDataHandler ? m_lpDataHandler->boIsStaticMesh() : false;

    bit16 sIts  = GetAsyncKeyState(VK_SUBTRACT) & 0x8001 ? -1 : GetAsyncKeyState(VK_ADD) & 0x8001 ? +1 : 0;

    if(sIts)
        {
        m_OpenGLWnd->vModifyIntensOfSelectedVertices(sIts * 50 * sStep);
        m_lpDataHandler->vSetChanged();
        }

    if(!lpMsh  ||  (!boShifted  &&  !boCtrled)) return;

    bit16 sXSgn = GetAsyncKeyState(VK_LEFT)  & 0x8001 ? -1 : GetAsyncKeyState(VK_RIGHT) & 0x8001 ? +1 : 0;
    bit16 sYSgn = GetAsyncKeyState(VK_PRIOR) & 0x8001 ? -1 : GetAsyncKeyState(VK_NEXT)  & 0x8001 ? +1 : 0;
    bit16 sZSgn = GetAsyncKeyState(VK_DOWN)  & 0x8001 ? -1 : GetAsyncKeyState(VK_UP)    & 0x8001 ? +1 : 0;
    
    if(!sXSgn  &&  !sYSgn  &&  !sZSgn) return;

    if(boShifted)
        {
        TR123_MESH* lpMesh = const_cast<TR123_MESH*>(lpMsh->lpGetMesh(dwNumMesh));

        CTRFile::vTranslateMesh(lpMesh,sStep * sXSgn,sStep * sYSgn,sStep * sZSgn,true,m_lpDataHandler->lpGetTRFile());
        }

    if(boCtrled)
        {
        if(boStatMesh  &&  m_lpToolbox)
            {
            CTrStatMsh* lpStatMesh = static_cast<CTrStatMsh*>(lpMsh);

            CVect3D vecTrans(sStep * sXSgn, -sStep * sYSgn, -sStep * sZSgn);

            BOOL boBB1IsSel = m_lpToolbox->boStaticBB1IsSel();
            BOOL boBB2IsSel = m_lpToolbox->boStaticBB2IsSel();

            if(boBB1IsSel)
                {
                CBoundingBox& BB = lpStatMesh->GetBBVisibility();

                BB.vSetBB(BB.GetBBLo() + vecTrans,BB.GetBBHi() + vecTrans);
                }

            if(boBB2IsSel)
                {
                CBoundingBox& BB = lpStatMesh->GetBBCollision();

                BB.vSetBB(BB.GetBBLo() + vecTrans,BB.GetBBHi() + vecTrans);
                }
            }
        else
            {
            TR123_MESH* lpMesh = const_cast<TR123_MESH*>(lpMsh->lpGetMesh(dwNumMesh));

            lpMesh->Center.sX += sStep * sXSgn;
            lpMesh->Center.sY += sStep * sYSgn;
            lpMesh->Center.sZ += sStep * sZSgn;
            }
        }

    m_lpDataHandler->vSetChanged();
}


void CEditMesh::OnActivateView(BOOL bActivate, CView* pActivateView, CView* pDeactiveView) 
{
    if(m_OpenGLWndCreated  &&  bActivate) m_OpenGLWnd->vActivateWnd();

    if(bActivate)
        {
        CMainFrame* lpFrame = (CMainFrame*)(AfxGetApp()->m_pMainWnd);

        if(lpFrame->boToolboxIsVisible()) lpFrame->vSetToolboxType(TOOLBOX_MESH);
        m_lpToolbox->vSetDataTypeHandler(bActivate ? m_lpDataHandler : NULL,bActivate ? this : NULL);
        }

    if(pActivateView  ||  pDeactiveView)
	    CFormView::OnActivateView(bActivate, pActivateView, pDeactiveView);
}


void CEditMesh::vReloadTextures()
{
    if(m_OpenGLWnd) m_OpenGLWnd->vReleaseTextures();
}


void CEditMesh::OnLButtonDown(UINT nFlags, CPoint point) 
{
    if(m_OpenGLWnd)
        {
        RECT rect;

        GetClientRect(&rect);

        bool boNoOk = (point.x < OPENGLWND_LEFT_OFFSET  ||  point.x > rect.right - OPENGLWND_RIGHT_OFFSET);
        
        boNoOk |= (point.y < OPENGLWND_TOP_OFFSET  ||  point.y > rect.bottom - OPENGLWND_BOTTOM_OFFSET);

        if(!boNoOk)
            {
            m_OpenGLWnd->vPick(point.x - OPENGLWND_LEFT_OFFSET,point.y - OPENGLWND_TOP_OFFSET);
            }

        m_OpenGLWnd->OnLButtonDown(nFlags, point);
        }
	
	CFormView::OnLButtonDown(nFlags, point);
}


void CEditMesh::OnLButtonUp(UINT nFlags, CPoint point) 
{
    if(m_OpenGLWndCreated) m_OpenGLWnd->OnLButtonUp(nFlags, point);
	
	CFormView::OnLButtonUp(nFlags, point);
}


void CEditMesh::OnMouseMove(UINT nFlags, CPoint point) 
{
    if(m_OpenGLWndCreated) m_OpenGLWnd->OnMouseMove(nFlags, point);
	
	CFormView::OnMouseMove(nFlags, point);
}


void CEditMesh::OnRButtonDown(UINT nFlags, CPoint point) 
{
    if(m_OpenGLWndCreated) m_OpenGLWnd->OnRButtonDown(nFlags, point);
	
	CFormView::OnRButtonDown(nFlags, point);
}


void CEditMesh::OnRButtonUp(UINT nFlags, CPoint point) 
{
    if(m_OpenGLWndCreated) m_OpenGLWnd->OnRButtonUp(nFlags, point);
	
	CFormView::OnRButtonUp(nFlags, point);
}


BOOL CEditMesh::OnMouseWheel(UINT nFlags, short zDelta, CPoint pt) 
{
    if(m_OpenGLWndCreated) m_OpenGLWnd->OnMouseWheel(nFlags, zDelta, pt);
	
	return CFormView::OnMouseWheel(nFlags, zDelta, pt);
}

void CEditMesh::OnLButtonDblClk(UINT nFlags, CPoint point) 
{
    if(m_OpenGLWndCreated) m_OpenGLWnd->OnLButtonDblClk(nFlags, point);
	
	CFormView::OnLButtonDblClk(nFlags, point);
}
