#include "stdafx.h"

#include "Resource.h"
#include "FileTree.h"
#include "HandleSprites.h"
#include "HandleSprite.h"
#include "EditEmpty.h"


void    CHandleSprites::vCreatePane()
{
    m_lpSplitterWnd->CreateView(0,1,RUNTIME_CLASS(CEditEmpty),CSize(1,1),NULL);
}


void    CHandleSprites::vExpandTree(bool boAlreadyExpanded)
{
    if(boAlreadyExpanded)
        {
        m_lpTree->m_boDeleting = true;
        m_lpTree->vReleaseNode(m_hParentItem,true,true);
        m_lpTree->m_boDeleting = false;
        }

    CString             s;
    DWORD               dwI = 0;
    HTREEITEM           h;
    CTRFile*            lpTRFile    = m_lpTree->lpGetTRFile();
    TR_FILE             trFile      = lpTRFile->GetTRFile();
    enumTRType          trType      = lpTRFile->GetTRType();

    TCHAR               szLevelName[19+1];
    
    lstrcpyn(szLevelName,lpTRFile->lpszGetShortLevelName(),19+1);

    switch(trType)
        {
        //
        case TR1_FILE:
        case TR2_FILE:
        case TR3_FILE:
            {
            TRListSpriteSeq* lpList = trFile.tr123File.lpListSpriteSeq;
            for(TRListSpriteSeq::iterator theIterator = lpList->begin(); theIterator != lpList->end(); theIterator++)
                {
                s.Format(_T("%s_SPR%.3d"),szLevelName,dwI++);
                INSERT(m_lpTree,h,m_hParentItem,new CHandleSprite(m_lpTree,m_lpSplitterWnd,*theIterator))
                }
            break;
            }
        }
}


CMenu*  CHandleSprites::lpGetPopupupMenu(HTREEITEM hSelectedItem)
{
    m_lpMenu = new CMenu;
    m_lpMenu->LoadMenu(IDR_MENU_SPRITES);

    return m_lpMenu->GetSubMenu(0);
}


void    CHandleSprites::vReleasePopupupMenu()
{
    delete m_lpMenu;
    
    m_lpMenu = NULL;
}


CHandleSprite* CHandleSprites::lpFindSprite(DWORD dwObjId)
{
    HTREEITEM hNextItem;
    HTREEITEM hChildItem = m_lpTree->GetChildItem(m_hParentItem);

    while(hChildItem != NULL)
        {
        hNextItem = m_lpTree->GetNextItem(hChildItem, TVGN_NEXT);

        CHandleSprite* lpNode = (CHandleSprite*) m_lpTree->GetItemData(hChildItem);

        if(lpNode->GetSpr()->dwGetObjectId() == dwObjId) return lpNode;

        hChildItem = hNextItem;
        }

    return NULL;
}


void    CHandleSprites::vHandleMenu(WORD wMenuId)
{
    switch(wMenuId)
        {
        // Add a sprite from a .TRSPR file
        case ID_SPRITES_IMPORTFROMTRSPR:
            {
            vImportFromSPR();
            break;
            }
        // Export all sprites to .SPR files
        case ID_SPRITES_EXPORTASSPR:
            {
            vExportAsSPR();
            break;
            }
        }
}


void    CHandleSprites::vExportAsSPR()
{
    BROWSEINFO   bi;
    ZeroMemory(&bi,sizeof(BROWSEINFO));

    TCHAR szBuf[MAX_PATH];
    lstrcpy(szBuf, __TEXT("Choose a folder to save the TRSPR in:"));

    bi.hwndOwner  = NULL;
    bi.ulFlags    = 0;
    bi.lpszTitle  = static_cast<LPCSTR>(szBuf);
    bi.lpfn       = NULL;
    bi.lParam     = (long)this;

    LPITEMIDLIST  pidl = SHBrowseForFolder(&bi);

    if(!pidl) return;

    SHGetPathFromIDList(pidl,szBuf);

    CWaitCursor dummy;

    CTRFile*    lpTRFile    = m_lpTree->lpGetTRFile();
    TR_FILE     trFile      = lpTRFile->GetTRFile();

    CString     s;
    TCHAR       szLevelName[19+1];
    DWORD       dwI = 0;
    
    lstrcpyn(szLevelName,lpTRFile->lpszGetShortLevelName(),19+1);

    TRListSpriteSeq* lpList = trFile.tr123File.lpListSpriteSeq;
    for(TRListSpriteSeq::iterator theIterator = lpList->begin(); theIterator != lpList->end(); theIterator++)
        {
        CTrLstSpr*  lpSprSeq = *theIterator;

        DWORD   dwSPRSize;
        BYTE*   lpSPR = lpSprSeq->lpbGetData(dwSPRSize);

        s.Format(_T("%s\\%s_SPR%.3d.trspr"),szBuf,szLevelName,dwI++);

        DWORD   dwNumWritten;
        HANDLE hFile = CreateFile(s,GENERIC_WRITE,0,NULL,CREATE_ALWAYS,FILE_ATTRIBUTE_ARCHIVE,NULL);
    
        WriteFile(hFile,lpSPR,dwSPRSize,&dwNumWritten,NULL);
    
        CloseHandle(hFile);

        delete [] lpSPR;
        }
}


//
//
// Import a sprite from a .TRSPR
//
//
void    CHandleSprites::vImportFromSPR(CTrLstSpr* lpSpr_)
{
    CTrLstSpr* lpSpr = new CTrLstSpr;

    if(lpSpr == NULL)
        {
        // Load the SPR file
        CString strFilter = _T("TR Sprite files (*.trspr)|*.trspr||");

        CFileDialog FileDialog(TRUE,_T(""),_T(""),OFN_HIDEREADONLY,strFilter);

        if(FileDialog.DoModal() != IDOK)
            {
            delete lpSpr;
            return;
            }

        CWaitCursor dummy;

        CTrLstSpr* Spr = CTrLstSpr::objLoad(FileDialog.GetPathName());

        if(Spr)
            {
            lpSpr->vRelease();
            lpSpr->vStealData(*Spr);
            delete Spr;
            }
        else
            {
            delete lpSpr;
            AfxMessageBox(_T("Could not import this sprite because of a version mismatch:\n\n - the sprite was produced with an old TRViewer release and its format is not supported anymore !"),MB_ICONSTOP | MB_TOPMOST);
            return;
            }
        }
    else
        {
        lpSpr->vRelease();
        lpSpr->vStealData(*lpSpr_);
        }

    CWaitCursor dummy;

    CTRFile* lpTRFile = m_lpTree->lpGetTRFile();

    lpTRFile->GetTRFile().tr123File.lpListSpriteSeq->push_back(lpSpr);

    vSetChanged();

    if(m_boExpanded)
        {
        vExpandTree(true);
        }
}
