#include "stdafx.h"

#include "Resource.h"
#include "FileTree.h"
#include "HandleStaticMeshes.h"
#include "HandleMesh.h"
#include "EditEmpty.h"


void    CHandleStaticMeshes::vCreatePane()
{
    m_lpSplitterWnd->CreateView(0,1,RUNTIME_CLASS(CEditEmpty),CSize(1,1),NULL);
}


void    CHandleStaticMeshes::vExpandTree(bool boAlreadyExpanded)
{
    if(boAlreadyExpanded)
        {
        m_lpTree->m_boDeleting = true;
        m_lpTree->vReleaseNode(m_hParentItem,true,true);
        m_lpTree->m_boDeleting = false;
        }

    CString             s;
    DWORD               dwI = 0;
    HTREEITEM           h;
    CTRFile*            lpTRFile    = m_lpTree->lpGetTRFile();
    TR_FILE             trFile      = lpTRFile->GetTRFile();
    enumTRType          trType      = lpTRFile->GetTRType();

    TCHAR               szLevelName[19+1];
    
    lstrcpyn(szLevelName,lpTRFile->lpszGetShortLevelName(),19+1);

    switch(trType)
        {
        //
        case TR1_FILE:
        case TR2_FILE:
        case TR3_FILE:
            {
            TRListStaticMesh* lpList = trFile.tr123File.lpListStaticMesh;
            for(TRListStaticMesh::iterator theIterator = lpList->begin(); theIterator != lpList->end(); theIterator++)
                {
                s.Format(_T("%s_SMH%.3d"),szLevelName,dwI++);
                INSERT(m_lpTree,h,m_hParentItem,new CHandleMesh(m_lpTree,m_lpSplitterWnd,(void*)*theIterator,true))
                }
            break;
            }
        }
}


CMenu*  CHandleStaticMeshes::lpGetPopupupMenu(HTREEITEM hSelectedItem)
{
    m_lpMenu = new CMenu;
    m_lpMenu->LoadMenu(IDR_MENU_STATICMESHES);

    return m_lpMenu->GetSubMenu(0);
}


void    CHandleStaticMeshes::vReleasePopupupMenu()
{
    delete m_lpMenu;
    
    m_lpMenu = NULL;
}


CHandleMesh* CHandleStaticMeshes::lpFindStaticMesh(DWORD dwObjId)
{
    HTREEITEM hNextItem;
    HTREEITEM hChildItem = m_lpTree->GetChildItem(m_hParentItem);
    DWORD     dwNumMesh;

    while(hChildItem != NULL)
        {
        hNextItem = m_lpTree->GetNextItem(hChildItem, TVGN_NEXT);

        CHandleMesh* lpNode = (CHandleMesh*) m_lpTree->GetItemData(hChildItem);

        if(lpNode->GetMesh(dwNumMesh)->dwGetObjectId() == dwObjId) return lpNode;

        hChildItem = hNextItem;
        }

    return NULL;
}


void    CHandleStaticMeshes::vHandleMenu(WORD wMenuId)
{
    switch(wMenuId)
        {
        // Load a 3ds file and add a new static mesh
        case ID_STATICMESHES_IMPORTFROM3DS:
            {
            vImportFrom3DS();
            break;
            }
        // Add a static mesh from a .TRSMH file
        case ID_STATICMESHES_IMPORTFROMTRSMH:
            {
            vImportFromMSH();
            break;
            }
        // Export all meshes to .MSH files
        case ID_STATICMESHES_EXPORTASMSH:
            {
            vExportAsMSH();
            break;
            }
        }
}


void    CHandleStaticMeshes::vExportAsMSH()
{
    BROWSEINFO   bi;
    ZeroMemory(&bi,sizeof(BROWSEINFO));

    TCHAR szBuf[MAX_PATH];
    lstrcpy(szBuf, __TEXT("Choose a folder to save the TRSMH in:"));

    bi.hwndOwner  = NULL;
    bi.ulFlags    = 0;
    bi.lpszTitle  = static_cast<LPCSTR>(szBuf);
    bi.lpfn       = NULL;
    bi.lParam     = (long)this;

    LPITEMIDLIST  pidl = SHBrowseForFolder(&bi);

    if(!pidl) return;

    SHGetPathFromIDList(pidl,szBuf);

    CWaitCursor dummy;

    CTRFile*    lpTRFile    = m_lpTree->lpGetTRFile();
    TR_FILE     trFile      = lpTRFile->GetTRFile();

    CString     s;
    TCHAR       szLevelName[19+1];
    
    lstrcpyn(szLevelName,lpTRFile->lpszGetShortLevelName(),19+1);

    for(DWORD dwI = 0; dwI < trFile.tr123File.dwNumStaticMeshes; dwI++)
        {
        CTrStatMsh  Msh(&(trFile.tr123File.lpStaticMeshes[dwI]),lpTRFile);

        DWORD   dwMSHSize;
        BYTE*   lpMSH = Msh.lpbGetData(dwMSHSize);

        s.Format(_T("%s\\%s_SMH%.3d.trsmh"),szBuf,szLevelName,dwI);

        DWORD   dwNumWritten;
        HANDLE hFile = CreateFile(s,GENERIC_WRITE,0,NULL,CREATE_ALWAYS,FILE_ATTRIBUTE_ARCHIVE,NULL);
    
        WriteFile(hFile,lpMSH,dwMSHSize,&dwNumWritten,NULL);
    
        CloseHandle(hFile);

        delete [] lpMSH;
        }
}


//
//
// Import a static mesh from a .3DS
//
//
void    CHandleStaticMeshes::vImportFrom3DS()
{
    BYTE* lpData = NULL;

    // Load the 3DS file
    CString strFilter(_T("3DS files (*.3ds)|*.3ds||"));
    CFileDialog FileDialog(TRUE,_T(""),_T(""),OFN_HIDEREADONLY,strFilter);

    if(FileDialog.DoModal() != IDOK) return;

    CWaitCursor dummy;

    CTRFile* lpTRFile = m_lpTree->lpGetTRFile();

    TR123_MESH  MeshCopy;
    
    if(lpTRFile->boImportMeshFrom3DS(&MeshCopy,(TCHAR*)(LPCTSTR)FileDialog.GetPathName(),false))
        {
        CTrLstMsh  Msh(&MeshCopy,lpTRFile);

        CTrStatMsh* lpStatMsh = new CTrStatMsh;

        lpStatMsh->CTrLstMsh::vRelease();
        lpStatMsh->CTrLstMsh::vStealData(Msh);

        TR123_MESH* lpMesh = const_cast<TR123_MESH*>(lpStatMsh->lpGetMesh(0));

        lpMesh->lCollisionSize = 0;
        CTRFile::vComputeMeshExtents(lpMesh);

        CBoundingBox    BB;

        lpStatMsh->vSetObjectId(0);
        lpStatMsh->vComputeBoundingBox(BB,0);
        lpStatMsh->vSetBBVisibility(BB);
        lpStatMsh->vSetBBCollision(BB);
        lpStatMsh->vSetFlags(TRSTATMSH_FLAGS_VISIBLE);

        lpTRFile->GetTRFile().tr123File.lpListStaticMesh->push_back(lpStatMsh);

        vSetChanged();
        }

    CTRFile::vReleaseMeshMemory(&MeshCopy);

    if(m_boExpanded)
        {
        vExpandTree(true);
        }
}


//
//
// Import a static mesh from a .TRSMH
//
//
void    CHandleStaticMeshes::vImportFromMSH(CTrLstMsh* lpMshSrc)
{
    CTrStatMsh* lpStaticMesh = new CTrStatMsh;

    // Load the MSH file
    if(lpMshSrc == NULL)
        {
        CString strFilter = _T("TR Static Mesh files (*.trsmh)|*.trsmh||");

        CFileDialog FileDialog(TRUE,_T(""),_T(""),OFN_HIDEREADONLY,strFilter);

        if(FileDialog.DoModal() != IDOK)
            {
            delete lpStaticMesh;
            return;
            }

        CWaitCursor dummy;

        CTrStatMsh* Msh = CTrStatMsh::objLoad(FileDialog.GetPathName());

        if(Msh)
            {
            lpStaticMesh->vRelease();
            lpStaticMesh->vStealData(*Msh);
            delete Msh;
            }
        else
            {
            delete lpStaticMesh;
            AfxMessageBox(_T("Could not import this static mesh because of a version mismatch:\n\n - the static mesh was produced with an old TRViewer release and its format is not supported anymore !"),MB_ICONSTOP | MB_TOPMOST);
            return;
            }
        }
    else
        {
        lpStaticMesh->vRelease();
        lpStaticMesh->vStealData(*(static_cast<CTrStatMsh*>(lpMshSrc)));
        }

    CWaitCursor dummy;

    CTRFile* lpTRFile = m_lpTree->lpGetTRFile();

    lpTRFile->GetTRFile().tr123File.lpListStaticMesh->push_back(lpStaticMesh);

    vSetChanged();

    if(m_boExpanded)
        {
        vExpandTree(true);
        }
}
