#include "stdafx.h"

#include <GL/glut.h>
#include "OpenGLWnd1.h"
#include "TextureTools.h"

#include "AnimData.h"
#include "AnimKey.h"
#include "AnimTrack.h"
#include "AnimMgr.h"


#define COMPUTE_LIGHTING2(numvert)                                                      \
        fIntens = 1.0f;                                                                 \
        if(boInverseIntens  &&  !lpLstLighting) fIntens = 0.5f;                         \
        if(lpLstLighting)                                                               \
            {                                                                           \
            fIntens = lpLstLighting[numvert] / 8192.0f;                                 \
            }                                                                           \
        if(boInverseIntens) fIntens = 1.0f - fIntens;                                   \


inline void vComputeUV(const TR2_OBJECT_TEXTURE_VERT* lpVertTexture,float* lpfU,float* lpfV)
{
    *lpfU = lpVertTexture->bXPixel; *lpfU = *lpfU + lpVertTexture->cXCoordinate;
    *lpfV = lpVertTexture->bYPixel; *lpfV = *lpfV + lpVertTexture->cYCoordinate;

    *lpfU = *lpfU / 255.0f;
    *lpfV = *lpfV / 255.0f;
}


inline void vSetTexturedVertex(TR2_VERTEX* lpVertex,float fLighting,float fU,float fV)
{
    float       fX =  (float)lpVertex->sX;
    float       fY = -(float)lpVertex->sY;
    float       fZ = -(float)lpVertex->sZ;

    float fR = fLighting, fG = fLighting, fB = fLighting;

    glColor3f( fR,fG,fB );
    glTexCoord2f( fU,fV );
    glVertex3f( fX,fY,fZ );
}


void    COpenGLWnd::vRenderMeshVertices(const TR123_MESH* lpMesh)
{
    glPointSize(2.0f);
    glColor3f(1.0f,1.0f,1.0f);

    for(bit16 sI = 0; sI < lpMesh->sNumVertices; sI++)
        {
        TR2_VERTEX* lpVertex = &lpMesh->lpVertices[sI];

        float fX =  (float)lpVertex->sX;
        float fY = -(float)lpVertex->sY;
        float fZ = -(float)lpVertex->sZ;

        if(m_boPickingMode) glLoadName(sI+1);

        glBegin(GL_POINTS);
        glVertex3f(fX,fY,fZ);
        glEnd();
        }
}


void    COpenGLWnd::vDrawBoundingBox(CBoundingBox& BB, CVect3D vecColor)
{
    glDisable(GL_TEXTURE_2D);

    glBegin(GL_LINES);
    glColor3f( (float) vecColor[0], (float) vecColor[1], (float) vecColor[2] );

    const CVect3D& v1 = BB.GetBBLo();
    const CVect3D& v2 = BB.GetBBHi();

    glVertex3f( (float) v1[0], (float) v1[1], (float) v1[2]);
    glVertex3f( (float) v2[0], (float) v1[1], (float) v1[2]);
    glVertex3f( (float) v2[0], (float) v1[1], (float) v1[2]);
    glVertex3f( (float) v2[0], (float) v2[1], (float) v1[2]);
    glVertex3f( (float) v2[0], (float) v2[1], (float) v1[2]);
    glVertex3f( (float) v1[0], (float) v2[1], (float) v1[2]);
    glVertex3f( (float) v1[0], (float) v2[1], (float) v1[2]);
    glVertex3f( (float) v1[0], (float) v1[1], (float) v1[2]);

    glVertex3f( (float) v1[0], (float) v1[1], (float) v2[2]);
    glVertex3f( (float) v2[0], (float) v1[1], (float) v2[2]);
    glVertex3f( (float) v2[0], (float) v1[1], (float) v2[2]);
    glVertex3f( (float) v2[0], (float) v2[1], (float) v2[2]);
    glVertex3f( (float) v2[0], (float) v2[1], (float) v2[2]);
    glVertex3f( (float) v1[0], (float) v2[1], (float) v2[2]);
    glVertex3f( (float) v1[0], (float) v2[1], (float) v2[2]);
    glVertex3f( (float) v1[0], (float) v1[1], (float) v2[2]);

    glVertex3f( (float) v1[0], (float) v1[1], (float) v1[2]);
    glVertex3f( (float) v1[0], (float) v1[1], (float) v2[2]);
    glVertex3f( (float) v2[0], (float) v1[1], (float) v1[2]);
    glVertex3f( (float) v2[0], (float) v1[1], (float) v2[2]);
    glVertex3f( (float) v1[0], (float) v2[1], (float) v1[2]);
    glVertex3f( (float) v1[0], (float) v2[1], (float) v2[2]);
    glVertex3f( (float) v2[0], (float) v2[1], (float) v1[2]);
    glVertex3f( (float) v2[0], (float) v2[1], (float) v2[2]);

    glEnd();

    glEnable(GL_TEXTURE_2D);
}


void COpenGLWnd::vRenderFace(TR2_VERTEX* lpVertices,const TR2_OBJECT_TEXTURE* lpOText,bitu16* lpFace,WORD wTextureIndex,bit16* lpLstLighting,bool boInverseIntens,bool boQuad,bool boWireframe)
{
    BYTE    bNumVertices    = boQuad ? 3 : 2;

    TR2_OBJECT_TEXTURE_VERT* lpVertTexture;
    float   fU, fV, fIntens;

    const TR2_OBJECT_TEXTURE* lpTexture   = &lpOText[wTextureIndex];
    lpVertTexture = (TR2_OBJECT_TEXTURE_VERT*)lpTexture->Vertices;

    if(m_lOldTextureBinded != m_LstTextId[lpTexture->wTile])
		{
		m_lOldTextureBinded = m_LstTextId[lpTexture->wTile];
		glBindTexture(GL_TEXTURE_2D, m_lOldTextureBinded);
		}

    if(boQuad) glBegin( boWireframe ? GL_LINE_LOOP : GL_QUADS ); else glBegin( boWireframe ? GL_LINE_LOOP : GL_TRIANGLES );

    for(BYTE bI = bNumVertices+1; bI != 0; bI--)
        {
        vComputeUV(&lpVertTexture[bI-1],&fU,&fV);
        COMPUTE_LIGHTING2(lpFace[bI-1])
        vSetTexturedVertex(&lpVertices[lpFace[bI-1]],fIntens,fU,fV);
        }

    glEnd();
}


void COpenGLWnd::vRenderFaceArray(DWORD dwNumFaces,const TR123_MESH* lpMesh,const TR2_OBJECT_TEXTURE* lpOText,void* lpArray,bool boInversIntens,bool boQuad,bool boWireframe,ListFacePtr& ChromaFaces, ListFacePtr& AlphaFaces)
{
    if(!dwNumFaces) return;

    TR2_VERTEX* lpVertices      = lpMesh->lpVertices;
    TR2_FACE3*  lpTriangles     = (TR2_FACE3*) lpArray;
    TR2_FACE4*  lpRectangles    = (TR2_FACE4*) lpArray;

    bit16*      lpLstLighting   = NULL;
    DWORD       dwIndex         = 0;

    if(lpMesh->Lighting.LightingType == MESH_LIGHTING_INTERNAL)
        {
        lpLstLighting = lpMesh->Lighting.lpLights;
        }

    while(dwNumFaces--)
        {
        bitu16* lpFace;
        WORD    wTextureIndex;
        void*   lpFacePtr;
        bool    boDontRender = false;

        if(boQuad)
            {
            TR2_FACE4* lpRectangle = &lpRectangles[dwNumFaces];
            lpFacePtr = lpRectangle;
            lpFace = lpRectangle->awVertices;
            wTextureIndex = lpRectangle->wTextureIndex;
            }
        else
            {
            TR2_FACE3* lpTriangle = &lpTriangles[dwNumFaces];
            lpFacePtr = lpTriangle;
            lpFace = lpTriangle->awVertices;
            wTextureIndex = lpTriangle->wTextureIndex;
            }

        const TR2_OBJECT_TEXTURE* lpTexture   = &lpOText[wTextureIndex];

        if(lpTexture->wAttribute & 2)       { AlphaFaces.push_back(lpFacePtr);    boDontRender = true; }
        else if(lpTexture->wAttribute & 1)  { ChromaFaces.push_back(lpFacePtr);   boDontRender = true; }

        if(!boDontRender) vRenderFace(lpVertices,lpOText,lpFace,wTextureIndex,lpLstLighting,boInversIntens,boQuad,boWireframe);

        dwIndex++;
        }
}


void COpenGLWnd::vRenderFaceList(const TR123_MESH* lpMesh,const TR2_OBJECT_TEXTURE* lpOText,ListFacePtr& ListFaces,bool boInversIntens,bool boQuad,bool boWireframe)
{
    DWORD dwNumFaces = ListFaces.size();

    if(!dwNumFaces) return;

    TR2_VERTEX* lpVertices      = lpMesh->lpVertices;

    bit16*      lpLstLighting   = NULL;
    DWORD       dwIndex         = 0;

    if(lpMesh->Lighting.LightingType == MESH_LIGHTING_INTERNAL)
        {
        lpLstLighting = lpMesh->Lighting.lpLights;
        }

    while(dwNumFaces--)
        {
        bitu16* lpFace;
        WORD    wTextureIndex;

        if(boQuad)
            {
            TR2_FACE4* lpRectangle = (TR2_FACE4*)ListFaces[dwIndex];
            lpFace = lpRectangle->awVertices;
            wTextureIndex = lpRectangle->wTextureIndex;
            }
        else
            {
            TR2_FACE3* lpTriangle = (TR2_FACE3*)ListFaces[dwIndex];
            lpFace = lpTriangle->awVertices;
            wTextureIndex = lpTriangle->wTextureIndex;
            }

        vRenderFace(lpVertices,lpOText,lpFace,wTextureIndex,lpLstLighting,boInversIntens,boQuad,boWireframe);

        dwIndex++;
        }
}


void    COpenGLWnd::vRenderMesh(CTrLstMsh* lpTrMsh,DWORD dwNumMesh,bool boShowCollisionSphere,bool boWireframe,bool boShowVertices)
{
    const TR123_MESH* lpMesh = lpTrMsh->lpGetMesh(dwNumMesh);

    if(!lpMesh->boIsDummy)
        {
        ListFacePtr   ChromaFacesTri, ChromaFacesQuad;
        ListFacePtr   AlphaFacesTri,  AlphaFacesQuad;

        // normale faces
        glEnable(GL_TEXTURE_2D);

        vRenderFaceArray(lpMesh->sNumTexturedTriangles,lpMesh,lpTrMsh->lpGetFragText(),lpMesh->lpTexturedTriangles,false,false,boWireframe,ChromaFacesTri,AlphaFacesTri);
        vRenderFaceArray(lpMesh->sNumTexturedRectangles,lpMesh,lpTrMsh->lpGetFragText(),lpMesh->lpTexturedRectangles,false,true,boWireframe,ChromaFacesQuad,AlphaFacesQuad);

        // Chroma-keyed faces
        glEnable(GL_ALPHA_TEST);

        if(ChromaFacesTri.size())
            vRenderFaceList(lpMesh,lpTrMsh->lpGetFragText(),ChromaFacesTri,false,false,boWireframe);

        if(ChromaFacesQuad.size())
            vRenderFaceList(lpMesh,lpTrMsh->lpGetFragText(),ChromaFacesQuad,false,true,boWireframe);

        glDisable(GL_ALPHA_TEST);

        // Alpha faces
        glBlendFunc(GL_ONE,GL_ONE_MINUS_SRC_COLOR);
        glDepthMask(GL_FALSE);

        if(AlphaFacesTri.size())
            vRenderFaceList(lpMesh,lpTrMsh->lpGetFragText(),AlphaFacesTri,true,false,boWireframe);

        if(AlphaFacesQuad.size())
            vRenderFaceList(lpMesh,lpTrMsh->lpGetFragText(),AlphaFacesQuad,true,true,boWireframe);

        glDepthMask(GL_TRUE);
        glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);

        glDisable(GL_TEXTURE_2D);
        }
    else
        {
        glDisable(GL_TEXTURE_2D);

        glColor3f( 0.0f, 0.0f, 1.0f );
        glutWireCube(20);
        }

    // Show collision sphere
    if(boShowCollisionSphere)
        {
        glPushMatrix();
        glTranslatef(+lpMesh->Center.sX,-lpMesh->Center.sY,-lpMesh->Center.sZ);

        glColor3f( 1.0f, 1.0f, 1.0f );
        glutWireSphere(lpMesh->lCollisionSize,10,10);

        glPopMatrix();
        }

    // Show origin plane
    if(m_boShowAxis)
        {
        float fExtX = lpMesh->fMaxX - lpMesh->fMinX;
        float fExtY = lpMesh->fMaxY - lpMesh->fMinY;
        float fExtZ = lpMesh->fMaxZ - lpMesh->fMinZ;
        float fExt = fExtX;

        if(fExt < fExtY) fExt = fExtY;
        if(fExt < fExtZ) fExt = fExtZ;

        glBegin( GL_QUADS );
        glColor4f( 0.0f, 1.0f, 0.0f, 0.3f );
        glVertex3f( -fExt ,0.0f ,-fExt );
        glVertex3f( -fExt ,0.0f ,+fExt );
        glVertex3f( +fExt ,0.0f ,+fExt );
        glVertex3f( +fExt ,0.0f ,-fExt );
        glEnd();

        glBegin(GL_LINES);
        glColor4f( 1.0f, 0.0f, 0.0f, 0.5f );
        glVertex3f( -fExt ,1.0f ,0.0f );
        glVertex3f( +fExt ,1.0f ,0.0f );
        glVertex3f(  0.0f ,1.0f ,-fExt );
        glVertex3f(  0.0f ,1.0f ,+fExt );
        glEnd();
        }

    // Show vertices
    if(boShowVertices) vRenderMeshVertices(lpMesh);

    // Show selected vertex
    if(m_dwShowNumVertex != -1)
        {
        glPointSize(2.0f);
        glColor3f(1.0f,0.0f,0.0f);
        TR2_VERTEX* lpVertex = &lpMesh->lpVertices[m_dwShowNumVertex];

        float fX =  (float)lpVertex->sX;
        float fY = -(float)lpVertex->sY;
        float fZ = -(float)lpVertex->sZ;

        glDisable( GL_DEPTH_TEST );
        glBegin(GL_POINTS);
        glVertex3f(fX,fY,fZ);
        glEnd();
        }

    // Show bounding box (for static meshes only)
    if(m_trObjType == TrObj_Type_StaticMesh)
        {
        CVect3D vecWhite((FLOAT)1.0,(FLOAT)1.0,(FLOAT)1.0);
        CVect3D vecRed((FLOAT)1.0,(FLOAT)0.0,(FLOAT)0.0);

        if(m_boShowStaticBB1)
            {
            if(m_boStaticBB1IsSel)
                vDrawBoundingBox(static_cast<CTrStatMsh*>(lpTrMsh)->GetBBVisibility(),vecRed);
            else
                vDrawBoundingBox(static_cast<CTrStatMsh*>(lpTrMsh)->GetBBVisibility(),vecWhite);
            }

        if(m_boShowStaticBB2)
            {
            if(m_boStaticBB2IsSel)
                vDrawBoundingBox(static_cast<CTrStatMsh*>(lpTrMsh)->GetBBCollision(),vecRed);
            else
                vDrawBoundingBox(static_cast<CTrStatMsh*>(lpTrMsh)->GetBBCollision(),vecWhite);
            }
        }
}


void    COpenGLWnd::vRenderMoveable(CTrMvb* lpTrMvb)
{
    TR2_MESHTREE* lpMeshTree = lpTrMvb->lpGetMeshTree();

    CAnimMgr*   lpAnimMgr = lpTrMvb->lpGetAnimMgr();

    if(m_boAnimShow  &&  lpAnimMgr->lpGetCurrentTrack())
        {
        CBoundingBox BBKey;

        lpAnimMgr->lpGetCurrentTrack()->vGetInterpolatedBB(&BBKey);

        vDrawBoundingBox(BBKey);
        }

    for(DWORD dwI = 0; dwI < lpTrMvb->dwGetNumMeshes(); dwI++)
        {
        const TR123_MESH* lpMesh = lpTrMvb->lpGetMesh(dwI);

        if(dwI != 0)
            {
            bitu32  dwStack = lpMeshTree->lCoord; lpMeshTree++;
            bit32   lTransX = lpMeshTree->lCoord; lpMeshTree++;
            bit32   lTransY = lpMeshTree->lCoord; lpMeshTree++;
            bit32   lTransZ = lpMeshTree->lCoord; lpMeshTree++;

            if(dwStack & 1) glPopMatrix();
            if(dwStack & 2) glPushMatrix();

            glTranslatef((float)lTransX,(float)-lTransY,(float)-lTransZ);
            }

        if(m_boAnimShow  &&  lpAnimMgr->lpGetCurrentTrack())
            {
            CAnimData   AnimData;

            lpAnimMgr->lpGetCurrentTrack()->vGetInterpolatedData(dwI,&AnimData);

            const CVect3D& vecTrans = AnimData.vecGetTrans();

            glTranslatef((float)vecTrans[0],(float)vecTrans[1],(float)vecTrans[2]);

            CMat4x4 Mat;

            const CQuaternion& Quat = AnimData.quatGetQuaternion();

            Quat.vToRotation(&Mat);

            GLfloat MatGL[16];

            MatGL[0] = Mat[0][0];
            MatGL[1] = Mat[1][0];
            MatGL[2] = Mat[2][0];
            MatGL[3] = Mat[3][0];
            MatGL[4] = Mat[0][1];
            MatGL[5] = Mat[1][1];
            MatGL[6] = Mat[2][1];
            MatGL[7] = Mat[3][1];
            MatGL[8] = Mat[0][2];
            MatGL[9] = Mat[1][2];
            MatGL[10] = Mat[2][2];
            MatGL[11] = Mat[3][2];
            MatGL[12] = Mat[0][3];
            MatGL[13] = Mat[1][3];
            MatGL[14] = Mat[2][3];
            MatGL[15] = Mat[3][3];

            glMultMatrixf(MatGL);
            }

        if(lpMesh) vRenderMesh(lpTrMvb,dwI,false,false,m_boShowVertices);
        }
}


void    COpenGLWnd::vRenderSprite(const WORLD_POSITIONNING* lpWorldP,const TR2_OBJECT_TEXTURE* lpText,bool boAlign /*= false*/)
{
    TR2_OBJECT_TEXTURE_VERT* lpVertTexture = const_cast<TR2_OBJECT_TEXTURE_VERT*>(lpText->Vertices);

    TR2_VERTEX  Vert[4];

    if(!boAlign)
        {
        Vert[0].sX = lpWorldP->sLeftSide;
        Vert[0].sY = lpWorldP->sTopSide;
        Vert[0].sZ = 0;

        Vert[3].sX = lpWorldP->sLeftSide;
        Vert[3].sY = lpWorldP->sBottomSide;
        Vert[3].sZ = 0;

        Vert[2].sX = lpWorldP->sRightSide;
        Vert[2].sY = lpWorldP->sBottomSide;
        Vert[2].sZ = 0;

        Vert[1].sX = lpWorldP->sRightSide;
        Vert[1].sY = lpWorldP->sTopSide;
        Vert[1].sZ = 0;
        }
    else
        {
        Vert[0].sX = Vert[0].sY = Vert[0].sZ = 0;

        Vert[3].sX = Vert[3].sZ = 0;
        Vert[3].sY = lpWorldP->sBottomSide - lpWorldP->sTopSide;

        Vert[2].sX = lpWorldP->sRightSide - lpWorldP->sLeftSide;
        Vert[2].sY = lpWorldP->sBottomSide - lpWorldP->sTopSide;
        Vert[2].sZ = 0;

        Vert[1].sX = lpWorldP->sRightSide - lpWorldP->sLeftSide;
        Vert[1].sY = Vert[1].sZ = 0;
        }

    glEnable(GL_TEXTURE_2D);

    if(m_lOldTextureBinded != m_LstTextId[lpText->wTile])
		{
		m_lOldTextureBinded = m_LstTextId[lpText->wTile];
		glBindTexture(GL_TEXTURE_2D, m_lOldTextureBinded);
		}

    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);

    glBegin( GL_QUADS );

    for(BYTE bI = 4; bI != 0; bI--)
        {
        float fU, fV;
        vComputeUV(&lpVertTexture[bI-1],&fU,&fV);
        vSetTexturedVertex(&Vert[bI-1],1.0f,fU,fV);
        }

    glEnd();

    glDisable(GL_TEXTURE_2D);

    if(m_boSprSeqShowBorder)
        {
        glDisable(GL_DEPTH_TEST);

        glBegin(GL_LINE_LOOP);
        glColor3f( 1.0f, 1.0f, 1.0f );

        glVertex3f( Vert[3].sX, -Vert[3].sY, -Vert[3].sZ );
        glVertex3f( Vert[2].sX, -Vert[2].sY, -Vert[2].sZ );
        glVertex3f( Vert[1].sX, -Vert[1].sY, -Vert[1].sZ );
        glVertex3f( Vert[0].sX, -Vert[0].sY, -Vert[0].sZ );
        glEnd();

        glEnable(GL_DEPTH_TEST);
        }
}


void    COpenGLWnd::vRenderSprSeq(CTrLstSpr* lpSprSeq)
{
    static float fSprNum = 0.0f;

    const WORLD_POSITIONNING* lpWorldP  = lpSprSeq->lpGetWorldPositionning();
    const TR2_OBJECT_TEXTURE* lpText    = lpSprSeq->lpGetFragText();

    if(m_boSprSeqTypeAnimated)
        {
        fSprNum += (m_fSprSeqSpeed / m_fFrameRate);
        if(fSprNum >= lpSprSeq->dwGetNumSprites()) fSprNum = 0.0f;

        DWORD dwSprNum = (DWORD) fSprNum;

        vRenderSprite(&lpWorldP[dwSprNum],&lpText[lpWorldP[dwSprNum].wFragIndex]);
        }
    else
        {
        float   fWidth      = 0;
        float   fY          = 0;
        float   fMaxY       = 0;
        float   fWidthMax   = 0;
        float   fWidthTot   = 0;
        float   fHeightTot  = 0.0f;

        // Calculate the width of the rectangle in which to display the sprite sequence
        for(DWORD dwI = 0; dwI < lpSprSeq->dwGetNumSprites(); dwI++)
            {
            fWidthTot += lpWorldP[dwI].sRightSide - lpWorldP[dwI].sLeftSide;
            
            if(fWidthMax < lpWorldP[dwI].sRightSide - lpWorldP[dwI].sLeftSide) fWidthMax = lpWorldP[dwI].sRightSide - lpWorldP[dwI].sLeftSide;
            }

        if(fWidthTot > m_fSprSeqListWidth)
            {
            fWidthTot = m_fSprSeqListWidth;
            }

        if(fWidthMax > fWidthTot) fWidthTot = fWidthMax;

        // Calculate the height of the rectangle in which to display the sprite sequence
        for(dwI = 0; dwI < lpSprSeq->dwGetNumSprites(); dwI++)
            {
            if(fMaxY < lpWorldP[dwI].sBottomSide - lpWorldP[dwI].sTopSide) fMaxY = lpWorldP[dwI].sBottomSide - lpWorldP[dwI].sTopSide;

            fWidth += lpWorldP[dwI].sRightSide - lpWorldP[dwI].sLeftSide;

            if(fWidth >= fWidthTot)
                {
                fWidth      = 0;
                fHeightTot += fMaxY;
                fMaxY       = 0;
                }
            }

        fHeightTot += fMaxY;

        fWidth  = 0.0f;
        fY      = 0.0f;
        fMaxY   = 0.0f;

        // Draw the sprites
        for(dwI = 0; dwI < lpSprSeq->dwGetNumSprites(); dwI++)
            {
            glPushMatrix();
            glTranslatef(fWidth - fWidthTot / 2.0f,-fY + fHeightTot / 2.0f,0);
            vRenderSprite(&lpWorldP[dwI],&lpText[lpWorldP[dwI].wFragIndex],true);
            glPopMatrix();

            if(fMaxY < lpWorldP[dwI].sBottomSide - lpWorldP[dwI].sTopSide) fMaxY = lpWorldP[dwI].sBottomSide - lpWorldP[dwI].sTopSide;

            fWidth += lpWorldP[dwI].sRightSide - lpWorldP[dwI].sLeftSide;

            if(fWidth >= m_fSprSeqListWidth)
                {
                fWidth = 0;
                fY += fMaxY;
                fMaxY = 0;
                }
            }
        }
}

