#include "stdafx.h"

#include "Split.h"
#include "ResMgr.h"


CResItem::CResItem(TCHAR* lpszName, TCHAR* lpszLevelName)
{
    lstrcpyn(m_szName,lpszName,128);
    lstrcpyn(m_szLevelName,lpszLevelName,128);
}


void    CResItem::vAddObject(enumTrObjType ObjType, DWORD dwObjId)
{
    m_ListObjects.push_back(CResObject(ObjType, dwObjId));
}


CResMgr::CResMgr()
{
    lstrcpy(m_szPath, _T(".\\"));
}


CResMgr::~CResMgr()
{
    vRelease();
}


void    CResMgr::vRelease()
{
    for(DWORD dwI = 0; dwI < __RESMGR_MAXTR__; dwI++)
        {
        m_Menu[dwI].DestroyMenu();
        for(ListResItem::iterator theIter = m_List[dwI].begin(); theIter != m_List[dwI].end(); theIter++)
            {
            delete *theIter;
            }
        m_List[dwI].clear();
        for(ListSfx::iterator theIter2 = m_ListSFX[dwI].begin(); theIter2 != m_ListSFX[dwI].end(); theIter2++)
            {
            delete *theIter2;
            }
        m_ListSFX[dwI].clear();
        }
}


void    CResMgr::vInitialize(TCHAR* lpszPath)
{
    vRelease();

    if(lpszPath) lstrcpyn(m_szPath, lpszPath, MAX_PATH);

    CString cstrMenuResource(m_szPath);

    cstrMenuResource += _T("MenuResource.txt");

    for(DWORD dwI = 0; dwI < __RESMGR_MAXTR__; dwI++)
        {
        // Create the "add resource" menu
        m_Menu[dwI].CreatePopupMenu();
    
        CMenu   SubMenu;

        SubMenu.CreatePopupMenu();
    
        m_Menu[dwI].AppendMenu(MF_POPUP,(UINT)SubMenu.m_hMenu,_T("Add resource"));

        enumTRType trType = TR_UNKNOWN;

        switch(dwI)
            {
            case __RESMGR_TR1__: trType = TR1_FILE; break;
            case __RESMGR_TR2__: trType = TR2_FILE; break;
            case __RESMGR_TR3__: trType = TR3_FILE; break;
            }

        vParseFile_MenuResource(SubMenu, (LPTSTR)(LPCTSTR)cstrMenuResource, trType, m_List[dwI]);

        SubMenu.Detach();

        // Load the SFXs
        CString cstrSFXFilename(m_szPath);

        switch(dwI)
            {
            case __RESMGR_TR1__:        cstrSFXFilename += _T("SFX_TR1.txt"); break;
            case __RESMGR_TR1GOLD__:    cstrSFXFilename += _T("SFX_TR1Gold.txt"); break;
            case __RESMGR_TR2__:        cstrSFXFilename += _T("SFX_TR2.txt"); break;
            case __RESMGR_TR2GOLD__:    cstrSFXFilename += _T("SFX_TR2Gold.txt"); break;
            case __RESMGR_TR3__:        cstrSFXFilename += _T("SFX_TR3.txt"); break;
            case __RESMGR_TR3GOLD__:    cstrSFXFilename += _T("SFX_TR3Gold.txt"); break;
            }

        vParseFile_SFX(cstrSFXFilename, m_ListSFX[dwI]);
        }
}


void    CResMgr::vSetDataPath(BYTE bTRResId, CString& cstrDataPath)
{
    m_DataPath[bTRResId] = cstrDataPath;

    CString cstrName;

    switch(bTRResId)
        {
        case __RESMGR_TR1__:        cstrName = _T("TR1");       break;
        case __RESMGR_TR1GOLD__:    cstrName = _T("TR1Gold");   break;
        case __RESMGR_TR2__:        cstrName = _T("TR2");       break;
        case __RESMGR_TR2GOLD__:    cstrName = _T("TR2Gold");   break;
        case __RESMGR_TR3__:        cstrName = _T("TR3");       break;
        case __RESMGR_TR3GOLD__:    cstrName = _T("TR3Gold");   break;
        }

    cstrName += _T(" Data Path");

    AfxGetApp()->WriteProfileString(_T("Settings"),cstrName,m_DataPath[bTRResId]);
}


void    CResMgr::vParseFile_MenuResource(CMenu& Menu,TCHAR* lpszFilename, enumTRType trType, ListResItem& List)
{
    FILE*   hFile = _tfopen(lpszFilename,_T("rt"));

    if(hFile ==  NULL) return;

    bool    boSectionFound = false;
    DWORD   dwMenuId = 0;

    CMenu*  lpStack[10];
    DWORD   dwStackIndex = 1;

    ZeroMemory(lpStack,sizeof(CMenu*) * 10);

    lpStack[0] = &Menu;

    while(!feof(hFile))
        {
        TCHAR   szBuf[1025];

        _fgetts(szBuf,1024,hFile);

        if(!boSectionFound)
            {
            _tcsupr(szBuf);
            switch(trType)
                {
                case TR1_FILE:
                    {
                    if(_tcsncmp(szBuf,_T("[TR1]"),5) == 0) boSectionFound = true;
                    break;
                    }
                case TR2_FILE:
                    {
                    if(_tcsncmp(szBuf,_T("[TR2]"),5) == 0) boSectionFound = true;
                    break;
                    }
                case TR3_FILE:
                    {
                    if(_tcsncmp(szBuf,_T("[TR3]"),5) == 0) boSectionFound = true;
                    break;
                    }
                }
            }
        else
            {
            CSplit Split(szBuf,":",TRUE);

            if(Split.lUBound() == 0)
                {
                if(strnicmp(Split[0],"[END]",5) == 0) break;

                // Menu delimiter
                char*   lpsz1 = strchr(szBuf,'[');
                char*   lpsz2 = NULL;

                if(lpsz1) lpsz2 = strchr(lpsz1 + 1,']');

                if(lpsz1 && lpsz2)
                    {
                    *lpsz2 = 0;
                
                    dwStackIndex = lpsz1 - szBuf;

                    if(lpStack[dwStackIndex])
                        {
                        lpStack[dwStackIndex]->Detach();
                        delete lpStack[dwStackIndex];
                        }

                    lpStack[dwStackIndex] = new CMenu;
                    lpStack[dwStackIndex]->CreatePopupMenu();

                    lpStack[dwStackIndex-1]->AppendMenu(MF_POPUP,(UINT)lpStack[dwStackIndex]->m_hMenu,lpsz1 + 1);
                    }
                }
            else
                {
                // Menu item
                char* lpsz1 = Split[0];

                while(*lpsz1 == 32  ||  *lpsz1 == 9) lpsz1++;

                lpStack[dwStackIndex]->AppendMenu(MF_STRING, dwMenuId++, lpsz1);

                CResItem*   lpResItem = new CResItem(lpsz1,Split[1]);

                CSplit Split2(Split[2],";",TRUE);

                for(DWORD dwI = 0; dwI < Split2.lUBound(); dwI++)
                    {
                    enumTrObjType   ObjType = TrObj_NoObj;
                    DWORD           dwObjId = atol(&(Split2[dwI][4]));

                    if(strnicmp(Split2[dwI],"mvb",3) == 0) ObjType = TrObj_Type_Moveable;
                    else if(strnicmp(Split2[dwI],"spr",3) == 0) ObjType = TrObj_Type_Sprite;
                    else if(strnicmp(Split2[dwI],"msh",3) == 0) ObjType = TrObj_Type_StaticMesh;

                    lpResItem->vAddObject(ObjType,dwObjId);
                    }

                List.push_back(lpResItem);
                }
            }
        }

    for(DWORD dwI = 1; dwI < 10; dwI++)
        {
        if(lpStack[dwI])
            {
            lpStack[dwI]->Detach();
            delete lpStack[dwI];
            }
        }

    fclose(hFile);
}


void    CResMgr::vParseFile_SFX(CString& cstrFilename, ListSfx& ListSFX)
{
    FILE*   hFile = _tfopen((LPTSTR)(LPCTSTR)cstrFilename,_T("rt"));

    if(hFile ==  NULL) return;

    long        lSoundIndex = -1;
    CResSFX*    lpResSFX = NULL;
    DWORD       dwFirstSoundIndex = 0;

    while(!feof(hFile))
        {
        TCHAR   szBuf[1025];

        _fgetts(szBuf,1024,hFile);

        CSplit Split(szBuf,"--",TRUE);

        if(Split.lUBound() > 0)
            {
            TCHAR   sz[5];

            memcpy(sz, Split[0], 4); sz[4] = 0;

            DWORD dwInternalIndex = _ttol(sz);

            if(dwInternalIndex != lSoundIndex)
                {
                CString s(&Split[1][1]);
                s = s.Left(s.GetLength() - 1);
                lpResSFX = new CResSFX((LPTSTR)(LPCTSTR)s, 1, dwFirstSoundIndex);
                for(long lI = lSoundIndex + 1; lI < (long)dwInternalIndex; lI++)
                    {
                    s.Format(_T("! Not used (id #%.3d) !"),lI);
                    ListSFX.push_back(new CResSFX((LPTSTR)(LPCTSTR)s,0,0));
                    }
                ListSFX.push_back(lpResSFX);
                lSoundIndex = dwInternalIndex;
                dwFirstSoundIndex++;
                }
            else
                {
                if(lpResSFX) lpResSFX->m_dwNumSamples++;
                dwFirstSoundIndex++;
                }
            }
        }

    fclose(hFile);
}


CString CResMgr::cstrGetSfxName(BYTE bTRResId, DWORD dwInternalIndex) const
{
    if(dwInternalIndex >= m_ListSFX[bTRResId].size())
        {
        CString s;
        s.Format(_T("! Unknown internal index #%.3d !"),dwInternalIndex);
        return CString(s);
        }

    return m_ListSFX[bTRResId][dwInternalIndex]->m_cstrName;
}


DWORD   CResMgr::dwGetSfxNumSamples(BYTE bTRResId, DWORD dwInternalIndex) const
{
    if(dwInternalIndex >= m_ListSFX[bTRResId].size()) return 0;

    return m_ListSFX[bTRResId][dwInternalIndex]->m_dwNumSamples;
}


DWORD   CResMgr::dwGetSfxFirstSoundIndex(BYTE bTRResId, DWORD dwInternalIndex) const
{
    if(dwInternalIndex >= m_ListSFX[bTRResId].size()) return 0;

    return m_ListSFX[bTRResId][dwInternalIndex]->m_dwFirstSoundIndex;
}


BYTE    CResMgr::bGetResId(enumTRType trType) const
{
    switch(trType)
        {
        case TR1_FILE:  return __RESMGR_TR1__;
        case TR2_FILE:  return __RESMGR_TR2__;
        case TR3_FILE:  return __RESMGR_TR3__;
        }

    return __RESMGR_TR2__;
}
