#include "stdafx.h"

#include "AnimData.h"
#include "AnimKey.h"
#include "AnimTrack.h"
#include "TRAnimTrack.h"


CTRAnimCommand::CTRAnimCommand(TR2_ANIM_COMMAND* lpsData, DWORD dwFrameStart /*= -1*/)
{
    m_sCommand = *lpsData++;

    m_wNumParams = 0xFFFF;

    switch(m_sCommand.sValue)
        {
        // 1
        case ANIMCMD_POSITIONREF3_3:
            {
            m_wNumParams = 3;
            break;
            }
        // 2
        case ANIMCMD_POSITIONREF2_2:
            {
            m_wNumParams = 2;
            break;
            }
        // 3
        case ANIMCMD_SLAVEANIM_0:
            {
            m_wNumParams = 0;
            break;
            }
        // 4
        case ANIMCMD_DEATHANIM_0:
            {
            m_wNumParams = 0;
            break;
            }
        // 5
        case ANIMCMD_PLAYSOUNDONFRAME_2:
            {
            m_wNumParams = 2;
            break;
            }
        // 6
        case ANIMCMD_MISCACTIONONFRAME_2:
            {
            switch(lpsData[1].sValue)
                {
                case ANIMCMD_MISC_REVERTCAMERA:
                case ANIMCMD_MISC_UNK6:
                case ANIMCMD_MISC_UNK0:
                case ANIMCMD_MISC_MAKEBUBBLE:
                case ANIMCMD_MISC_UNK7:
                case ANIMCMD_MISC_REACTIVATEACTIONKEY:
                case ANIMCMD_MISC_UNK8:
                case ANIMCMD_MISC_CAMERACTRL1:
                case ANIMCMD_MISC_CAMERACTRL2:
                case ANIMCMD_MISC_UNK1:
                case ANIMCMD_MISC_UNK2:
                case ANIMCMD_MISC_LARACHANGECLOTH:
                case ANIMCMD_MISC_UNK3:
                case ANIMCMD_MISC_UNK4:
                case ANIMCMD_MISC_HIDEOBJECT:
                case ANIMCMD_MISC_SHOWOBJECT:
                case ANIMCMD_MISC_CAMERACTRL3:
                case ANIMCMD_MISC_ADDITIONALSOUND1:
                case ANIMCMD_MISC_UNK5:
                case ANIMCMD_MISC_ADDITIONALSOUND2:
                    {
                    m_wNumParams = 2;
                    break;
                    }
                }
            break;
            }
        }

    ASSERT(m_wNumParams < ANIMCMD_NUMMAX_PARAMS);

    if(m_wNumParams != 0xFFFF)
        {
        for(WORD wI = 0; wI < m_wNumParams; wI++)
            {
            m_sParams[wI] = *lpsData++;
            if(wI == 0  &&  boFirstParamIsFrame( m_sCommand )  &&  dwFrameStart != -1) m_sParams[0].sValue -= dwFrameStart;
            }
        }
    else
        {
        CString s;

        s.Format(_T("Command %ld unknown in Anim commands list. Aborting TRViewer !!"),(long)m_sCommand.sValue);
        AfxMessageBox(s,MB_OK | MB_TOPMOST);

        abort();
        }
}


/*static*/ const bool  CTRAnimCommand::boFirstParamIsFrame( const TR2_ANIM_COMMAND sCmd )
{
    bool boIsFrame = false;

    switch(sCmd.sValue)
        {
        case ANIMCMD_PLAYSOUNDONFRAME_2:
        case ANIMCMD_MISCACTIONONFRAME_2:
            {
            boIsFrame = true;
            break;
            }
        }

    return boIsFrame;
}


const bool  CTRAnimCommand::boIsSound() const
{
    return
        m_sCommand.sValue == ANIMCMD_PLAYSOUNDONFRAME_2  ||  
        (
        m_sCommand.sValue == ANIMCMD_MISCACTIONONFRAME_2  && 
            (
                m_sParams[1].sValue == ANIMCMD_MISC_ADDITIONALSOUND1 || 
                m_sParams[1].sValue == ANIMCMD_MISC_ADDITIONALSOUND2
            )
        );
}


/*static*/ const DWORD CTRAnimCommand::dwGetNumParams( const TR2_ANIM_COMMAND sCmd )
{
    DWORD   dwNumParams = 0xFFFF;

    switch(sCmd.sValue)
        {
        // 1
        case ANIMCMD_POSITIONREF3_3:
            {
            dwNumParams = 3;
            break;
            }
        // 2
        case ANIMCMD_POSITIONREF2_2:
            {
            dwNumParams = 2;
            break;
            }
        // 3
        case ANIMCMD_SLAVEANIM_0:
            {
            dwNumParams = 0;
            break;
            }
        // 4
        case ANIMCMD_DEATHANIM_0:
            {
            dwNumParams = 0;
            break;
            }
        // 5
        case ANIMCMD_PLAYSOUNDONFRAME_2:
            {
            dwNumParams = 2;
            break;
            }
        // 6
        case ANIMCMD_MISCACTIONONFRAME_2:
            {
            dwNumParams = 2;
            break;
            }
        }

    return dwNumParams;
}


void    CTRAnimCommand::vSetCommand(TR2_ANIM_COMMAND sCommand)
{
    m_sCommand = sCommand;

    m_wNumParams = dwGetNumParams( m_sCommand );
}


const void  CTRStateChange::vRemoveAnimDispatch( const DWORD dwTag)
{
    for( TRLstAnimDispatch::iterator theIter = m_LstAnimDispatch.begin(); theIter != m_LstAnimDispatch.end(); ++theIter )
        {
        if( theIter->m_dwTag == dwTag )
            {
            m_LstAnimDispatch.erase( theIter );
            break;
            }
        }
}


const CTRAnimDispatch&  CTRStateChange::GetAnimDispatch(DWORD dwKey, DWORD& dwOtherKey, bool boKeyIsTag) const
{
    if( !boKeyIsTag )
        {
        _ASSERTE( dwKey < dwGetNumAnimDispatch() );
    
        dwOtherKey = m_LstAnimDispatch[ dwKey ].m_dwTag;
    
        return m_LstAnimDispatch[ dwKey ].m_AnimDispatch;
        }

    bool    boOk = false;
    
    for(DWORD dwAnmDisp = 0; dwAnmDisp < dwGetNumAnimDispatch(); dwAnmDisp++ )
        {
        if( m_LstAnimDispatch[ dwAnmDisp ].m_dwTag == dwKey )
            {
            boOk = true;
            break;
            }
        }

    _ASSERTE( boOk );

    dwOtherKey = dwAnmDisp;

    return m_LstAnimDispatch[ dwAnmDisp ].m_AnimDispatch;
}


const CTRStateChange& CTRAnimTrack::GetStateChangeFromStateId(DWORD dwStateId, DWORD* lpdwIndex) const
{
    DWORD dwIndex = -1;

    for(DWORD dwStateChange = 0; dwStateChange < dwGetNumStateChanges(); dwStateChange++)
        {
        if((*m_lpListStateChanges)[dwStateChange].dwGetStateId() == dwStateId)
            {
            dwIndex = dwStateChange;
            break;
            }
        }

    ASSERT(dwIndex != -1);

    if( lpdwIndex ) *lpdwIndex = dwIndex;

    return (*m_lpListStateChanges)[dwIndex];
}


CTRAnimTrack::CTRAnimTrack(DWORD dwNumKeys, DWORD dwNumFrames, DWORD dwFrameRate, FLOAT fAnimFPS) :
        CAnimTrack(dwNumKeys, dwNumFrames, dwFrameRate, fAnimFPS)
{
    vInitialize();

    m_lpListStateChanges = new TRAnimListStateChanges;
    m_lpListAnimCommands = new TRAnimListAnimCommands;
}


const void  CTRAnimTrack::vInitialize()
{
    m_dwTRStateId           = 0;
    m_sTRUnknown            = 0;
    m_lTRSpeed              = 15;
    m_fTRAcceleration       = 0.0f;
    m_lpListStateChanges    = NULL;
    m_lpListAnimCommands    = NULL;
}


const void  CTRAnimTrack::vRelease()
{
    if( m_lpListStateChanges ) m_lpListStateChanges->clear();
    if( m_lpListAnimCommands ) m_lpListAnimCommands->clear();

    delete m_lpListStateChanges;
    delete m_lpListAnimCommands;

    vInitialize();
}


DWORD CTRAnimTrack::dwGetNumAnimDispatches() const
{
    DWORD dwNum = 0;

    for(DWORD dwStateChange = 0; dwStateChange < dwGetNumStateChanges(); dwStateChange++)
        {
        dwNum += (*m_lpListStateChanges)[dwStateChange].dwGetNumAnimDispatch();
        }

    return dwNum;
}


DWORD CTRAnimTrack::dwGetNumParamsAnimCommands() const
{
    DWORD dwNum = 0;

    for(DWORD dwAnimCommand = 0; dwAnimCommand < dwGetNumAnimCommands(); dwAnimCommand++)
        {
        dwNum += (*m_lpListAnimCommands)[dwAnimCommand].wGetNumParams() + 1;
        }

    return dwNum;
}


const void  CTRAnimTrack::vRemoveAnimCommand(const DWORD dwNumAnimCommand)
{
    ASSERT(dwNumAnimCommand < dwGetNumAnimCommands());
    
    (*m_lpListAnimCommands).erase((*m_lpListAnimCommands).begin() + dwNumAnimCommand);
}


const void  CTRAnimTrack::vSwapAnimCmde( const DWORD dwAnimCmd1, const DWORD dwAnimCmd2 )
{
    DWORD   dwNumAnmCmde = dwGetNumAnimCommands();

    ASSERT( dwAnimCmd1 < dwNumAnmCmde  &&  dwAnimCmd2 < dwNumAnmCmde );

    if (dwAnimCmd1 == dwAnimCmd2 ) return;

    CTRAnimCommand  AnimCmd1 = (*m_lpListAnimCommands)[ dwAnimCmd1 ];
    (*m_lpListAnimCommands)[ dwAnimCmd1 ] = (*m_lpListAnimCommands)[ dwAnimCmd2 ];
    (*m_lpListAnimCommands)[ dwAnimCmd2 ] = AnimCmd1;
}


const void  CTRAnimTrack::vStealData( CTRAnimTrack& Track )
{
    m_dwTRStateId           = Track.m_dwTRStateId;
    m_sTRUnknown            = Track.m_sTRUnknown;
    m_lTRSpeed              = Track.m_lTRSpeed;
    m_fTRAcceleration       = Track.m_fTRAcceleration;
                            
    m_lpListStateChanges    = Track.m_lpListStateChanges;
    m_lpListAnimCommands    = Track.m_lpListAnimCommands;

    Track.vInitialize();
}


const void  CTRAnimTrack::vRemoveAnimDispatch( const DWORD dwStateId, const DWORD dwTag )
{
    DWORD   dwIndex = -1;

    CTRStateChange& StateChange = const_cast<CTRStateChange&>( GetStateChangeFromStateId( dwStateId, &dwIndex ) );

    if( dwIndex != -1 )
        {
        StateChange.vRemoveAnimDispatch( dwTag );

        if( StateChange.dwGetNumAnimDispatch() == 0 )
            {
            m_lpListStateChanges->erase( m_lpListStateChanges->begin() + dwIndex );
            }
        }
}


void    CTRAnimTrack::vPushStateChange(CTRStateChange& StateChange)
{
    StateChange.vPackAnimDispatches();

    DWORD   dwStateId = StateChange.dwGetStateId();
    TRAnimListStateChanges::iterator theIter = m_lpListStateChanges->begin();

    for(DWORD dwStateChange = 0; dwStateChange < dwGetNumStateChanges(); dwStateChange++)
        {
        CTRStateChange&   StCh = const_cast<CTRStateChange&>( GetStateChange( dwStateChange ) );

        if( dwStateId == StCh.dwGetStateId() )
            {
            for( DWORD dwAnmDisp = 0; dwAnmDisp < StateChange.dwGetNumAnimDispatch(); ++dwAnmDisp )
                {
                DWORD   dwTag_;
                const CTRAnimDispatch& AnimDispatch = StateChange.GetAnimDispatch( dwAnmDisp, dwTag_, false );

                 StCh.vPushAnimDispatch( AnimDispatch );
                }
            return;
            }

        if( dwStateId < StCh.dwGetStateId() )
            {
            m_lpListStateChanges->insert( theIter, StateChange );
            return;
            }

        ++theIter;
        }

    m_lpListStateChanges->push_back( StateChange );
}
