#include "stdafx.h"

#include "Resource.h"
#include "Mainfrm.h"
#include "DlgLevelLoad.h"
#include "DlgLevelSave.h"
#include "TextureTools.h"


//#define __NO_CONVERSION__


//----------------------------------------------------------------------------
//
//  Function:   boLoad
//
//  Synopsis:   Load a TR level, then parse it and populate the m_trFile
//              data member
//
//  Arguments:  [lpszFileName] -- file name of the level
//
//  Return:     true if the level could be loaded / parsed, else false
//
//----------------------------------------------------------------------------
bool    CTRFile::boLoad(TCHAR* lpszFileName)
{
    HANDLE  hSrcFile = INVALID_HANDLE_VALUE;
    bool    boOk = false;
    BYTE*   lpData = NULL;
    DWORD   dwDataSize = 0;
    
    CString s;
    lstrcpy(m_szFileName,lpszFileName);

    _splitpath(m_szFileName,NULL,NULL,m_szLevelName,NULL);

    ((CMainFrame*)(AfxGetApp()->m_pMainWnd))->vSetToolboxType(TOOLBOX_EMPTY,true,false);

    // Pump messages to let the toolbox window disapear
    MSG Msg;

    while(PeekMessage(&Msg,NULL,0,0,TRUE))
        {
        TranslateMessage(&Msg);
        DispatchMessage(&Msg);
        }

try
    {
    // Open the file
    hSrcFile = CreateFile(lpszFileName,GENERIC_READ,0,NULL,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);

    if(hSrcFile  ==  INVALID_HANDLE_VALUE)
        {
        s.Format(_T("File %s not found.\n"),lpszFileName);
        AfxMessageBox(s,MB_ICONSTOP | MB_OK | MB_TOPMOST);
        throw(1);
        }

    // Get the file size
    dwDataSize = GetFileSize(hSrcFile,NULL);

    if(dwDataSize  ==  0xFFFFFFFF)
        {
        dwDataSize = 0;
        s.Format(_T("Can't get the size of the file (error %ld).\n"),GetLastError());
        AfxMessageBox(s,MB_ICONSTOP | MB_OK | MB_TOPMOST);
        throw(2);
        }

    // Allocate the memory and load the data
    DWORD   dwNumRead;

    lpData = new BYTE[dwDataSize];

    ReadFile(hSrcFile,lpData,dwDataSize,&dwNumRead,NULL);

    if(dwNumRead  !=  dwDataSize)
        {
        s.Format(_T("Can't read the data (error %ld).\n"),GetLastError());
        AfxMessageBox(s,MB_ICONSTOP | MB_OK | MB_TOPMOST);
        throw(3);
        }

    boOk = true;
    throw(0);
    }
catch(...)
    {
    if(hSrcFile != INVALID_HANDLE_VALUE) CloseHandle(hSrcFile);
    }

    if(boOk)
        {
        CWaitCursor Dummy;

        // Parse the level
        boOk = boParseTRLevel(lpData,dwDataSize,lpszFileName);
        if(!boOk)
            {
            vRelease();
            AfxMessageBox(_T("Unknown file type !"),MB_ICONSTOP | MB_OK | MB_TOPMOST);
            }
        }

    delete [] lpData;

    return boOk;
}


//----------------------------------------------------------------------------
//
//  Function:   boSave
//
//  Synopsis:   Save the current TR level
//
//  Arguments:  [lpszFileName] -- file name of the level
//
//  Return:     true if the level could be saved, else false
//
//----------------------------------------------------------------------------
bool    CTRFile::boSave(LPCTSTR lpszFileName)
{
    if(m_TRType == TR_UNKNOWN) return false;

    HANDLE  hSrcFile = INVALID_HANDLE_VALUE;
    bool    boOk = false;
    BYTE*   lpData = NULL;
    DWORD   dwDataSize = 0;
    
    CString s;

    ((CMainFrame*)(AfxGetApp()->m_pMainWnd))->vSetToolboxType(TOOLBOX_EMPTY,true,false);

    // Pump messages to let the toolbox window disapear
    MSG Msg;

    while(PeekMessage(&Msg,NULL,0,0,TRUE))
        {
        TranslateMessage(&Msg);
        DispatchMessage(&Msg);
        }

try
    {
    // Open the file
    hSrcFile = CreateFile(lpszFileName,GENERIC_WRITE,0,NULL,CREATE_ALWAYS,FILE_ATTRIBUTE_NORMAL,NULL);

    if(hSrcFile  ==  INVALID_HANDLE_VALUE)
        {
        s.Format(_T("Can't create the file %s.\n"),lpszFileName);
        AfxMessageBox(s,MB_ICONSTOP | MB_OK | MB_TOPMOST);
        throw(1);
        }

    // Get the data
    if(m_TRType == TR4_FILE)
        {
        CWaitCursor Dummy;

        lpData = lpbMakeTR4Level(&dwDataSize);
        }
    else
        {
        CWaitCursor Dummy;

        lpData = lpbMakeTR123Level(&dwDataSize);
        }

    // Write the data
    DWORD   dwNbWritten;

    WriteFile(hSrcFile,lpData,dwDataSize,&dwNbWritten,NULL);

    if(dwNbWritten  !=  dwDataSize)
        {
        s.Format(_T("Can't write the data (error %ld).\n"),GetLastError());
        AfxMessageBox(s,MB_ICONSTOP | MB_OK | MB_TOPMOST);
        throw(3);
        }

    boOk = true;
    throw(0);
    }
catch(...)
    {
    if(hSrcFile != INVALID_HANDLE_VALUE) CloseHandle(hSrcFile);
    }

    if(boOk)
        {
        lstrcpy(m_szFileName,lpszFileName);
        }

    delete [] lpData;

    return boOk;
}


//----------------------------------------------------------------------------
//
//  Function:   vRelease
//
//  Synopsis:   Release the memory of a level
//
//  Arguments:  -
//
//  Return:     -
//
//----------------------------------------------------------------------------
void    CTRFile::vRelease()
{
    if(m_TRType != TR4_FILE)
        {
        TR123_FILE_STRUCT tr123File = m_trFile.tr123File;

        if(tr123File.lplpTextiles16)
            {
            for(DWORD dwI = 0; dwI < tr123File.dwNumTextiles; dwI++)
                {
                delete tr123File.lplpTextiles16[dwI];
                }
            }

        delete [] tr123File.lplpTextiles16;

        for(DWORD dwI = 0; dwI < tr123File.wNumRooms; dwI++)
            {
            TR123_ROOM* lpRoom = &tr123File.lpRooms[dwI];

            delete [] lpRoom->RoomData.lpVertices;
            delete [] lpRoom->RoomData.lpRectangles;
            delete [] lpRoom->RoomData.lpTriangles;
            delete [] lpRoom->RoomData.lpSprites;
            delete [] lpRoom->lpPortals;
            delete [] lpRoom->lpSectorList;
            delete [] lpRoom->lpLights;
            delete [] lpRoom->lpStaticMeshes;
            }

        delete [] tr123File.lpRooms;
        delete [] tr123File.lpFloorData;

        for(dwI = 0; dwI < tr123File.dwNumMeshes; dwI++)
            {
            TR123_MESH* lpMesh = tr123File.lplpMeshes[dwI];

            delete [] lpMesh->lpVertices;

            if(lpMesh->Lighting.LightingType == MESH_LIGHTING_EXTERNAL)
                delete [] lpMesh->Lighting.lpNormals;
            else
                delete [] lpMesh->Lighting.lpLights;

            delete [] lpMesh->lpTexturedRectangles;
            delete [] lpMesh->lpTexturedTriangles;
            delete [] lpMesh->lpColoredRectangles;
            delete [] lpMesh->lpColoredTriangles;

            delete lpMesh;
            }

        TRListStaticMesh* lpList = m_trFile.tr123File.lpListStaticMesh;

        if(lpList)
            for(TRListStaticMesh::iterator theIterator = lpList->begin(); theIterator != lpList->end(); theIterator++)
                {
                CTrStatMsh* lpStatMsh = *theIterator;

                delete lpStatMsh;
                }

        delete m_trFile.tr123File.lpListStaticMesh;

        TRListSpriteSeq* lpList2 = m_trFile.tr123File.lpListSpriteSeq;

        if(lpList2)
            for(TRListSpriteSeq::iterator theIterator2 = lpList2->begin(); theIterator2 != lpList2->end(); theIterator2++)
                {
                CTrLstSpr* lpSprSeq = *theIterator2;

                delete lpSprSeq;
                }

        delete m_trFile.tr123File.lpListSpriteSeq;

        TRListMoveable* lpList3 = m_trFile.tr123File.lpListMoveable;

        if(lpList3)
            for(TRListMoveable::iterator theIterator3 = lpList3->begin(); theIterator3 != lpList3->end(); theIterator3++)
                {
                CTrMvb* lpMoveable = *theIterator3;

                delete lpMoveable;
                }

        delete m_trFile.tr123File.lpListMoveable;

        TRListItem* lpList4 = m_trFile.tr123File.lpListItem;

        if(lpList4)
            for(TRListItem::iterator theIterator4 = lpList4->begin(); theIterator4 != lpList4->end(); theIterator4++)
                {
                TR123_ITEM* lpItem= *theIterator4;

                delete lpItem;
                }

        delete m_trFile.tr123File.lpListItem;

        delete [] tr123File.lplpMeshes;
        delete [] tr123File.lpMeshPointers;
        delete [] tr123File.lpAnimations;
        delete [] tr123File.lpStateChanges;
        delete [] tr123File.lpAnimDispatches;
        delete [] tr123File.lpAnimCommands;
        delete [] tr123File.lpMeshTrees;
        delete [] tr123File.lpFrames;
        delete [] tr123File.lpMoveables;
        delete [] tr123File.lpStaticMeshes;
        delete [] tr123File.lpSpriteSequences;
        delete [] tr123File.lpItems;
        delete [] tr123File.lpSpriteTextures;
        delete [] tr123File.lpAnimatedTextures;
        delete [] tr123File.lpObjectTextures;
        delete [] tr123File.lpSoundSources;
        delete [] tr123File.lpSoundDetails;
        delete [] tr123File.lpSampleIndices;
        delete [] tr123File.lpSamples;
        
        delete  tr123File.lpSndMgr;
        delete  tr123File.lpListUnusedObjText;

        delete [] tr123File.lpCameras;
        delete [] tr123File.lpBoxes;
        delete [] tr123File.lpOverlaps;
        delete [] tr123File.lpZones;
        delete [] tr123File.lpCinematicFrames;
        delete [] tr123File.lpDemoData;
        }
    else
        {
        TR4_FILE_STRUCT tr4File = m_trFile.tr4File;

        delete [] tr4File.Texture1.lpbTextData;
        delete [] tr4File.Texture2.lpbTextData;
        delete [] tr4File.Texture3.lpbTextData;
        delete [] tr4File.lpStuff;
        }

    ZeroMemory(&m_trFile,sizeof(TR_FILE));

    m_TRType = TR_UNKNOWN;
    m_szFileName[0] = 0;
}


//----------------------------------------------------------------------------
//
//  Function:   dwCalcLevelTR123DataSize
//
//  Synopsis:   Calculate the data size of a whole level for TR1/2 or 3
//
//  Arguments:  -
//
//  Return:     Data size of the level
//
//----------------------------------------------------------------------------
DWORD   CTRFile::dwCalcLevelTR123DataSize()
{
    TR123_FILE_STRUCT* lp = (TR123_FILE_STRUCT*) &m_trFile.tr123File;
    bool    boTR1 = (m_TRType == TR1_FILE);

    DWORD   dwSize =
        sizeof(bitu32)                                          // Version
      + sizeof(TR2_PALETTE8)                                    // Palette 8 bits
      + (boTR1 ? 0 : sizeof(TR2_PALETTE16))                     // Palette 16 bits
      + sizeof(bitu32)                                          // Num textiles
      + lp->dwNumTextiles * sizeof(TR2_TEXTILE8)                // Textiles 8 bits
      + (boTR1 ? 0 : lp->dwNumTextiles * sizeof(TR2_TEXTILE16)) // Textiles 16 bits
      + sizeof(bitu32)                                          // unused
      + sizeof(bitu16);                                         // Num rooms

    for(DWORD dwI = 0; dwI < lp->wNumRooms; dwI++)
        {
        dwSize += dwCalcRoomDataSize(&lp->lpRooms[dwI],m_TRType);
        }

    dwSize += 
        sizeof(bitu32)                                          // Num floor data
      + sizeof(bitu16) * lp->dwNumFloorData                     // Floor data
      + sizeof(bitu32);                                         // Num mesh datas

    DWORD dwStartingMeshOffset = dwSize;

    for(dwI = 0; dwI < lp->dwNumMeshes; dwI++)
        {
        dwSize += dwCalcMeshDataSize(lp->lplpMeshes[dwI],dwSize - dwStartingMeshOffset);
        }

    dwSize +=
        sizeof(bitu32)                                          // Num mesh pointers
      + sizeof(bitu32) * lp->dwNumMeshPointers                  // Mesh pointers
      + sizeof(bitu32)                                          // Num animations
      + sizeof(TR2_ANIMATION) * lp->dwNumAnimations             // Animations
      + sizeof(bitu32)                                          // Num state changes
      + sizeof(TR2_STATE_CHANGE) * lp->dwNumStateChanges        // State changes
      + sizeof(bitu32)                                          // Num anim dispatches
      + sizeof(TR2_ANIM_DISPATCH) * lp->dwNumAnimDispatches     // Anim dispatches
      + sizeof(bitu32)                                          // Num anim commands
      + sizeof(TR2_ANIM_COMMAND) * lp->dwNumAnimCommands        // Anim commands
      + sizeof(bitu32)                                          // Num mesh trees
      + sizeof(TR2_MESHTREE) * lp->dwNumMeshTrees               // Mesh trees
      + sizeof(bitu32)                                          // Num frames
      + sizeof(bitu16) * lp->dwNumFrames                        // Frames
      + sizeof(bitu32)                                          // Num moveables
      + sizeof(TR2_MOVEABLE) * lp->dwNumMoveables               // Moveables
      + sizeof(bitu32)                                          // Num static meshes
      + sizeof(TR2_STATICMESH) * lp->dwNumStaticMeshes          // Static meshes
      + sizeof(bitu32)                                          // Num sprite sequences
      + sizeof(TR2_SPRITE_SEQUENCE) * lp->dwNumSpriteSequences  // Sprite sequences
      + sizeof(bitu32)                                          // Num items
      + lp->dwNumItems * (boTR1 ? sizeof(TR1_ITEM) : sizeof(TR2_ITEM))  // Items
      + sizeof(bitu32)                                          // Num sprite textures
      + sizeof(TR2_SPRITE_TEXTURE) * lp->dwNumSpriteTextures    // Sprite textures
      + sizeof(bitu32)                                          // Num animated textures
      + sizeof(bitu16) * lp->dwNumAnimatedTextures              // Animated textures
      + sizeof(bitu32)                                          // Num object textures
      + sizeof(TR2_OBJECT_TEXTURE) * lp->dwNumObjectTextures    // Object textures
      + sizeof(bitu32)                                          // Num sound sources
      + sizeof(TR2_SOUND_SOURCE) * lp->dwNumSoundSources        // Sound sources
      + (boTR1 ? sizeof(TR1_SOUNDMAP) : sizeof(TR2_SOUNDMAP))   // Sound map
      + sizeof(bitu32)                                          // Num sound details
      + sizeof(TR2_SOUND_DETAILS) * lp->dwNumSoundDetails       // Sound details
      + (boTR1 ? sizeof(bitu32) : 0)                            // Num samples
      + (boTR1 ? sizeof(bitu8) * lp->dwNumSamples : 0)          // Samples
      + sizeof(bitu32)                                          // Num sample indices
      + sizeof(bitu32) * lp->dwNumSampleIndices                 // Sample indices
      + sizeof(bitu32)                                          // Num cameras
      + sizeof(TR2_CAMERA) * lp->dwNumCameras                   // Cameras
      + sizeof(bitu32)                                          // Num boxes
      + lp->dwNumBoxes * (boTR1 ? sizeof(TR1_BOX) : sizeof(TR2_BOX))    // Boxes
      + sizeof(bitu32)                                          // Num overlaps
      + sizeof(bitu16) * lp->dwNumOverlaps                      // Overlaps
      + lp->dwNumBoxes * (boTR1 ? sizeof(TR1_ZONE) : sizeof(TR2_ZONE))  // Zones
      + sizeof(TR2_LIGHTMAP)                                    // Lightmap
      + sizeof(bitu16)                                          // Num cinematic frames
      + sizeof(TR2_CINEMATIC_FRAME) * lp->wNumCinematicFrames   // Cinematic frames
      + sizeof(bitu16)                                          // Num demo data
      + sizeof(bitu8) * lp->wNumDemoData;                        // Demo data

    return dwSize;
}


//----------------------------------------------------------------------------
//
//  Function:   dwCalcLevelTR4DataSize
//
//  Synopsis:   Calculate the data size of a whole level for TR4
//
//  Arguments:  -
//
//  Return:     Data size of the level
//
//----------------------------------------------------------------------------
DWORD   CTRFile::dwCalcLevelTR4DataSize()
{
    TR4_FILE_STRUCT* lp = (TR4_FILE_STRUCT*) &m_trFile.tr4File;

    return
        sizeof(bitu32)                                          // Version
      + sizeof(bitu16)                                          // Unknown 1
      + sizeof(bitu16)                                          // Unknown 2
      + sizeof(bitu16)                                          // Unknown 3
      + sizeof(bitu32)                                          // Texture 1
      + sizeof(bitu32)
      + sizeof(bitu8) * lp->Texture1.dwTextCompSize
      + sizeof(bitu32)                                          // Texture 2
      + sizeof(bitu32)
      + sizeof(bitu8) * lp->Texture2.dwTextCompSize
      + sizeof(bitu32)                                          // Texture 3
      + sizeof(bitu32)
      + sizeof(bitu8) * lp->Texture3.dwTextCompSize
      + sizeof(bitu8) * lp->dwStuffSize;                        // Other stuff
}


//----------------------------------------------------------------------------
//
//  Function:   lpbMakeTR123Level
//
//  Synopsis:   Write (into memory) a level to the TR1/2 or 3 format
//
//  Arguments:  [lpdwDataSize] -- at return, hold the data size
//
//  Return:     pointer on data size.
//              This pointer needs to be freed after use (delete [] 'pointer')
//
//----------------------------------------------------------------------------
BYTE*   CTRFile::lpbMakeTR123Level(DWORD* lpdwDataSize)
{
    CDlgLevelSave   DlgSave;

    DlgSave.Create(IDD_LEVEL_SAVE);

    bool boKeepUnused = AfxMessageBox( _T("Do you want to keep the unused textures in the level ?"), MB_YESNO | MB_TOPMOST) == IDYES;

    // Pump messages to let the dialog box disappear
    MSG Msg;

    while(PeekMessage(&Msg,NULL,0,0,TRUE))
        {
        TranslateMessage(&Msg);
        DispatchMessage(&Msg);
        }

    vConvertToTRLevel(DlgSave, boKeepUnused);

    *lpdwDataSize = dwCalcLevelTR123DataSize();

    DlgSave.vSetSaveRange(*lpdwDataSize);

    DWORD   dwI,dwSize;
    BYTE*   lpbData = new BYTE [*lpdwDataSize];
    BYTE*   lpData = lpbData;

    ZeroMemory(lpbData,*lpdwDataSize);

    TR123_FILE_STRUCT* lp = (TR123_FILE_STRUCT*) &m_trFile.tr123File;

    bool boTR1 = (m_TRType == TR1_FILE);

    //
    // Write version
    //
    * ((bitu32*) lpData) = dwGetVersionNumber(m_TRType);    lpData += sizeof(bitu32);

    //
    // Write palettes data
    //
    if(!boTR1)
        {
        * ((TR2_PALETTE8*) lpData)  = lp->Palette8;     lpData += sizeof(TR2_PALETTE8);
        * ((TR2_PALETTE16*) lpData) = lp->Palette16;    lpData += sizeof(TR2_PALETTE16);
        }

    DlgSave.vSetSaveProgress(lpData - lpbData);

    //
    // Write textiles data
    //
    TR2_TEXTILE8    Textile8;

    memset(&Textile8, 255, sizeof(TR2_TEXTILE8));

    * ((bitu32*) lpData) = lp->dwNumTextiles; lpData += sizeof(bitu32);

    // Create the 8-bits textiles
    BYTE*   lpMap = new BYTE [32 * 32 * 32];

    ZeroMemory(lpMap,32 * 32 * 32);

    for(DWORD dwR = 0; dwR < 32; dwR++)
        for(DWORD dwG = 0; dwG < 32; dwG++)
            for(DWORD dwB = 0; dwB < 32; dwB++)
                lpMap[dwB + dwG * 32 + dwR * 32 * 32] = bFindColor((BYTE*)lp->Palette8.aRGB,dwR << 3,dwG << 3,dwB << 3);

    for(dwI = 0; dwI < lp->dwNumTextiles; dwI++)
        {
        vConvert16BitsTo8Bits((BYTE*)lp->lplpTextiles16[dwI]->awTile,256,256,Textile8.abTile,lpMap);

        * ((TR2_TEXTILE8*) lpData) = Textile8; lpData += sizeof(TR2_TEXTILE8);
        }

    delete [] lpMap;

    // Save the 16-bits textiles
    if(!boTR1)
        {
        for(dwI = 0; dwI < lp->dwNumTextiles; dwI++)
            {
            * ((TR2_TEXTILE16*) lpData) = *lp->lplpTextiles16[dwI]; lpData += sizeof(TR2_TEXTILE16);
            }
        }

    * ((bitu32*) lpData) = 0; lpData += sizeof(bitu32);   // unused

    DlgSave.vSetSaveProgress(lpData - lpbData);

    //
    // Write rooms data
    //
    * ((bitu16*) lpData) = lp->wNumRooms; lpData += sizeof(bitu16);
    for(dwI = 0; dwI < lp->wNumRooms; dwI++)
        {
        TR123_ROOM*    lpRoom = &lp->lpRooms[dwI];

        * ((TR2_ROOM_INFO*) lpData) = lpRoom->RoomInfo; lpData += sizeof(TR2_ROOM_INFO);

        bitu32* lpdwNumData = (bitu32*) lpData; lpData += sizeof(bitu32);

        TR2_ROOM_DATA* lpRoomData = (TR2_ROOM_DATA*) &lpRoom->RoomData;

        WRITE_ARRAYNUM_TR(bitu16,lpRoomData->wNumVertices,lpRoomData->lpVertices,TR1_VERTEX_ROOM,TR2_VERTEX_ROOM)
        WRITE_ARRAYNUM(bitu16,TR2_FACE4,lpRoomData->wNumRectangles,lpRoomData->lpRectangles)
        WRITE_ARRAYNUM(bitu16,TR2_FACE3,lpRoomData->wNumTriangles,lpRoomData->lpTriangles)
        WRITE_ARRAYNUM(bitu16,TR2_ROOM_SPRITE,lpRoomData->wNumSprites,lpRoomData->lpSprites)
        *lpdwNumData = (lpData - ((BYTE*)lpdwNumData + sizeof(bitu32))) / 2;
        WRITE_ARRAYNUM(bitu16,TR2_ROOM_PORTAL,lpRoom->wNumPortals,lpRoom->lpPortals)

        * ((bitu16*) lpData) = lpRoom->wNumZSectors; lpData += sizeof(bitu16);
        * ((bitu16*) lpData) = lpRoom->wNumXSectors; lpData += sizeof(bitu16);
        WRITE_ARRAY(lpRoom->wNumZSectors * lpRoom->wNumXSectors,TR2_ROOM_SECTOR,lpRoom->lpSectorList)

        * ((bit16*) lpData) = lpRoom->sAmbientIntensity1; lpData += sizeof(bit16);
        if(m_TRType != TR1_FILE)                                                            
            {
            * ((bit16*) lpData) = lpRoom->sAmbientIntensity2; lpData += sizeof(bit16);
            }
        if(m_TRType == TR2_FILE)                                                            
            {
            * ((bit16*) lpData) = lpRoom->sLightMode; lpData += sizeof(bit16);
            }

        WRITE_ARRAYNUM_TR(bitu16,lpRoom->wNumLights,lpRoom->lpLights,TR1_ROOM_LIGHT,TR2_ROOM_LIGHT)
        WRITE_ARRAYNUM_TR(bitu16,lpRoom->wNumStaticMeshes,lpRoom->lpStaticMeshes,TR1_ROOM_STATICMESH,TR2_ROOM_STATICMESH)

        * ((bit16*) lpData) = lpRoom->sAlternateRoom; lpData += sizeof(bit16);
        * ((bit16*) lpData) = lpRoom->sFlags; lpData += sizeof(bit16);
        if(m_TRType == TR3_FILE)
            {
            * ((TR2_COLOR*) lpData) = lpRoom->RoomLightColor; lpData += sizeof(TR2_COLOR);
            }
        }

    DlgSave.vSetSaveProgress(lpData - lpbData);

    //
    // Write behaviours data
    //

    WRITE_ARRAYNUM(bitu32,bitu16,lp->dwNumFloorData,lp->lpFloorData)

    DlgSave.vSetSaveProgress(lpData - lpbData);

    //
    // Write Meshes geometry data
    //
    bitu32* lpdwNumMeshData = (bitu32*) lpData; lpData += sizeof(bitu32);
    BYTE* lpbAddrStartMesh = lpData;

    bitu32* lpLstAddr = new bitu32 [lp->dwNumMeshPointers];

    for(dwI = 0; dwI < lp->dwNumMeshes; dwI++)
        {
        lpLstAddr[dwI] = (bitu32) (lpData - lpbAddrStartMesh);
        lpData = lpbWriteMesh(lp->lplpMeshes[dwI],lpData,lpbAddrStartMesh);
        }

    *lpdwNumMeshData = (lpData - lpbAddrStartMesh) / 2;

    bitu32* lpMeshPointers = new bitu32 [lp->dwNumMeshPointers];
    memcpy(lpMeshPointers,lp->lpMeshPointers,lp->dwNumMeshPointers * sizeof(bitu32));

    for(dwI = 0; dwI < lp->dwNumMeshPointers; dwI++)
        {
        lpMeshPointers[dwI] = lpLstAddr[lpMeshPointers[dwI]];
        }

    WRITE_ARRAYNUM(bitu32,bitu32,lp->dwNumMeshPointers,lpMeshPointers)

    delete [] lpMeshPointers;
    delete [] lpLstAddr;

    DlgSave.vSetSaveProgress(lpData - lpbData);

    //
    // Write meshes animations data
    //

    WRITE_ARRAYNUM(bitu32,TR2_ANIMATION,lp->dwNumAnimations,lp->lpAnimations)
    WRITE_ARRAYNUM(bitu32,TR2_STATE_CHANGE,lp->dwNumStateChanges,lp->lpStateChanges)
    WRITE_ARRAYNUM(bitu32,TR2_ANIM_DISPATCH,lp->dwNumAnimDispatches,lp->lpAnimDispatches)
    WRITE_ARRAYNUM(bitu32,TR2_ANIM_COMMAND,lp->dwNumAnimCommands,lp->lpAnimCommands)
    WRITE_ARRAYNUM(bitu32,TR2_MESHTREE,lp->dwNumMeshTrees,lp->lpMeshTrees)
    WRITE_ARRAYNUM(bitu32,bitu16,lp->dwNumFrames,lp->lpFrames)

    DlgSave.vSetSaveProgress(lpData - lpbData);

    //
    // Write meshes lists data
    //
    WRITE_ARRAYNUM(bitu32,TR2_MOVEABLE,lp->dwNumMoveables,lp->lpMoveables)
    WRITE_ARRAYNUM(bitu32,TR2_STATICMESH,lp->dwNumStaticMeshes,lp->lpStaticMeshes)

    if(m_TRType != TR3_FILE)
        {
        WRITE_ARRAYNUM(bitu32,TR2_OBJECT_TEXTURE,lp->dwNumObjectTextures,lp->lpObjectTextures)
        }

    WRITE_ARRAYNUM(bitu32,TR2_SPRITE_TEXTURE,lp->dwNumSpriteTextures,lp->lpSpriteTextures)
    WRITE_ARRAYNUM(bitu32,TR2_SPRITE_SEQUENCE,lp->dwNumSpriteSequences,lp->lpSpriteSequences)

    DlgSave.vSetSaveProgress(lpData - lpbData);

    //
    // Write other & sound data
    //
    if(m_TRType == TR1_FILE  &&  m_boTUBFile)
        {
        * ((TR2_PALETTE8*) lpData) = lp->Palette8;  lpData += sizeof(TR2_PALETTE8);
        }

    WRITE_ARRAYNUM(bitu32,TR2_CAMERA,lp->dwNumCameras,lp->lpCameras)
    WRITE_ARRAYNUM(bitu32,TR2_SOUND_SOURCE,lp->dwNumSoundSources,lp->lpSoundSources)

    DlgSave.vSetSaveProgress(lpData - lpbData);

    //
    // Write NPCs data
    //
    WRITE_ARRAYNUM_TR(bitu32,lp->dwNumBoxes,lp->lpBoxes,TR1_BOX,TR2_BOX)
    WRITE_ARRAYNUM(bitu32,bitu16,lp->dwNumOverlaps,lp->lpOverlaps)
    WRITE_ARRAY_TR(lp->dwNumBoxes,lp->lpZones,TR1_ZONE,TR2_ZONE)

    DlgSave.vSetSaveProgress(lpData - lpbData);

    //
    // Write textures data
    //
    WRITE_ARRAYNUM(bitu32,bitu16,lp->dwNumAnimatedTextures,lp->lpAnimatedTextures)

    if(m_TRType == TR3_FILE)
        {
        WRITE_ARRAYNUM(bitu32,TR2_OBJECT_TEXTURE,lp->dwNumObjectTextures,lp->lpObjectTextures)
        }

    DlgSave.vSetSaveProgress(lpData - lpbData);

    //
    // Write other data
    //
    WRITE_ARRAYNUM_TR(bitu32,lp->dwNumItems,lp->lpItems,TR1_ITEM,TR2_ITEM)

    * ((TR2_LIGHTMAP*) lpData) = lp->LightMap;  lpData += sizeof(TR2_LIGHTMAP);

    if(m_TRType == TR1_FILE  &&  !m_boTUBFile)
        {
        * ((TR2_PALETTE8*) lpData) = lp->Palette8;  lpData += sizeof(TR2_PALETTE8);
        }

    WRITE_ARRAYNUM(bitu16,TR2_CINEMATIC_FRAME,lp->wNumCinematicFrames,lp->lpCinematicFrames)
    WRITE_ARRAYNUM(bitu16,bitu8,lp->wNumDemoData,lp->lpDemoData)

    DlgSave.vSetSaveProgress(lpData - lpbData);

    //
    // Write sound data
    //
    if(m_TRType != TR1_FILE)
        {
        * ((TR2_SOUNDMAP*) lpData) = lp->SoundMap.tr2;
        lpData += sizeof(TR2_SOUNDMAP);
        }
    else
        {
        * ((TR1_SOUNDMAP*) lpData) = lp->SoundMap.tr1;
        lpData += sizeof(TR1_SOUNDMAP);
        }

    WRITE_ARRAYNUM(bitu32,TR2_SOUND_DETAILS,lp->dwNumSoundDetails,lp->lpSoundDetails)

    if(m_TRType == TR1_FILE)
        {
        WRITE_ARRAYNUM(bitu32,bitu8,lp->dwNumSamples,lp->lpSamples)
        }

    WRITE_ARRAYNUM(bitu32,bitu32,lp->dwNumSampleIndices,lp->lpSampleIndices)

    DlgSave.vSetSaveProgress(lpData - lpbData);

    return lpbData;
}


//----------------------------------------------------------------------------
//
//  Function:   lpbMakeTR4Level
//
//  Synopsis:   Write (into memory) a level to the TR4 format
//
//  Arguments:  [lpdwDataSize] -- at return, hold the data size
//
//  Return:     pointer on data size.
//              This pointer needs to be freed after use (delete [] 'pointer')
//
//----------------------------------------------------------------------------
BYTE*   CTRFile::lpbMakeTR4Level(DWORD* lpdwDataSize)
{
    *lpdwDataSize = dwCalcLevelTR4DataSize();

    DWORD   dwSize;
    BYTE*   lpbData = new BYTE [*lpdwDataSize];
    BYTE*   lpData = lpbData;

    ZeroMemory(lpbData,*lpdwDataSize);

    TR4_FILE_STRUCT* lp = (TR4_FILE_STRUCT*) &m_trFile.tr4File;

    //
    // Write version
    //
    * ((bitu32*) lpData) = dwGetVersionNumber(m_TRType); lpData += sizeof(bitu32);

    //
    // Write unknowns
    //
    * ((bitu16*) lpData) = lp->wUnknown1; lpData += sizeof(bitu16);
    * ((bitu16*) lpData) = lp->wUnknown2; lpData += sizeof(bitu16);
    * ((bitu16*) lpData) = lp->wUnknown3; lpData += sizeof(bitu16);

    //
    // Write textures
    //
    * ((bitu32*) lpData) = lp->Texture1.dwTextUncompSize;   lpData += sizeof(bitu32);
    * ((bitu32*) lpData) = lp->Texture1.dwTextCompSize;     lpData += sizeof(bitu32);
    WRITE_ARRAY(lp->Texture1.dwTextCompSize,bitu8,lp->Texture1.lpbTextData)

    * ((bitu32*) lpData) = lp->Texture2.dwTextUncompSize;   lpData += sizeof(bitu32);
    * ((bitu32*) lpData) = lp->Texture2.dwTextCompSize;     lpData += sizeof(bitu32);
    WRITE_ARRAY(lp->Texture2.dwTextCompSize,bitu8,lp->Texture2.lpbTextData)

    * ((bitu32*) lpData) = lp->Texture3.dwTextUncompSize;   lpData += sizeof(bitu32);
    * ((bitu32*) lpData) = lp->Texture3.dwTextCompSize;     lpData += sizeof(bitu32);
    WRITE_ARRAY(lp->Texture3.dwTextCompSize,bitu8,lp->Texture3.lpbTextData)

    //
    // Write stuff
    //
    WRITE_ARRAY(lp->dwStuffSize,bitu8,lp->lpStuff)

    return lpbData;
}


//----------------------------------------------------------------------------
//
//  Function:   vParseTR123File
//
//  Synopsis:   Read a TR1/2/3 level and build internal structures accordingly
//
//  Arguments:  [lpData] -- pointer on the level data
//              [dwDataSize] -- size of the data
//
//  Return:     -
//
//----------------------------------------------------------------------------
void    CTRFile::vParseTR123File(BYTE*  lpData, DWORD dwDataSize)
{
    DWORD   dwI, dwSize;
    TR123_FILE_STRUCT* lp = (TR123_FILE_STRUCT*) &m_trFile.tr123File;
    BYTE*   lpbData = lpData;

    CDlgLevelLoad   DlgLoad;

    DlgLoad.Create(IDD_LEVEL_LOAD);

    DlgLoad.vSetLoadRange(dwDataSize);

    TR2_TEXTILE8**  lplpTextiles8;

    //
    // Read palettes data
    //
    if(m_TRType != TR1_FILE)
        {
        lp->Palette8    = *(TR2_PALETTE8*)      lpData;     lpData += sizeof(TR2_PALETTE8);
        lp->Palette16   = *(TR2_PALETTE16*)     lpData;     lpData += sizeof(TR2_PALETTE16);
        }

    DlgLoad.vSetLoadProgress(lpData - lpbData);

    //
    // Read Textiles data
    //
    lp->dwNumTextiles       = *(bitu32*)            lpData;     lpData += sizeof(bitu32);

    lplpTextiles8 = new TR2_TEXTILE8* [lp->dwNumTextiles];

    for(dwI = 0; dwI < lp->dwNumTextiles; dwI++)
        {
        lplpTextiles8 [dwI] = new TR2_TEXTILE8;
        *lplpTextiles8[dwI] = *(TR2_TEXTILE8*)  lpData;     lpData += sizeof(TR2_TEXTILE8);
        }

    if(m_TRType != TR1_FILE)
        {
        lp->lplpTextiles16  = new TR2_TEXTILE16* [lp->dwNumTextiles];
        for(dwI = 0; dwI < lp->dwNumTextiles; dwI++)
            {
            lp->lplpTextiles16 [dwI]    = new TR2_TEXTILE16;
            *lp->lplpTextiles16[dwI]    = *(TR2_TEXTILE16*) lpData; lpData += sizeof(TR2_TEXTILE16);
            }
        }

    lpData += sizeof(bitu32); // unused

    DlgLoad.vSetLoadProgress(lpData - lpbData);

    //
    // Read Rooms data
    //
    lp->wNumRooms           = *(bitu16*)            lpData;     lpData += sizeof(bitu16);
    lp->lpRooms = new TR123_ROOM [lp->wNumRooms];

    for(dwI = 0; dwI < lp->wNumRooms; dwI++)
        {
        TR123_ROOM*    lpRoom = &lp->lpRooms[dwI];

        lpRoom->RoomInfo = *(TR2_ROOM_INFO*)    lpData;         lpData += sizeof(TR2_ROOM_INFO);
        lpData += sizeof(bitu32); // NumData - unused

        TR2_ROOM_DATA* lpRoomData = (TR2_ROOM_DATA*) &lpRoom->RoomData;

        READ_ARRAYNUM_TR(bitu16,TR123_VERTEX_ROOM,lpRoomData->wNumVertices,lpRoomData->lpVertices,TR1_VERTEX_ROOM,TR2_VERTEX_ROOM)
        READ_ARRAYNUM(bitu16,TR2_FACE4,lpRoomData->wNumRectangles,lpRoomData->lpRectangles)
        READ_ARRAYNUM(bitu16,TR2_FACE3,lpRoomData->wNumTriangles,lpRoomData->lpTriangles)
        READ_ARRAYNUM(bitu16,TR2_ROOM_SPRITE,lpRoomData->wNumSprites,lpRoomData->lpSprites)
        READ_ARRAYNUM(bitu16,TR2_ROOM_PORTAL,lpRoom->wNumPortals,lpRoom->lpPortals)

        lpRoom->wNumZSectors = *(bitu16*) lpData;        lpData += sizeof(bitu16);
        lpRoom->wNumXSectors = *(bitu16*) lpData;        lpData += sizeof(bitu16);
        READ_ARRAY(lpRoom->wNumZSectors * lpRoom->wNumXSectors,TR2_ROOM_SECTOR,lpRoom->lpSectorList)

        lpRoom->sAmbientIntensity1 = *(bit16*) lpData;        lpData += sizeof(bit16);
        if(m_TRType != TR1_FILE)                                                            
            {
            lpRoom->sAmbientIntensity2 = *(bit16*) lpData;        lpData += sizeof(bit16);
            }
        if(m_TRType == TR2_FILE)                                                            
            {
            lpRoom->sLightMode = *(bit16*) lpData;        lpData += sizeof(bit16);
            }

        READ_ARRAYNUM_TR(bitu16,TR123_ROOM_LIGHT,lpRoom->wNumLights,lpRoom->lpLights,TR1_ROOM_LIGHT,TR2_ROOM_LIGHT)
        READ_ARRAYNUM_TR(bitu16,TR123_ROOM_STATICMESH,lpRoom->wNumStaticMeshes,lpRoom->lpStaticMeshes,TR1_ROOM_STATICMESH,TR2_ROOM_STATICMESH)

        lpRoom->sAlternateRoom = *(bit16*) lpData;        lpData += sizeof(bit16);
        lpRoom->sFlags = *(bit16*) lpData;        lpData += sizeof(bit16);
        if(m_TRType == TR3_FILE)
            {
            lpRoom->RoomLightColor        = *(TR2_COLOR*) lpData;  lpData += sizeof(TR2_COLOR);
            }
        }

    DlgLoad.vSetLoadProgress(lpData - lpbData);

    //
    // Read behaviours data
    //

    READ_ARRAYNUM(bitu32,bitu16,lp->dwNumFloorData,lp->lpFloorData)

    //
    // Read Meshes geometry data
    //
    lp->dwNumMeshes = 0;

    bitu32 dwNumMeshData      = *(bitu32*)             lpData;  lpData += sizeof(bitu32);
    BYTE* lpbAddrStartMesh = lpData;
    BYTE* lpbMeshEndAddr = lpData + dwNumMeshData * 2;

    TR123_MESH   Mesh;

    while(lpData != lpbMeshEndAddr)
        {
        lpData = lpbReadMesh(&Mesh,lpData,lpbAddrStartMesh);
        vReleaseMeshMemory(&Mesh);
        lp->dwNumMeshes++;
        }

    lp->lplpMeshes = new TR123_MESH* [lp->dwNumMeshes];

    dwI = 0;
    lpData = lpbAddrStartMesh;

    DWORD*  lpLstAddr = new DWORD [lp->dwNumMeshes];

    while(lpData != lpbMeshEndAddr)
        {
        lpLstAddr[dwI] = lpData - lpbAddrStartMesh;
        lp->lplpMeshes[dwI] = new TR123_MESH;
        lpData = lpbReadMesh(lp->lplpMeshes[dwI],lpData,lpbAddrStartMesh);
        dwI++;
        }

    READ_ARRAYNUM(bitu32,bitu32,lp->dwNumMeshPointers,lp->lpMeshPointers)

    for(dwI = 0; dwI < lp->dwNumMeshPointers; dwI++)
        {
        bitu32  dwAddr = lp->lpMeshPointers[dwI];
        bool    boOk = false;
        for(DWORD dwJ = 0; dwJ < lp->dwNumMeshes; dwJ++)
            {
            if(dwAddr == lpLstAddr[dwJ])
                {
                lp->lpMeshPointers[dwI] = dwJ;
                boOk = true;
                break;
                }
            }
        if(!boOk)
            {
            CString s;
            s.Format("Can't find the mesh address of the MeshPointer num #%ld !!",dwI);
            AfxMessageBox(s,MB_OK | MB_ICONSTOP | MB_TOPMOST);
            }
        }

    delete [] lpLstAddr;

    DlgLoad.vSetLoadProgress(lpData - lpbData);

    //
    // Read meshes animations data
    //

    READ_ARRAYNUM(bitu32,TR2_ANIMATION,lp->dwNumAnimations,lp->lpAnimations)
    READ_ARRAYNUM(bitu32,TR2_STATE_CHANGE,lp->dwNumStateChanges,lp->lpStateChanges)
    READ_ARRAYNUM(bitu32,TR2_ANIM_DISPATCH,lp->dwNumAnimDispatches,lp->lpAnimDispatches)
    READ_ARRAYNUM(bitu32,TR2_ANIM_COMMAND,lp->dwNumAnimCommands,lp->lpAnimCommands)
    READ_ARRAYNUM(bitu32,TR2_MESHTREE,lp->dwNumMeshTrees,lp->lpMeshTrees)
    READ_ARRAYNUM(bitu32,bitu16,lp->dwNumFrames,lp->lpFrames)

    DlgLoad.vSetLoadProgress(lpData - lpbData);

    //
    // Read meshes lists data
    //
    READ_ARRAYNUM(bitu32,TR2_MOVEABLE,lp->dwNumMoveables,lp->lpMoveables)
    READ_ARRAYNUM(bitu32,TR2_STATICMESH,lp->dwNumStaticMeshes,lp->lpStaticMeshes)

    if(m_TRType != TR3_FILE)
        {
        READ_ARRAYNUM(bitu32,TR2_OBJECT_TEXTURE,lp->dwNumObjectTextures,lp->lpObjectTextures)
        }

    READ_ARRAYNUM(bitu32,TR2_SPRITE_TEXTURE,lp->dwNumSpriteTextures,lp->lpSpriteTextures)
    READ_ARRAYNUM(bitu32,TR2_SPRITE_SEQUENCE,lp->dwNumSpriteSequences,lp->lpSpriteSequences)

    DlgLoad.vSetLoadProgress(lpData - lpbData);

    //
    // Read other & sound data
    //
    if(m_TRType == TR1_FILE  &&  m_boTUBFile)
        {
        lp->Palette8        = *(TR2_PALETTE8*)      lpData;     lpData += sizeof(TR2_PALETTE8);
        }

    READ_ARRAYNUM(bitu32,TR2_CAMERA,lp->dwNumCameras,lp->lpCameras)
    READ_ARRAYNUM(bitu32,TR2_SOUND_SOURCE,lp->dwNumSoundSources,lp->lpSoundSources)

    DlgLoad.vSetLoadProgress(lpData - lpbData);

    //
    // Read NPCs data
    //
    READ_ARRAYNUM_TR(bitu32,TR123_BOX,lp->dwNumBoxes,lp->lpBoxes,TR1_BOX,TR2_BOX)
    READ_ARRAYNUM(bitu32,bitu16,lp->dwNumOverlaps,lp->lpOverlaps)
    READ_ARRAY_TR(lp->dwNumBoxes,TR123_ZONE,lp->lpZones,TR1_ZONE,TR2_ZONE)

    //
    // Read textures data
    //
    READ_ARRAYNUM(bitu32,bitu16,lp->dwNumAnimatedTextures,lp->lpAnimatedTextures)

    if(m_TRType == TR3_FILE)
        {
        READ_ARRAYNUM(bitu32,TR2_OBJECT_TEXTURE,lp->dwNumObjectTextures,lp->lpObjectTextures)
        }

    DlgLoad.vSetLoadProgress(lpData - lpbData);

    //
    // Read other data
    //
    READ_ARRAYNUM_TR(bitu32,TR123_ITEM,lp->dwNumItems,lp->lpItems,TR1_ITEM,TR2_ITEM)

    lp->LightMap                   = *(TR2_LIGHTMAP*)       lpData;  lpData += sizeof(TR2_LIGHTMAP);

    if(m_TRType == TR1_FILE  &&  !m_boTUBFile)
        {
        lp->Palette8    = *(TR2_PALETTE8*)      lpData;     lpData += sizeof(TR2_PALETTE8);
        }

    READ_ARRAYNUM(bitu16,TR2_CINEMATIC_FRAME,lp->wNumCinematicFrames,lp->lpCinematicFrames)
    READ_ARRAYNUM(bitu16,bitu8,lp->wNumDemoData,lp->lpDemoData)

    DlgLoad.vSetLoadProgress(lpData - lpbData);

    //
    // Read sound data
    //
    if(m_TRType != TR1_FILE)
        {
        lp->SoundMap.tr2 = *(TR2_SOUNDMAP*) lpData;
        lpData += sizeof(TR2_SOUNDMAP);
        }
    else
        {
        lp->SoundMap.tr1 = *(TR1_SOUNDMAP*) lpData;
        lpData += sizeof(TR1_SOUNDMAP);
        }

    READ_ARRAYNUM(bitu32,TR2_SOUND_DETAILS,lp->dwNumSoundDetails,lp->lpSoundDetails)

    if(m_TRType == TR1_FILE)
        {
        READ_ARRAYNUM(bitu32,bitu8,lp->dwNumSamples,lp->lpSamples)
        }

    READ_ARRAYNUM(bitu32,bitu32,lp->dwNumSampleIndices,lp->lpSampleIndices)

    DlgLoad.vSetLoadProgress(lpData - lpbData);

    //
    // Build the 16 bits textiles if TR1
    //
    if(m_TRType == TR1_FILE)
        {
        lp->lplpTextiles16  = new TR2_TEXTILE16* [lp->dwNumTextiles];

        for(DWORD dwTile = 0; dwTile < lp->dwNumTextiles; dwTile++)
            {
            lp->lplpTextiles16 [dwTile]    = new TR2_TEXTILE16;

            for(DWORD dwJ = 0; dwJ < 256; dwJ++)
                {
                for(DWORD dwI = 0; dwI < 256; dwI++)
                    {
                    BYTE bIndex = *((BYTE*)(lplpTextiles8[dwTile]->abTile + dwI + dwJ * 256));

                    BYTE bR = lp->Palette8.aRGB[bIndex].bRed    >> 1;
                    BYTE bG = lp->Palette8.aRGB[bIndex].bGreen  >> 1;
                    BYTE bB = lp->Palette8.aRGB[bIndex].bBlue   >> 1;

                    WORD wColor = (bR << 10) + (bG << 5) + bB;

                    if(bIndex != 0) wColor |= 0x8000;

                    lp->lplpTextiles16[dwTile]->awTile[dwI + dwJ * 256] = wColor;
                    }
                }
            }
        }

    //
    // Delete the 8 bits textiles
    //
    for(DWORD dwTile = 0; dwTile < lp->dwNumTextiles; dwTile++)
        {
        delete lplpTextiles8[dwTile];
        }

    delete [] lplpTextiles8;

    //
    // Build a list of unused object textures
    //
    BYTE*   lpbObjText = new BYTE [ lp->dwNumObjectTextures ];

    ZeroMemory(lpbObjText, lp->dwNumObjectTextures * sizeof(BYTE) );

    // Flag object textures used by meshes
    for(DWORD dwMesh = 0; dwMesh < lp->dwNumMeshes; dwMesh++)
        {
        TR123_MESH* lpMesh = lp->lplpMeshes[ dwMesh ];

        for(DWORD dwPoly = 0; dwPoly < lpMesh->sNumTexturedRectangles; dwPoly++)
            {
            lpbObjText[ lpMesh->lpTexturedRectangles[ dwPoly ].wTextureIndex ]++;
            }

        for(dwPoly = 0; dwPoly < lpMesh->sNumTexturedTriangles; dwPoly++)
            {
            lpbObjText[ lpMesh->lpTexturedTriangles[ dwPoly ].wTextureIndex ]++;
            }
        }

    // Flag object textures used by rooms
    for(DWORD dwNumRoom = 0; dwNumRoom < lp->wNumRooms; dwNumRoom++)
        {
        TR123_ROOM*     lpRoom = &lp->lpRooms[dwNumRoom];
        TR2_ROOM_DATA*  lpRoomData = &lpRoom->RoomData;

        for(DWORD dwPoly = 0; dwPoly < lpRoomData->wNumRectangles; dwPoly++)
            {
            lpbObjText[ lpRoomData->lpRectangles[ dwPoly ].wTextureIndex & 0x7FFF ]++;
            }

        for(dwPoly = 0; dwPoly < lpRoomData->wNumTriangles; dwPoly++)
            {
            lpbObjText[ lpRoomData->lpTriangles[ dwPoly ].wTextureIndex & 0x7FFF ]++;
            }
        }

    // Flag object textures used by animated textures
    bitu16* lpsAnimatedTextures     = lp->lpAnimatedTextures;
    WORD    wNumAnimatedTextures    = *lpsAnimatedTextures++;

    while(wNumAnimatedTextures--)
        {
        bitu16   wNum = *lpsAnimatedTextures++;  wNum++;
        while(wNum--)
            {
            bitu16  wNumObjText = *lpsAnimatedTextures;

            lpbObjText[ wNumObjText ]++;

            lpsAnimatedTextures++;
            }
        }

    // Build the list of non-flagged object textures
    lp->lpListUnusedObjText = new TRListObjText;

    for(DWORD dwObjText = 0; dwObjText < lp->dwNumObjectTextures; dwObjText++)
        {
        if( lpbObjText[ dwObjText ] ) continue;

        lp->lpListUnusedObjText->push_back( lp->lpObjectTextures[ dwObjText ] );
        }

    delete [] lpbObjText;

#ifndef __NO_CONVERSION__
    dwReplaceColoredPolys(lp->Palette8, lp->Palette16);

    //
    // Convert to internal representation of the level
    //
    vConvertFromTRLevel(DlgLoad);
#endif

    DlgLoad.DestroyWindow();
}


//----------------------------------------------------------------------------
//
//  Function:   vParseTR4File
//
//  Synopsis:   Read a TR4 level and build internal structures accordingly
//
//  Arguments:  [lpData] -- pointer on the level data
//              [dwDataSize] -- data size of the level
//
//  Return:     -
//
//----------------------------------------------------------------------------
void    CTRFile::vParseTR4File(BYTE* lpData,DWORD dwDataSize)
{
    DWORD   dwSize;
    BYTE*   lpDataSave = lpData;
    TR4_FILE_STRUCT* lp = (TR4_FILE_STRUCT*) &m_trFile.tr4File;

    lp->wUnknown1 = * (bitu16*) lpData;     lpData += sizeof(bitu16);
    lp->wUnknown2 = * (bitu16*) lpData;     lpData += sizeof(bitu16);
    lp->wUnknown3 = * (bitu16*) lpData;     lpData += sizeof(bitu16);

    lp->Texture1.dwTextUncompSize = * (bitu32*) lpData;  lpData += sizeof(bitu32);
    lp->Texture1.dwTextCompSize = * (bitu32*) lpData;  lpData += sizeof(bitu32);
    READ_ARRAY(lp->Texture1.dwTextCompSize,bitu8,lp->Texture1.lpbTextData)

    lp->Texture2.dwTextUncompSize = * (bitu32*) lpData;  lpData += sizeof(bitu32);
    lp->Texture2.dwTextCompSize = * (bitu32*) lpData;  lpData += sizeof(bitu32);
    READ_ARRAY(lp->Texture2.dwTextCompSize,bitu8,lp->Texture2.lpbTextData)

    lp->Texture3.dwTextUncompSize = * (bitu32*) lpData;  lpData += sizeof(bitu32);
    lp->Texture3.dwTextCompSize = * (bitu32*) lpData;  lpData += sizeof(bitu32);
    READ_ARRAY(lp->Texture3.dwTextCompSize,bitu8,lp->Texture3.lpbTextData)

    lp->dwStuffSize = dwDataSize - (lpData - lpDataSave);    
    READ_ARRAY(lp->dwStuffSize,bitu8,lp->lpStuff)
}


//----------------------------------------------------------------------------
//
//  Function:   boParseTRLevel
//
//  Synopsis:   Read a TR level and build internal structures accordingly
//
//  Arguments:  [lpData] -- pointer on the level data
//              [dwDataSize] -- data size of the level
//              [lpszFileName] -- name of the file
//
//  Return:     true if the level could be parsed, else false
//
//----------------------------------------------------------------------------
bool    CTRFile::boParseTRLevel(BYTE* lpData,DWORD dwDataSize,TCHAR* lpszFileName)
{
    _tcsupr(lpszFileName);

    DWORD dwVersion = *(bitu32*) lpData; lpData += sizeof(bitu32);

    switch(dwVersion)
        {
        case 0x00000020:    m_TRType        = TR1_FILE; break;
        case 0x0000002d:    m_TRType        = TR2_FILE; break;
        case 0xFF080038:
        case 0xFF180038:    m_TRType        = TR3_FILE; break;
        case 0x00345254:    m_TRType        = TR4_FILE; break;
        default:            m_TRType        = TR_UNKNOWN;
        }

    if(m_TRType == TR_UNKNOWN) return false;

    if(m_TRType == TR1_FILE  &&  (_tcsstr(lpszFileName,_T(".tub")) != NULL))
        {
        m_boTUBFile = true;
        }

    if(m_TRType == TR4_FILE)
        {
        vParseTR4File(lpData,dwDataSize - sizeof(bitu32));
        return true;
        }

    vParseTR123File(lpData,dwDataSize);

    return true;
}
