#include "stdafx.h"

#include "Resource.h"
#include "Fragment.h"
#include "DlgLevelLoad.h"
#include "DlgLevelSave.h"
#include "AnimData.h"
#include "AnimKey.h"
#include "AnimTrack.h"
#include "AnimMgr.h"
#include "TRAnimTrack.h"


#ifdef __EXACT_TRANGLE__
DWORD dwGetFrameSize(CTRAnimTrack* lpTrack)
{
    DWORD   dwFrameSize = 0;

    for(DWORD dwKey = 0; dwKey < lpTrack->dwGetNumKeys(); dwKey++)
        {
        CAnimKey* lpKey = lpTrack->lpGetKey(dwKey);

        DWORD   dwCurFrameSize = 0;

        for(DWORD dwData = 0; dwData < lpKey->dwGetNumData(); dwData++)
            {
            const CAnimData& Data = lpKey->GetData(dwData);

            WORD    wTRAngle1, wTRAngle2;

            Data.vGetTRAngle(wTRAngle1, wTRAngle2);

            dwCurFrameSize++;

            if(wTRAngle2 != 0x1234) dwCurFrameSize++;
            }

        if(dwFrameSize < dwCurFrameSize) dwFrameSize = dwCurFrameSize;
        }

    return dwFrameSize + 3 * 3;
}
#endif


//----------------------------------------------------------------------------
//
//  Function:   vConvertToTRLevel
//
//  Synopsis:   Convert internal representation of the level to TR's one
//
//  Arguments:  [DlgSave] -- save dialog (to update progress bars)
//              [boKeepUnusedTextures] -- if true, save the level with the
//                  unused textures
//
//  Return:     -
//
//----------------------------------------------------------------------------
void    CTRFile::vConvertToTRLevel(CDlgLevelSave& DlgSave, bool boKeepUnusedTextures)
{
    TR123_FILE_STRUCT& trFile = m_trFile.tr123File;

    bool    boTR1 = (m_TRType == TR1_FILE);

    DWORD   dwMaxRange = 5 + trFile.lpListSpriteSeq->size() * 2 + trFile.lpListMoveable->size() * 3 + trFile.lpListStaticMesh->size() * 2 + 1;

    DlgSave.vSetConvertRange(dwMaxRange);

    DWORD dwProgress = 0;

    //
    // Recalculate OBJECT_TEXTURE structures
    //
    DWORD       dwNumFragments = 0;
    CFragment*  lpLstFragments;

    // Collect fragments for the whole level and sort them
    CFragment::vCollectFragments(this,&lpLstFragments,&dwNumFragments);

    // Add fragments for the unused object textures
    if(boKeepUnusedTextures)
        {
        for(TRListObjText::iterator theIter = trFile.lpListUnusedObjText->begin(); theIter != trFile.lpListUnusedObjText->end(); theIter++)
            {
            TR2_OBJECT_TEXTURE* lpObjText = &(*theIter);
            bool    boTri = (lpObjText->Vertices[ 3 ].bXPixel == 0) &&
                            (lpObjText->Vertices[ 3 ].bYPixel == 0) &&
                            (lpObjText->Vertices[ 3 ].cXCoordinate == 0) &&
                            (lpObjText->Vertices[ 3 ].cYCoordinate == 0);
            BYTE    bNumVert = boTri ? 3 : 4;
            BYTE*   lpbTextureData = (BYTE*) trFile.lplpTextiles16[ lpObjText->wTile ]->awTile;

            CFragment::vAddFragment(lpLstFragments, dwNumFragments, lpObjText, bNumVert, OBJTEXT_TYPE_UNUSED, NULL, true, lpbTextureData);
            }
        }

    CFragment::vSortFragments(lpLstFragments,0,dwNumFragments-1);
    CFragment::dwRemoveDuplicated(dwNumFragments,lpLstFragments);

    // Build the new texture
    BYTE*   lpbArea         = NULL;
    DWORD   dwAreaWidth     = 0;
    DWORD   dwAreaHeight    = 0;
    DWORD   dwTextureDepth  = 16;

    DWORD   dwNumObjText = CFragment::dwBuildTextureFromFragments(lpLstFragments,dwNumFragments,&lpbArea,&dwAreaWidth,&dwAreaHeight,true);

    // Create the new TR2_OBJECT_TEXTURE structures
    delete [] trFile.lpObjectTextures;

    trFile.dwNumObjectTextures = dwNumObjText;
    trFile.lpObjectTextures = new TR2_OBJECT_TEXTURE [ dwNumObjText ];

    DWORD dwIndex = 0;

    for(DWORD dwN = 0; dwN < dwNumFragments; dwN++)
        {
        for(LISTOBJTEXT::iterator theIter = lpLstFragments[dwN].m_LstTextCoord.begin(); theIter != lpLstFragments[dwN].m_LstTextCoord.end(); theIter++)
            {
            FRAGMENT_TEXTURE* lpFragText    = &(*theIter);
            TR2_OBJECT_TEXTURE*  lpObjText  = &lpFragText->ObjText;

            if(!lpFragText->boAlreadyDone) trFile.lpObjectTextures[dwIndex++] = *lpObjText;
            }
        }

    trFile.dwNumObjectTextures = dwIndex;

    // Delete the old tiles
    DWORD dwOldNumTiles = trFile.dwNumTextiles;

    for(dwN = 0; dwN < trFile.dwNumTextiles; dwN++)
        {
        delete trFile.lplpTextiles16[dwN];
        }

    delete [] trFile.lplpTextiles16;

    // Create the new tiles
    trFile.dwNumTextiles = (dwAreaHeight - 1) / 256 + 1;;
    trFile.lplpTextiles16 = new TR2_TEXTILE16* [ trFile.dwNumTextiles ];

    for(dwN = 0; dwN < trFile.dwNumTextiles; dwN++)
        {
        TR2_TEXTILE16* lp16 = new TR2_TEXTILE16;
        memset(lp16,0,sizeof(TR2_TEXTILE16));
        DWORD dwHeight = 256;
        if((dwN + 1) * 256 > dwAreaHeight) dwHeight = dwAreaHeight & 0xFF;
        memcpy(lp16,lpbArea + dwN * sizeof(TR2_TEXTILE16), dwHeight * 256 * 2);
        trFile.lplpTextiles16[dwN] = lp16;
        }

    delete [] lpLstFragments;
    delete [] lpbArea;

    DlgSave.vSetNumTiles(dwOldNumTiles, trFile.dwNumTextiles);

    dwProgress = 5;

    DlgSave.vSetConvertProgress(dwProgress);

    //
    // Process sprites
    //
    TRListSpriteSeq*    lpListSpriteSeq     = trFile.lpListSpriteSeq;
    DWORD               dwNumSpriteSeq      = lpListSpriteSeq->size();
    DWORD               dwNumSpriteTexture  = 0;

    for(TRListSpriteSeq::iterator theIterator3 = lpListSpriteSeq->begin(); theIterator3 != lpListSpriteSeq->end(); theIterator3++)
        {
        CTrLstSpr* lpLstSpr = *theIterator3;

        dwNumSpriteTexture += lpLstSpr->dwGetNumSprites();
        }

    TR2_SPRITE_SEQUENCE*    lpSpriteSequence    = new TR2_SPRITE_SEQUENCE [ dwNumSpriteSeq ];
    TR2_SPRITE_TEXTURE*     lpSpriteTexture     = new TR2_SPRITE_TEXTURE [ dwNumSpriteTexture ];

    DWORD   dwSpriteSeq             = 0;
    DWORD   dwSpriteTextureOffset   = 0;

    for(theIterator3 = lpListSpriteSeq->begin(); theIterator3 != lpListSpriteSeq->end(); theIterator3++, dwSpriteSeq++)
        {
        CTrLstSpr* lpLstSpr = *theIterator3;

        DWORD   dwNumSprite = lpLstSpr->dwGetNumSprites();

        const TR2_OBJECT_TEXTURE* lpOText = lpLstSpr->lpGetFragText();
        const WORLD_POSITIONNING* lpWorld = lpLstSpr->lpGetWorldPositionning();

        lpSpriteSequence[dwSpriteSeq].lObjectId         = lpLstSpr->dwGetObjectId();
        lpSpriteSequence[dwSpriteSeq].sNegativeLength   = -dwNumSprite;
        lpSpriteSequence[dwSpriteSeq].sOffset           = dwSpriteTextureOffset;

        for(DWORD dwSprite = 0; dwSprite < dwNumSprite; dwSprite++, dwSpriteTextureOffset++)
            {
            const TR2_OBJECT_TEXTURE* lpObjText = &lpOText[lpWorld[dwSprite].wFragIndex];

            DWORD   dwWidth     = lpObjText->Vertices[1].bXPixel - lpObjText->Vertices[0].bXPixel;
            DWORD   dwHeight    = lpObjText->Vertices[2].bYPixel - lpObjText->Vertices[0].bYPixel;

            lpSpriteTexture[dwSpriteTextureOffset].bX           = lpObjText->Vertices[0].bXPixel;
            lpSpriteTexture[dwSpriteTextureOffset].bY           = lpObjText->Vertices[0].bYPixel;
            lpSpriteTexture[dwSpriteTextureOffset].sBottomSide  = lpWorld[dwSprite].sBottomSide;
            lpSpriteTexture[dwSpriteTextureOffset].sLeftSide    = lpWorld[dwSprite].sLeftSide;
            lpSpriteTexture[dwSpriteTextureOffset].sRightSide   = lpWorld[dwSprite].sRightSide;
            lpSpriteTexture[dwSpriteTextureOffset].sTopSide     = lpWorld[dwSprite].sTopSide;
            lpSpriteTexture[dwSpriteTextureOffset].wHeight      = dwHeight * 256 + 255;
            lpSpriteTexture[dwSpriteTextureOffset].wWidth       = dwWidth * 256 + 255;
            lpSpriteTexture[dwSpriteTextureOffset].wTile        = lpObjText->wTile;
            }

        delete lpLstSpr;

        DlgSave.vSetConvertProgress(++dwProgress);
        }

    delete [] trFile.lpSpriteTextures;
    delete [] trFile.lpSpriteSequences;

    trFile.dwNumSpriteSequences = dwNumSpriteSeq;
    trFile.lpSpriteSequences    = lpSpriteSequence;
    trFile.dwNumSpriteTextures  = dwNumSpriteTexture;
    trFile.lpSpriteTextures     = lpSpriteTexture;

    //
    // Calculate the new number of meshes / mesh pointers and reallocate the arrays
    //
    TRListMoveable*     lpListMoveable                  = trFile.lpListMoveable;
    DWORD               dwNumMeshMoveablesWithoutDummy  = 0;
    DWORD               dwNumMeshMoveables              = 0;
    DWORD               dwNumMoveables                  = lpListMoveable->size();

    for(TRListMoveable::iterator theIterator = lpListMoveable->begin(); theIterator != lpListMoveable->end(); theIterator++)
        {
        CTrMvb* lpMvb = *theIterator;

        for(DWORD dwMesh = 0; dwMesh < lpMvb->dwGetNumMeshes(); dwMesh++)
            {
            TR123_MESH* lpMesh  = const_cast<TR123_MESH*>(lpMvb->lpGetMesh(dwMesh));

            if(!lpMesh->boIsDummy) dwNumMeshMoveablesWithoutDummy++;
            }

        dwNumMeshMoveables += lpMvb->dwGetNumMeshes();
        }

    TRListStaticMesh*   lpListStatMsh                   = trFile.lpListStaticMesh;
    DWORD               dwNumStaticMeshesWithoutDummy   = 0;
    DWORD               dwNumStaticMeshes               = lpListStatMsh->size();
    
    for(TRListStaticMesh::iterator theIterator2 = lpListStatMsh->begin(); theIterator2 != lpListStatMsh->end(); theIterator2++)
        {
        CTrStatMsh* lpStatMsh = *theIterator2;

        TR123_MESH* lpMesh      = const_cast<TR123_MESH*>(lpStatMsh->lpGetMesh(0));

        if(!lpMesh->boIsDummy) dwNumStaticMeshesWithoutDummy++;
        }

    DWORD   dwNumMeshes             = 0;
    DWORD   dwNewNumMeshes          = dwNumMeshes + dwNumMeshMoveablesWithoutDummy + dwNumStaticMeshesWithoutDummy;
    DWORD   dwNumMeshPointers       = 0;
    DWORD   dwNewNumMeshPointers    = dwNumMeshPointers + dwNumMeshMoveables + dwNumStaticMeshes;

    TR123_MESH**    lplpMeshes      = new TR123_MESH* [ dwNewNumMeshes ];
    bitu32*         lpMeshPointers  = new bitu32 [ dwNewNumMeshPointers ];

    memcpy(lplpMeshes,      trFile.lplpMeshes,      dwNumMeshes * sizeof(TR123_MESH*));
    memcpy(lpMeshPointers,  trFile.lpMeshPointers,  dwNumMeshPointers * sizeof(bitu32));

    delete [] trFile.lplpMeshes;
    delete [] trFile.lpMeshPointers;
    
    trFile.dwNumMeshes          = dwNewNumMeshes;
    trFile.lplpMeshes           = lplpMeshes;
    trFile.dwNumMeshPointers    = dwNewNumMeshPointers;
    trFile.lpMeshPointers       = lpMeshPointers;

    //
    // Process moveables
    //
    DWORD   dwNewNumAnimations      = 0;
    DWORD   dwNewNumMeshTrees       = 0;
    DWORD   dwNewNumStateChanges    = 0;
    DWORD   dwNewNumAnimDispatches  = 0;
    DWORD   dwNewNumAnimCommands    = 0;
    DWORD   dwSizeOfKeys            = 0;

    // Calculate the new sizes of the different arrays related to moveables
    for(theIterator = lpListMoveable->begin(); theIterator != lpListMoveable->end(); theIterator++)
        {
        CTrMvb*     lpMvb = *theIterator;
        CAnimMgr*   lpAnimMgr = lpMvb->lpGetAnimMgr();

        dwNewNumAnimations += lpAnimMgr->dwGetNumTracks();
        dwNewNumMeshTrees += 4 * (lpMvb->dwGetNumMeshes() - 1);

        for(DWORD dwTrack = 0; dwTrack < lpAnimMgr->dwGetNumTracks(); dwTrack++)
            {
            CTRAnimTrack* lpTrack = static_cast<CTRAnimTrack*>(lpAnimMgr->lpGetTrack(dwTrack));

            dwNewNumStateChanges += lpTrack->dwGetNumStateChanges();
            dwNewNumAnimDispatches += lpTrack->dwGetNumAnimDispatches();
            dwNewNumAnimCommands += lpTrack->dwGetNumParamsAnimCommands();
#ifndef __EXACT_TRANGLE__
            dwSizeOfKeys += lpTrack->dwGetNumKeys() * (3 * 3 + lpMvb->dwGetNumMeshes() * 2);
#else
            dwSizeOfKeys += lpTrack->dwGetNumKeys() * dwGetFrameSize(lpTrack);
#endif
            if(boTR1) dwSizeOfKeys += lpTrack->dwGetNumKeys();
            }

        DlgSave.vSetConvertProgress(++dwProgress);
        }

    // Allocate the new arrays
    TR2_MOVEABLE*       lpMoveables         = new TR2_MOVEABLE [ dwNumMoveables ];
    TR2_ANIMATION*      lpAnimations        = new TR2_ANIMATION [ dwNewNumAnimations ];
    TR2_MESHTREE*       lpMeshTrees         = new TR2_MESHTREE [ dwNewNumMeshTrees ];
    TR2_STATE_CHANGE*   lpStateChanges      = new TR2_STATE_CHANGE [ dwNewNumStateChanges ];
    TR2_ANIM_DISPATCH*  lpAnimDispatches    = new TR2_ANIM_DISPATCH [ dwNewNumAnimDispatches ];
    TR2_ANIM_COMMAND*   lpAnimCommands      = new TR2_ANIM_COMMAND [ dwNewNumAnimCommands ];
    bitu16*             lpKeys              = new bitu16 [ dwSizeOfKeys ];

    ZeroMemory(lpKeys, dwSizeOfKeys * sizeof(bitu16));

    delete [] trFile.lpMoveables;
    delete [] trFile.lpAnimations;
    delete [] trFile.lpMeshTrees;
    delete [] trFile.lpStateChanges;
    delete [] trFile.lpAnimDispatches;
    delete [] trFile.lpAnimCommands;
    delete [] trFile.lpFrames;

    trFile.dwNumMoveables       = dwNumMoveables;
    trFile.lpMoveables          = lpMoveables;
    trFile.dwNumMeshTrees       = dwNewNumMeshTrees;
    trFile.lpMeshTrees          = lpMeshTrees;
    trFile.dwNumAnimations      = dwNewNumAnimations;
    trFile.lpAnimations         = lpAnimations;
    trFile.dwNumStateChanges    = dwNewNumStateChanges;
    trFile.lpStateChanges       = lpStateChanges;
    trFile.dwNumAnimDispatches  = dwNewNumAnimDispatches;
    trFile.lpAnimDispatches     = lpAnimDispatches;
    trFile.dwNumAnimCommands    = dwNewNumAnimCommands;
    trFile.lpAnimCommands       = lpAnimCommands;
    trFile.dwNumFrames          = dwSizeOfKeys;
    trFile.lpFrames             = lpKeys;

    DWORD   dwMoveable      = 0;
    DWORD   dwMeshTree      = 0;
    DWORD   dwAnimation     = 0;
    DWORD   dwStateChange   = 0;
    DWORD   dwAnimDispatch  = 0;
    DWORD   dwAnimCommand   = 0;
    DWORD   dwKey           = 0;
    DWORD   dwFrame         = 0;

    // Build the new arrays
    for(theIterator = lpListMoveable->begin(); theIterator != lpListMoveable->end(); theIterator++)
        {
        CTrMvb*     lpMvb = *theIterator;
        CAnimMgr*   lpAnimMgr = lpMvb->lpGetAnimMgr();

        // Build the TR2_MOVEABLE structure
        lpMoveables[dwMoveable].dwFrameOffset   = dwKey * 2;
        lpMoveables[dwMoveable].dwMeshTree      = dwMeshTree;
        lpMoveables[dwMoveable].dwObjectId      = lpMvb->dwGetObjectId();
        lpMoveables[dwMoveable].wAnimation      = lpMvb->boIsAnimatedByEngine() ? 0xFFFF : dwAnimation;
        lpMoveables[dwMoveable].wNumMeshes      = lpMvb->dwGetNumMeshes();
        lpMoveables[dwMoveable].wStartingMesh   = dwNumMeshPointers;

        DWORD   dwMoveableStartFrame = dwFrame;

        dwMoveable++;

        // Build the TR2_MESHTREE structure
        memcpy(&lpMeshTrees[dwMeshTree], lpMvb->lpGetMeshTree(), 4 * sizeof(TR2_MESHTREE) * (lpMvb->dwGetNumMeshes() - 1));

        dwMeshTree += 4 * (lpMvb->dwGetNumMeshes() - 1);

        // Add the meshes to the mesh list
        for(DWORD dwMesh = 0; dwMesh < lpMvb->dwGetNumMeshes(); dwMesh++)
            {
            TR123_MESH* lpMesh      = const_cast<TR123_MESH*>(lpMvb->lpGetMesh(dwMesh));
            TR123_MESH* lpMeshDst   = new TR123_MESH;

            vCopyMesh(lpMesh, lpMeshDst);

            if(m_TRType != TR3_FILE) CTRFile::vMeshReverseLighting(lpMeshDst);

            if(lpMeshDst->boIsDummy)
                {
                lpMeshPointers[ dwNumMeshPointers ] = 0;
                }
            else
                {
                lplpMeshes[ dwNumMeshes++ ]         = lpMeshDst;
                lpMeshPointers[ dwNumMeshPointers ] = dwNumMeshes - 1;
                }

            dwNumMeshPointers++;
            }

        DWORD   dwStartAnimation = dwAnimation;

        // Loop through the tracks for this moveable
        for(DWORD dwTrack = 0; dwTrack < lpAnimMgr->dwGetNumTracks(); dwTrack++)
            {
            CTRAnimTrack* lpTrack = static_cast<CTRAnimTrack*>(lpAnimMgr->lpGetTrack(dwTrack));

            short   sTRUnknown;
            long    lTRSpeed;
            float   fTRAcceleration;
            DWORD   dwNextTrackIndex, dwNextTrackFrame;

            lpTrack->vGetPlayParams(sTRUnknown, lTRSpeed, fTRAcceleration);
            lpAnimMgr->vGetNextAnimInfos(dwNextTrackIndex, dwNextTrackFrame, dwTrack);

            short   sAccelHi = 0;
            WORD    wAccelLo = 0;

            if(fTRAcceleration >= 0.0f)
                {
                sAccelHi = (short) fTRAcceleration;
                wAccelLo = (WORD) ((fTRAcceleration - sAccelHi) * 65536.0f);
                }
            else
                {
                float fAccel = floor(fTRAcceleration);

                sAccelHi = (short) fAccel;
                wAccelLo = (WORD) ((fTRAcceleration - fAccel) * 65536.0f);
                }

            // Build the TR_ANIMATION structure
            lpAnimations[dwAnimation].bFrameRate            = lpTrack->dwGetFrameRate();
#ifndef __EXACT_TRANGLE__
            lpAnimations[dwAnimation].bFrameSize            = (3 * 3 + lpMvb->dwGetNumMeshes() * 2);
#else
            lpAnimations[dwAnimation].bFrameSize            = dwGetFrameSize(lpTrack);
#endif
            if(boTR1) lpAnimations[dwAnimation].bFrameSize++;
            lpAnimations[dwAnimation].dwFrameOffset         = dwKey * 2;
            lpAnimations[dwAnimation].sAccelHi              = sAccelHi;
            lpAnimations[dwAnimation].sSpeed                = lTRSpeed;
            lpAnimations[dwAnimation].Unknown               = sTRUnknown;
            lpAnimations[dwAnimation].wAccelLo              = wAccelLo;
            lpAnimations[dwAnimation].wAnimCommand          = dwAnimCommand;
            lpAnimations[dwAnimation].wFrameEnd             = dwFrame + lpTrack->dwGetNumFrames() - 1 - dwMoveableStartFrame;
            lpAnimations[dwAnimation].wFrameStart           = dwFrame - dwMoveableStartFrame;
            lpAnimations[dwAnimation].wNextAnimation        = dwStartAnimation + dwNextTrackIndex;
            lpAnimations[dwAnimation].wNextFrame            = dwNextTrackFrame;
            lpAnimations[dwAnimation].wNumAnimCommands      = lpTrack->dwGetNumAnimCommands();
            lpAnimations[dwAnimation].wNumStateChanges      = lpTrack->dwGetNumStateChanges();
            lpAnimations[dwAnimation].wStateChangeOffset    = dwStateChange;
            lpAnimations[dwAnimation].wStateId              = lpTrack->dwGetStateId();

            dwAnimation++;

            // Build the TR_STATE_CHANGE structures
            for(DWORD dwNumStateChange = 0; dwNumStateChange < lpTrack->dwGetNumStateChanges(); dwNumStateChange++)
                {
                const CTRStateChange& StateChange = lpTrack->GetStateChange(dwNumStateChange);

                lpStateChanges[dwStateChange].wAnimDispatch = dwAnimDispatch;
                lpStateChanges[dwStateChange].wNumAnimDispatches = StateChange.dwGetNumAnimDispatch();
                lpStateChanges[dwStateChange].wStateId = StateChange.dwGetStateId();

                dwStateChange++;

                // Build the TR_ANIM_DISPATCH structures
                for(DWORD dwNumAnimDispatch = 0; dwNumAnimDispatch < StateChange.dwGetNumAnimDispatch(); dwNumAnimDispatch++)
                    {
                    DWORD   dwTag;
                    const   CTRAnimDispatch& AnimDispatch = StateChange.GetAnimDispatch(dwNumAnimDispatch, dwTag, false);

                    DWORD dwStartFrame, dwEndFrame;
                    DWORD dwNextAnimation, dwNextFrame;

                    AnimDispatch.vGetFrameRange(dwStartFrame, dwEndFrame);
                    AnimDispatch.vGetNextAnim(dwNextAnimation, dwNextFrame);

                    lpAnimDispatches[dwAnimDispatch].sHigh = dwFrame + dwEndFrame + 1 - dwMoveableStartFrame;
                    lpAnimDispatches[dwAnimDispatch].sLow = dwFrame + dwStartFrame - dwMoveableStartFrame;
                    lpAnimDispatches[dwAnimDispatch].sNextAnimation = dwStartAnimation + dwNextAnimation;
                    lpAnimDispatches[dwAnimDispatch].sNextFrame = dwNextFrame;

                    dwAnimDispatch++;
                    }
                }

            // Build the TR_ANIM_COMMAND structures
            for(DWORD dwNumAnimCommand = 0; dwNumAnimCommand < lpTrack->dwGetNumAnimCommands(); dwNumAnimCommand++)
                {
                const CTRAnimCommand& AnimCommand = lpTrack->GetAnimCommand(dwNumAnimCommand);

                lpAnimCommands[dwAnimCommand++] = AnimCommand.sGetCommand();

                bool boFirstParamIsFrame = CTRAnimCommand::boFirstParamIsFrame( lpAnimCommands[dwAnimCommand-1].sValue );

                for(BYTE bParam = 0; bParam < AnimCommand.wGetNumParams(); bParam++)
                    {
                    lpAnimCommands[dwAnimCommand++] = AnimCommand.sGetParam(bParam);
                    if(bParam == 0  &&  boFirstParamIsFrame)
                        {
                        lpAnimCommands[dwAnimCommand-1].sValue += dwFrame - dwMoveableStartFrame;
                        }
                    }
                }

            // Build the keys
            for(DWORD dwNumKey = 0; dwNumKey < lpTrack->dwGetNumKeys(); dwNumKey++)
                {
                CAnimKey*           lpKey = lpTrack->lpGetKey(dwNumKey);
                const CBoundingBox  BB = lpKey->GetBB();

                DWORD dwCurKey = dwKey;

                lpKeys[dwKey++] =   (bit16) BB.GetBBLo()[0];
                lpKeys[dwKey++] =   (bit16) BB.GetBBHi()[0];
                lpKeys[dwKey++] =  -(bit16) BB.GetBBHi()[1];
                lpKeys[dwKey++] =  -(bit16) BB.GetBBLo()[1];
                lpKeys[dwKey++] =  -(bit16) BB.GetBBHi()[2];
                lpKeys[dwKey++] =  -(bit16) BB.GetBBLo()[2];

                for(DWORD dwNumData = 0; dwNumData < lpKey->dwGetNumData(); dwNumData++)
                    {
                    const CAnimData& AnimData = lpKey->GetData(dwNumData);

                    if(dwNumData == 0)
                        {
                        const CVect3D&  vecTrans = AnimData.vecGetTrans();

                        lpKeys[dwKey++] =  (bit16) vecTrans[0];
                        lpKeys[dwKey++] = -(bit16) vecTrans[1];
                        lpKeys[dwKey++] = -(bit16) vecTrans[2];

                        if(boTR1) lpKeys[dwKey++] = lpKey->dwGetNumData();
                        }

                    WORD wAngle1, wAngle2;

                    AnimData.vGetTRAngle(wAngle1, wAngle2);

                    if(boTR1)
                        {
                        WORD wTemp = wAngle1;
                        wAngle1 = wAngle2;
                        wAngle2 = wTemp;
                        }

                    lpKeys[dwKey++] = wAngle1;

#ifndef __EXACT_TRANGLE__
                    lpKeys[dwKey++] = wAngle2;
#else
                    if(wAngle2 != 0x1234) lpKeys[dwKey++] = wAngle2;
#endif
                    }

                dwKey = dwCurKey + lpAnimations[dwAnimation - 1].bFrameSize;
                }

            if(boTR1) lpAnimations[dwAnimation - 1].bFrameSize = 0;

            dwFrame += lpTrack->dwGetNumFrames();
            }

        DlgSave.vSetConvertProgress(++dwProgress);
        }

    // Adjust the NextFrame field for each animation
    for(dwAnimation = 0; dwAnimation < dwNewNumAnimations; dwAnimation++)
        {
        TR2_ANIMATION*   lpAnimation = &lpAnimations[dwAnimation];

        lpAnimation->wNextFrame += lpAnimations[lpAnimation->wNextAnimation].wFrameStart;
        }

    // Adjust the NextFrame field for each anim_dispatch
    TR2_MOVEABLE*   lpMoveable = lpMoveables;

    for(theIterator = lpListMoveable->begin(); theIterator != lpListMoveable->end(); theIterator++, lpMoveable++)
        {
        CTrMvb*     lpMvb = *theIterator;
        CAnimMgr*   lpAnimMgr = lpMvb->lpGetAnimMgr();

        DWORD   dwStartAnimation = lpMoveable->wAnimation;

        for(DWORD dwAnimation = 0; dwAnimation < lpAnimMgr->dwGetNumTracks(); dwAnimation++)
            {
            TR2_ANIMATION* lpAnim = &lpAnimations[dwStartAnimation + dwAnimation];

            for(DWORD dwStateChange = 0; dwStateChange < lpAnim->wNumStateChanges; dwStateChange++)
                {
                TR2_STATE_CHANGE* lpStateChange = &lpStateChanges[lpAnim->wStateChangeOffset + dwStateChange];

                for(DWORD dwAnimDispatch = 0; dwAnimDispatch < lpStateChange->wNumAnimDispatches; dwAnimDispatch++)
                    {
                    TR2_ANIM_DISPATCH* lpAnimDispatch = &lpAnimDispatches[lpStateChange->wAnimDispatch + dwAnimDispatch];

                    if(lpAnimDispatch->sNextAnimation < dwNewNumAnimations)
                        lpAnimDispatch->sNextFrame += lpAnimations[lpAnimDispatch->sNextAnimation].wFrameStart;
                    else
                        lpAnimDispatch->sNextFrame += lpAnimations[lpAnimDispatch->sNextAnimation - dwStartAnimation].wFrameStart;
                    }
                }
            }

        delete lpMvb;
        }

    //
    // Process static meshes
    //
    TR2_STATICMESH* lpStaticMeshes  = new TR2_STATICMESH [ dwNumStaticMeshes ];

    DWORD   dwStaticMesh = 0;

    for(theIterator2 = lpListStatMsh->begin(); theIterator2 != lpListStatMsh->end(); theIterator2++)
        {
        CTrStatMsh* lpStatMsh = *theIterator2;

        TR123_MESH* lpMesh      = const_cast<TR123_MESH*>(lpStatMsh->lpGetMesh(0));
        TR123_MESH* lpMeshDst   = new TR123_MESH;

        vCopyMesh(lpMesh, lpMeshDst);

        if(m_TRType != TR3_FILE) CTRFile::vMeshReverseLighting(lpMeshDst);

        if(lpMeshDst->boIsDummy)
            {
            lpMeshPointers[ dwNumMeshPointers ] = 0;
            }
        else
            {
            lplpMeshes[ dwNumMeshes++ ]         = lpMeshDst;
            lpMeshPointers[ dwNumMeshPointers ] = dwNumMeshes - 1;
            }

        CBoundingBox BBVis = lpStatMsh->GetBBVisibility();
        CBoundingBox BBCol = lpStatMsh->GetBBCollision();

        lpStaticMeshes[ dwStaticMesh ].dwObjectId           = lpStatMsh->dwGetObjectId();
        lpStaticMeshes[ dwStaticMesh ].wFlags               = lpStatMsh->wGetFlags();
        lpStaticMeshes[ dwStaticMesh ].wMesh                = dwNumMeshPointers;
        lpStaticMeshes[ dwStaticMesh ].BoundingBox[0].sMinX =  (BBVis.GetBBLo())[0];
        lpStaticMeshes[ dwStaticMesh ].BoundingBox[0].sMaxY = -(BBVis.GetBBLo())[1];
        lpStaticMeshes[ dwStaticMesh ].BoundingBox[0].sMaxZ = -(BBVis.GetBBLo())[2];
        lpStaticMeshes[ dwStaticMesh ].BoundingBox[0].sMaxX =  (BBVis.GetBBHi())[0];
        lpStaticMeshes[ dwStaticMesh ].BoundingBox[0].sMinY = -(BBVis.GetBBHi())[1];
        lpStaticMeshes[ dwStaticMesh ].BoundingBox[0].sMinZ = -(BBVis.GetBBHi())[2];
        lpStaticMeshes[ dwStaticMesh ].BoundingBox[1].sMinX =  (BBCol.GetBBLo())[0];
        lpStaticMeshes[ dwStaticMesh ].BoundingBox[1].sMaxY = -(BBCol.GetBBLo())[1];
        lpStaticMeshes[ dwStaticMesh ].BoundingBox[1].sMaxZ = -(BBCol.GetBBLo())[2];
        lpStaticMeshes[ dwStaticMesh ].BoundingBox[1].sMaxX =  (BBCol.GetBBHi())[0];
        lpStaticMeshes[ dwStaticMesh ].BoundingBox[1].sMinY = -(BBCol.GetBBHi())[1];
        lpStaticMeshes[ dwStaticMesh ].BoundingBox[1].sMinZ = -(BBCol.GetBBHi())[2];

        dwStaticMesh++;
        dwNumMeshPointers++;
        
        delete lpStatMsh;

        DlgSave.vSetConvertProgress(++dwProgress);
        }

    delete [] trFile.lpStaticMeshes;

    trFile.dwNumStaticMeshes    = dwNumStaticMeshes;
    trFile.lpStaticMeshes       = lpStaticMeshes;

    //
    // Process items
    //
    TRListItem* lpListItem = trFile.lpListItem;
    DWORD       dwNewNumItems = lpListItem->size();

    TR123_ITEM* lpItems = new TR123_ITEM [ dwNewNumItems ];

    delete [] trFile.lpItems;

    trFile.lpItems = lpItems;
    trFile.dwNumItems = dwNewNumItems;

    DWORD   dwItem  = 0;

    for(TRListItem::iterator theIterator4 = lpListItem->begin(); theIterator4 != lpListItem->end(); theIterator4++)
        {
        TR123_ITEM* lpItem = *theIterator4;

        lpItems[dwItem++] = *lpItem;
        }

    DlgSave.vSetConvertProgress(++dwProgress);

    //
    // Process sounds
    //
    trFile.lpSndMgr->vConvertToTR(this);

    //
    // Rebuild the new list of moveables
    //
    lpListMoveable->clear();

    for(DWORD dwNumMoveable = 0; dwNumMoveable < m_trFile.tr123File.dwNumMoveables; dwNumMoveable++)
        {
        TR2_MOVEABLE* lpMoveable = &m_trFile.tr123File.lpMoveables[dwNumMoveable];

        lpListMoveable->push_back(new CTrMvb(lpMoveable,this));

        DlgSave.vSetConvertProgress(++dwProgress);
        }

    //
    // Rebuild the new list of static meshes
    //
    lpListStatMsh->clear();

    for(DWORD dwNumStaticMesh = 0; dwNumStaticMesh < m_trFile.tr123File.dwNumStaticMeshes; dwNumStaticMesh++)
        {
        TR2_STATICMESH* lpStatMesh = &m_trFile.tr123File.lpStaticMeshes[dwNumStaticMesh];

        lpListStatMsh->push_back(new CTrStatMsh(lpStatMesh,this));

        DlgSave.vSetConvertProgress(++dwProgress);
        }

    //
    // Rebuild the new list of sprites
    //
    lpListSpriteSeq->clear();

    for(DWORD dwNumSprSeq = 0; dwNumSprSeq < m_trFile.tr123File.dwNumSpriteSequences; dwNumSprSeq++)
        {
        TR2_SPRITE_SEQUENCE* lpSprSeq= &m_trFile.tr123File.lpSpriteSequences[dwNumSprSeq];

        lpListSpriteSeq->push_back(new CTrLstSpr(lpSprSeq,this));

        DlgSave.vSetConvertProgress(++dwProgress);
        }

    DlgSave.vSetConvertProgress(dwMaxRange);
}


//----------------------------------------------------------------------------
//
//  Function:   vConvertFromTRLevel
//
//  Synopsis:   Convert TR level to internal representation
//
//  Arguments:  [DlgLoad] -- load dialog (to update progress bars)
//
//  Return:     -
//
//----------------------------------------------------------------------------
void    CTRFile::vConvertFromTRLevel(CDlgLevelLoad& DlgLoad)
{
    DlgLoad.vSetConvertRange(   m_trFile.tr123File.dwNumMeshes + m_trFile.tr123File.dwNumMoveables + 
                                m_trFile.tr123File.dwNumStaticMeshes + m_trFile.tr123File.dwNumSpriteSequences +
                                m_trFile.tr123File.dwNumItems);

    DWORD   dwProgress = 0;

    //
    // Set the "intensity alpha" flag for meshes
    //
    for(DWORD dwNumMesh = 0; dwNumMesh < m_trFile.tr123File.dwNumMeshes; dwNumMesh++)
        {
        TR123_MESH* lpMesh = m_trFile.tr123File.lplpMeshes[dwNumMesh];

        lpMesh->boIntensityAlpha = CTRFile::boIsMeshIntensityAlpha(lpMesh,m_trFile.tr123File.lpObjectTextures);

        DlgLoad.vSetConvertProgress(++dwProgress);
        }

    //
    // Process moveables
    //
    m_trFile.tr123File.lpListMoveable = new TRListMoveable;

    for(DWORD dwNumMoveable = 0; dwNumMoveable < m_trFile.tr123File.dwNumMoveables; dwNumMoveable++)
        {
        TR2_MOVEABLE* lpMoveable = &m_trFile.tr123File.lpMoveables[dwNumMoveable];

        (*m_trFile.tr123File.lpListMoveable).push_back(new CTrMvb(lpMoveable,this));

        DlgLoad.vSetConvertProgress(++dwProgress);
        }

    //
    // Process static meshes
    //
    m_trFile.tr123File.lpListStaticMesh = new TRListStaticMesh;

    for(DWORD dwNumStaticMesh = 0; dwNumStaticMesh < m_trFile.tr123File.dwNumStaticMeshes; dwNumStaticMesh++)
        {
        TR2_STATICMESH* lpStatMesh = &m_trFile.tr123File.lpStaticMeshes[dwNumStaticMesh];

        (*m_trFile.tr123File.lpListStaticMesh).push_back(new CTrStatMsh(lpStatMesh,this));

        DlgLoad.vSetConvertProgress(++dwProgress);
        }

    //
    // Process sprite sequences 
    //
    m_trFile.tr123File.lpListSpriteSeq = new TRListSpriteSeq;

    for(DWORD dwNumSprSeq = 0; dwNumSprSeq < m_trFile.tr123File.dwNumSpriteSequences; dwNumSprSeq++)
        {
        TR2_SPRITE_SEQUENCE* lpSprSeq= &m_trFile.tr123File.lpSpriteSequences[dwNumSprSeq];

        (*m_trFile.tr123File.lpListSpriteSeq).push_back(new CTrLstSpr(lpSprSeq,this));

        DlgLoad.vSetConvertProgress(++dwProgress);
        }

    //
    // Process items
    //
    m_trFile.tr123File.lpListItem = new TRListItem;

    for(DWORD dwNumItem = 0; dwNumItem < m_trFile.tr123File.dwNumItems; dwNumItem++)
        {
        TR123_ITEM* lpItem = &m_trFile.tr123File.lpItems[dwNumItem];
        
        TR123_ITEM* lpItem2 = new TR123_ITEM;

        *lpItem2 = *lpItem;

        (*m_trFile.tr123File.lpListItem).push_back(lpItem2);

        DlgLoad.vSetConvertProgress(++dwProgress);
        }

    //
    // Process sounds
    //
    m_trFile.tr123File.lpSndMgr = new CTrSndMgr(this);
}
