#include "stdafx.h"

#include "TextureTools.h"


//----------------------------------------------------------------------------
//
//  Function:   lpbReadMesh
//
//  Synopsis:   Read a mesh geometry
//
//  Arguments:  [lpMesh] -- dest address: where to put data
//              [lpData] -- source address: where to read data
//              [lpbAddrStartMesh] -- starting address of the meshes stuff. We
//               need this address to offset the current pointer after having
//               read the mesh data
//              [boExtraInfos] -- true to take into account extra infos
//
//  Return:     source address after having read the mesh
//
//----------------------------------------------------------------------------
BYTE*   CTRFile::lpbReadMesh(TR123_MESH* lpMesh,BYTE* lpData,BYTE* lpbAddrStartMesh,bool boExtraInfos /*= false*/)
{
    DWORD dwSize;

    lpMesh->Center                    = *(TR2_VERTEX*)         lpData;  lpData += sizeof(TR2_VERTEX);
    lpMesh->lCollisionSize            = *(bit32*)              lpData;  lpData += sizeof(bit32);

    READ_ARRAYNUM(bit16,TR2_VERTEX,lpMesh->sNumVertices,lpMesh->lpVertices)

    bit16   sNumNormals               = *(bit16*)              lpData;  lpData += sizeof(bit16);
    if(sNumNormals > 0)
        {
        lpMesh->Lighting.LightingType   = MESH_LIGHTING_EXTERNAL;
        lpMesh->Lighting.sNumData       = sNumNormals;

        READ_ARRAY(sNumNormals,TR2_VERTEX,lpMesh->Lighting.lpNormals)
        }
    else
        {
        sNumNormals = -sNumNormals;
        lpMesh->Lighting.LightingType   = MESH_LIGHTING_INTERNAL;
        lpMesh->Lighting.sNumData       = sNumNormals;

        READ_ARRAY(sNumNormals,bit16,lpMesh->Lighting.lpLights)
        }

    READ_ARRAYNUM(bit16,TR2_FACE4,lpMesh->sNumTexturedRectangles,lpMesh->lpTexturedRectangles)
    READ_ARRAYNUM(bit16,TR2_FACE3,lpMesh->sNumTexturedTriangles,lpMesh->lpTexturedTriangles)
    READ_ARRAYNUM(bit16,TR2_FACE4,lpMesh->sNumColoredRectangles,lpMesh->lpColoredRectangles)
    READ_ARRAYNUM(bit16,TR2_FACE3,lpMesh->sNumColoredTriangles,lpMesh->lpColoredTriangles)

    if(boExtraInfos)
        {
        BYTE    bIsDummy = *(BYTE*)lpData;          lpData += sizeof(BYTE);
        lpMesh->boIsDummy = bIsDummy ? true : false;

        BYTE    bIntensityAlpha = *(BYTE*)lpData;   lpData += sizeof(BYTE);
        lpMesh->boIntensityAlpha = bIntensityAlpha ? true : false;
        }
    else
        {
        lpMesh->boIsDummy           = false;
        lpMesh->boIntensityAlpha    = false;
        }

    if((lpData - lpbAddrStartMesh) & 3) lpData += 4 - ((lpData - lpbAddrStartMesh) & 3);

    return lpData;
}


//----------------------------------------------------------------------------
//
//  Function:   lpbWriteMesh
//
//  Synopsis:   Write a mesh geometry into memory
//
//  Arguments:  [lpMesh] -- src address: where to read data from
//              [lpData] -- dst address: where to put data in
//              [lpbAddrStartMesh] -- starting address of the meshes stuff. We
//               need this address to offset the current pointer after having
//               written the mesh data
//              [boExtraInfos] -- true to take into account extra infos
//
//  Return:     dst address after having written the mesh
//
//----------------------------------------------------------------------------
BYTE*   CTRFile::lpbWriteMesh(TR123_MESH* lpMesh,BYTE* lpData,BYTE* lpbAddrStartMesh,bool boExtraInfos /*= false*/)
{
    DWORD dwSize;

    * ((TR2_VERTEX*) lpData) = lpMesh->Center;  lpData += sizeof(TR2_VERTEX);

    * ((bit32*) lpData) = lpMesh->lCollisionSize;  lpData += sizeof(bit32);

    WRITE_ARRAYNUM(bit16,TR2_VERTEX,lpMesh->sNumVertices,lpMesh->lpVertices)

    bit16   sNumNormals               = lpMesh->Lighting.sNumData;
    if(lpMesh->Lighting.LightingType == MESH_LIGHTING_INTERNAL) sNumNormals = -sNumNormals;

    * ((bit16*) lpData) = sNumNormals;  lpData += sizeof(bit16);

    if(lpMesh->Lighting.LightingType == MESH_LIGHTING_EXTERNAL)
        {
        WRITE_ARRAY(lpMesh->Lighting.sNumData,TR2_VERTEX,lpMesh->Lighting.lpNormals)
        }
    else
        {
        WRITE_ARRAY(lpMesh->Lighting.sNumData,bit16,lpMesh->Lighting.lpLights)
        }

    WRITE_ARRAYNUM(bit16,TR2_FACE4,lpMesh->sNumTexturedRectangles,lpMesh->lpTexturedRectangles)
    WRITE_ARRAYNUM(bit16,TR2_FACE3,lpMesh->sNumTexturedTriangles,lpMesh->lpTexturedTriangles)
    WRITE_ARRAYNUM(bit16,TR2_FACE4,lpMesh->sNumColoredRectangles,lpMesh->lpColoredRectangles)
    WRITE_ARRAYNUM(bit16,TR2_FACE3,lpMesh->sNumColoredTriangles,lpMesh->lpColoredTriangles)

    if(boExtraInfos)
        {
        * (BYTE*) lpData = lpMesh->boIsDummy ? 1 : 0;           lpData += sizeof(BYTE);
        * (BYTE*) lpData = lpMesh->boIntensityAlpha ? 1 : 0;    lpData += sizeof(BYTE);
        }

    if((lpData - lpbAddrStartMesh) & 3) lpData += 4 - ((lpData - lpbAddrStartMesh) & 3);

    return lpData;
}


//----------------------------------------------------------------------------
//
//  Function:   dwCalcMeshDataSize
//
//  Synopsis:   Calculate the data size of a mesh
//
//  Arguments:  [lpMesh] -- mesh to calculate the data size
//              [dwMeshOffset] -- offset of this mesh from the first mesh
//              [boExtraInfos] -- true to take into account extra infos
//
//  Return:     Data size of the mesh
//
//----------------------------------------------------------------------------
DWORD   CTRFile::dwCalcMeshDataSize(TR123_MESH* lpMesh,DWORD dwMeshOffset,bool boExtraInfos /*= false*/)
{
    DWORD   dwSize =
        sizeof(TR2_VERTEX)
      + sizeof(bit32)
      + sizeof(bit16)
      + sizeof(TR2_VERTEX) * lpMesh->sNumVertices
      + sizeof(bit16)
      + lpMesh->Lighting.sNumData * ((lpMesh->Lighting.LightingType == MESH_LIGHTING_EXTERNAL) ? sizeof(TR2_VERTEX) : sizeof(bit16))
      + sizeof(bit16)
      + sizeof(TR2_FACE4) * lpMesh->sNumTexturedRectangles
      + sizeof(bit16)
      + sizeof(TR2_FACE3) * lpMesh->sNumTexturedTriangles
      + sizeof(bit16)
      + sizeof(TR2_FACE4) * lpMesh->sNumColoredRectangles
      + sizeof(bit16)
      + sizeof(TR2_FACE3) * lpMesh->sNumColoredTriangles;

    if(boExtraInfos)
        dwSize +=   sizeof(BYTE) +  // boIsDummy
                    sizeof(BYTE);   // boIntensityAlpha

    dwMeshOffset += dwSize;

    if(dwMeshOffset & 3) dwSize += 4 - (dwMeshOffset & 3);

    return dwSize;
}


//----------------------------------------------------------------------------
//
//  Function:   vComputeMeshExtents
//
//  Synopsis:   Compute the bounding box of a mesh
//
//  Arguments:  [lpMesh] -- Mesh to compute the bounding box
//
//  Return:     lpMesh->fMinX / fMaxX / fMinY / fMaxY / fMinZ / fMaxZ hold
//               the bounding box extents
//
//----------------------------------------------------------------------------
void    CTRFile::vComputeMeshExtents(TR123_MESH* lpMesh)
{
    lpMesh->fMinX = 1e20f; lpMesh->fMaxX = -1e20f;
    lpMesh->fMinY = 1e20f; lpMesh->fMaxY = -1e20f;
    lpMesh->fMinZ = 1e20f; lpMesh->fMaxZ = -1e20f;

    short sNumVertices = lpMesh->sNumVertices;

    while(sNumVertices--)
        {
        TR2_VERTEX* lpVertex = &lpMesh->lpVertices[sNumVertices];
        
        float   fX =  (float)lpVertex->sX;
        float   fY = -(float)lpVertex->sY;
        float   fZ = -(float)lpVertex->sZ;

        if(fX < lpMesh->fMinX) lpMesh->fMinX = fX; else if(fX > lpMesh->fMaxX) lpMesh->fMaxX = fX;
        if(fY < lpMesh->fMinY) lpMesh->fMinY = fY; else if(fY > lpMesh->fMaxY) lpMesh->fMaxY = fY;
        if(fZ < lpMesh->fMinZ) lpMesh->fMinZ = fZ; else if(fZ > lpMesh->fMaxZ) lpMesh->fMaxZ = fZ;
        }
}


//----------------------------------------------------------------------------
//
//  Function:   vComputeMeshNormals
//
//  Synopsis:   Compute the normals for a mesh
//
//  Arguments:  [lpMesh] -- Mesh to compute the normals for
//
//  Return:     -
//
//----------------------------------------------------------------------------
void    CTRFile::vComputeMeshNormals(TR123_MESH* lpMesh)
{
    bit16 sNumVertices = lpMesh->sNumVertices;

    TR2_VERTEX* lpLstVertices = lpMesh->lpVertices;

    short sNumTexturedRectangles    = lpMesh->sNumTexturedRectangles;
    short sNumTexturedTriangles     = lpMesh->sNumTexturedTriangles;
    short sNumColoredRectangles     = lpMesh->sNumColoredRectangles;
    short sNumColoredTriangles      = lpMesh->sNumColoredTriangles;

    vec3*   lpLstPtNormals = new vec3[sNumVertices];

    ZeroMemory(lpLstPtNormals,sNumVertices * sizeof(vec3));

    vComputeFaceArrayNormals(lpLstVertices,lpLstPtNormals,sNumTexturedTriangles,lpMesh->lpTexturedTriangles,false);
    vComputeFaceArrayNormals(lpLstVertices,lpLstPtNormals,sNumColoredTriangles,lpMesh->lpColoredTriangles,false);
    vComputeFaceArrayNormals(lpLstVertices,lpLstPtNormals,sNumTexturedRectangles,lpMesh->lpTexturedRectangles,true);
    vComputeFaceArrayNormals(lpLstVertices,lpLstPtNormals,sNumColoredRectangles,lpMesh->lpColoredRectangles,true);

    for(bit16 sI = 0; sI < sNumVertices; sI++)
        {
        float fLength = lpLstPtNormals[sI].length();

        if(fLength != 0.0f) lpLstPtNormals[sI] /= fLength;

        lpLstPtNormals[sI] *= 8192.0f;

        lpMesh->Lighting.lpNormals[sI].sX = (bit16)lpLstPtNormals[sI][0];
        lpMesh->Lighting.lpNormals[sI].sY = (bit16)lpLstPtNormals[sI][1];
        lpMesh->Lighting.lpNormals[sI].sZ = (bit16)lpLstPtNormals[sI][2];
        }

    delete [] lpLstPtNormals;
}


//----------------------------------------------------------------------------
//
//  Function:   vCopyMesh
//
//  Synopsis:   Copy a mesh in another mesh.
//
//  Arguments:  [lpMeshSrc] -- Source mesh
//              [lpMeshDst] -- Destination mesh
//
//  Return:     -
//
//----------------------------------------------------------------------------
void    CTRFile::vCopyMesh(TR123_MESH* lpMeshSrc,TR123_MESH* lpMeshDst)
{
    *lpMeshDst = *lpMeshSrc;

    lpMeshDst->lpVertices = new TR2_VERTEX[lpMeshDst->sNumVertices];
    memcpy(lpMeshDst->lpVertices,lpMeshSrc->lpVertices,sizeof(TR2_VERTEX) * lpMeshDst->sNumVertices);

    if(lpMeshDst->Lighting.LightingType == MESH_LIGHTING_INTERNAL)
        {
        lpMeshDst->Lighting.lpLights = new bit16[lpMeshDst->Lighting.sNumData];
        memcpy(lpMeshDst->Lighting.lpLights,lpMeshSrc->Lighting.lpLights,lpMeshDst->Lighting.sNumData * sizeof(bit16));
        }
    else
        {
        lpMeshDst->Lighting.lpNormals = new TR2_VERTEX[lpMeshDst->Lighting.sNumData];
        memcpy(lpMeshDst->Lighting.lpNormals,lpMeshSrc->Lighting.lpNormals,lpMeshDst->Lighting.sNumData * sizeof(TR2_VERTEX));
        }

    lpMeshDst->lpTexturedRectangles = NULL;

    if(lpMeshDst->sNumTexturedRectangles)
        {
        lpMeshDst->lpTexturedRectangles = new TR2_FACE4[lpMeshDst->sNumTexturedRectangles];
        memcpy(lpMeshDst->lpTexturedRectangles,lpMeshSrc->lpTexturedRectangles,sizeof(TR2_FACE4) * lpMeshDst->sNumTexturedRectangles);
        }

    lpMeshDst->lpTexturedTriangles = NULL;

    if(lpMeshDst->sNumTexturedTriangles)
        {
        lpMeshDst->lpTexturedTriangles = new TR2_FACE3[lpMeshDst->sNumTexturedTriangles];
        memcpy(lpMeshDst->lpTexturedTriangles,lpMeshSrc->lpTexturedTriangles,sizeof(TR2_FACE3) * lpMeshDst->sNumTexturedTriangles);
        }

    lpMeshDst->lpColoredRectangles = NULL;

    if(lpMeshDst->sNumColoredRectangles)
        {
        lpMeshDst->lpColoredRectangles = new TR2_FACE4[lpMeshDst->sNumColoredRectangles];
        memcpy(lpMeshDst->lpColoredRectangles,lpMeshSrc->lpColoredRectangles,sizeof(TR2_FACE4) * lpMeshDst->sNumColoredRectangles);
        }

    lpMeshDst->lpColoredTriangles = NULL;

    if(lpMeshDst->sNumColoredTriangles)
        {
        lpMeshDst->lpColoredTriangles = new TR2_FACE3[lpMeshDst->sNumColoredTriangles];
        memcpy(lpMeshDst->lpColoredTriangles,lpMeshSrc->lpColoredTriangles,sizeof(TR2_FACE3) * lpMeshDst->sNumColoredTriangles);
        }
}


//----------------------------------------------------------------------------
//
//  Function:   vReleaseMeshMemory
//
//  Synopsis:   Release the memory eatten by a mesh. Caution: release only
//              the arrays in TR123_MESH*, not TR123_MESH* itself !
//
//  Arguments:  [lpMesh] -- Mesh to release memory
//
//  Return:     -
//
//----------------------------------------------------------------------------
void    CTRFile::vReleaseMeshMemory(TR123_MESH* lpMesh)
{
    delete [] lpMesh->lpColoredRectangles;  lpMesh->lpColoredRectangles     = NULL;
    delete [] lpMesh->lpColoredTriangles;   lpMesh->lpColoredTriangles      = NULL;
    delete [] lpMesh->lpTexturedRectangles; lpMesh->lpTexturedRectangles    = NULL;
    delete [] lpMesh->lpTexturedTriangles;  lpMesh->lpTexturedTriangles     = NULL;
    delete [] lpMesh->lpVertices;           lpMesh->lpVertices              = NULL;

    lpMesh->sNumColoredRectangles   = 0;
    lpMesh->sNumColoredTriangles    = 0;
    lpMesh->sNumTexturedRectangles  = 0;
    lpMesh->sNumTexturedTriangles   = 0;
    lpMesh->sNumVertices            = 0;

    if(lpMesh->Lighting.LightingType == MESH_LIGHTING_INTERNAL)
        {
        delete [] lpMesh->Lighting.lpLights;    lpMesh->Lighting.lpLights   = NULL;
        }
    else
        {
        delete [] lpMesh->Lighting.lpNormals;   lpMesh->Lighting.lpNormals  = NULL;
        }

    lpMesh->Lighting.sNumData = 0;
}


//----------------------------------------------------------------------------
//
//  Function:   vMeshComputeCollisionSphere
//
//  Synopsis:   Compute an acceptable center / radius of a collision sphere
//              for a mesh
//
//  Arguments:  [lpMesh] -- Mesh to compute collision sphere
//
//  Return:     -
//
//----------------------------------------------------------------------------
void    CTRFile::vComputeMeshCollisionSphere(TR123_MESH* lpMesh)
{
    vComputeMeshExtents(lpMesh);

    float fMidX = (lpMesh->fMinX + lpMesh->fMaxX) / 2.0f;
    float fMidY = (lpMesh->fMinY + lpMesh->fMaxY) / 2.0f;
    float fMidZ = (lpMesh->fMinZ + lpMesh->fMaxZ) / 2.0f;

    lpMesh->Center.sX = (bit16) -fMidX;
    lpMesh->Center.sY = (bit16) -fMidY;
    lpMesh->Center.sZ = (bit16) -fMidZ;

    float fRadius = ((lpMesh->fMaxX - lpMesh->fMinX) + (lpMesh->fMaxY - lpMesh->fMinY) + (lpMesh->fMaxZ - lpMesh->fMinZ)) / (3.0f * 2.0f);

    lpMesh->lCollisionSize = (bit32) fRadius;
}


//----------------------------------------------------------------------------
//
//  Function:   vOptimizeMesh
//
//  Synopsis:   Optimize a mesh -> eliminate duplicated vertices
//
//  Arguments:  [lpMesh] -- Mesh to optimize
//
//  Return:     -
//
//----------------------------------------------------------------------------
void    CTRFile::vOptimizeMesh(TR123_MESH* lpMesh)
{
    bit16 sNewNumVertices = 0;

    for(DWORD dwI = 0; dwI < (DWORD) lpMesh->sNumVertices; dwI++)
        {
        // Take a vertex one after the other
        TR2_VERTEX* lpVert = &lpMesh->lpVertices[dwI];

        bit16 sX = lpVert->sX;
        bit16 sY = lpVert->sY;
        bit16 sZ = lpVert->sZ;

        bool boNewVertex = true;    // false if a vertex with the same coordinates already exists

        // Look for another vertex with the same coordinates
        for(DWORD dwJ = 0; dwJ < (DWORD)sNewNumVertices; dwJ++)
            {
            if(lpMesh->lpVertices[dwJ].sX == sX  &&  lpMesh->lpVertices[dwJ].sY == sY  &&  lpMesh->lpVertices[dwJ].sZ == sZ)
                {
                boNewVertex = false;
                break;
                }
            }

        // We haven't found a vertex with same coordinates yet ?
        if(boNewVertex)
            {
            // No : add this vertex to the list
            dwJ = sNewNumVertices;
            lpMesh->lpVertices[sNewNumVertices++] = *lpVert;
            }

        // Set the new vertex index in TR2_FACE3 and TR2_FACE4 structures
        for(bit16 sI = 0; sI < lpMesh->sNumColoredRectangles; sI++)
            {
            bitu16* lpFace = lpMesh->lpColoredRectangles[sI].awVertices;

            if(lpFace[0] == dwI) lpFace[0] = (bit16) dwJ;
            if(lpFace[1] == dwI) lpFace[1] = (bit16) dwJ;
            if(lpFace[2] == dwI) lpFace[2] = (bit16) dwJ;
            if(lpFace[3] == dwI) lpFace[3] = (bit16) dwJ;
            }

        for(sI = 0; sI < lpMesh->sNumColoredTriangles; sI++)
            {
            bitu16* lpFace = lpMesh->lpColoredTriangles[sI].awVertices;

            if(lpFace[0] == dwI) lpFace[0] = (bit16) dwJ;
            if(lpFace[1] == dwI) lpFace[1] = (bit16) dwJ;
            if(lpFace[2] == dwI) lpFace[2] = (bit16) dwJ;
            }

        for(sI = 0; sI < lpMesh->sNumTexturedRectangles; sI++)
            {
            bitu16* lpFace = lpMesh->lpTexturedRectangles[sI].awVertices;

            if(lpFace[0] == dwI) lpFace[0] = (bit16) dwJ;
            if(lpFace[1] == dwI) lpFace[1] = (bit16) dwJ;
            if(lpFace[2] == dwI) lpFace[2] = (bit16) dwJ;
            if(lpFace[3] == dwI) lpFace[3] = (bit16) dwJ;
            }

        for(sI = 0; sI < lpMesh->sNumTexturedTriangles; sI++)
            {
            bitu16* lpFace = lpMesh->lpTexturedTriangles[sI].awVertices;
            
            if(lpFace[0] == dwI) lpFace[0] = (bit16) dwJ;
            if(lpFace[1] == dwI) lpFace[1] = (bit16) dwJ;
            if(lpFace[2] == dwI) lpFace[2] = (bit16) dwJ;
            }
        }

    // Allocate an array of vertices of the new vertex count and replace the old one
    TR2_VERTEX* lpNewVertices = new TR2_VERTEX[sNewNumVertices];

    memcpy(lpNewVertices,lpMesh->lpVertices,sNewNumVertices * sizeof(TR2_VERTEX));

    delete [] lpMesh->lpVertices;

    lpMesh->lpVertices      = lpNewVertices;
    lpMesh->sNumVertices    = sNewNumVertices;
}


//----------------------------------------------------------------------------
//
//  Function:   vUpdateTextureData
//
//  Synopsis:   Add new tiles / object textures according to the material map
//              passed
//
//  Arguments:  [MapMat] -- material map
//
//  Return:     -
//
//----------------------------------------------------------------------------
void    CTRFile::vUpdateTextureData(MATNAME2INFO& MapMat)
{
    TR123_FILE_STRUCT& trFile   = m_trFile.tr123File;

    DWORD dwNumTile             = trFile.dwNumTextiles;
    DWORD dwNumObjectTextures   = trFile.dwNumObjectTextures;
    DWORD dwNumNewTile          = 0;
    DWORD dwNumNewObjText       = 0;

    // Calculate the number of new tiles / object textures to create
    for(MATNAME2INFO::iterator theIter = MapMat.begin(); theIter != MapMat.end(); theIter++)
        {
        INFO_MAT* lpInfoMat = theIter->second;

        if(lpInfoMat->boUsed)
            {
            BYTE* lpbTexture = lpbEnlarge32BitsBitmap(lpInfoMat->lpbData,lpInfoMat->dwWidth,lpInfoMat->dwHeight,256,256);

            if(!lpbTexture)
                {
                CString s;
        
                s.Format(_T("Can't handle texture %s because its width and/or height is greater than 256 pixels !"),lpInfoMat->szTextName);
        
                AfxMessageBox(s,MB_OK | MB_ICONEXCLAMATION | MB_TOPMOST);
        
                delete [] lpInfoMat->lpbData;
                }
            else
                {
                delete [] lpInfoMat->lpbData;
                lpInfoMat->lpbData = lpbTexture;

                dwNumNewTile++;
                for(LISTMESHMAT::iterator theIter2 = lpInfoMat->ListMeshMat.begin(); theIter2 != lpInfoMat->ListMeshMat.end(); theIter2++)
                    {
                    INFO_MESH_MAT* lpMeshMat = *theIter2;
                    dwNumNewObjText += lpMeshMat->dwNumObjText;
                    }
                }
            }
        }

    // Allocate new arrays for tiles / object textures
    TR2_OBJECT_TEXTURE* lpObjText   = new TR2_OBJECT_TEXTURE[dwNumObjectTextures + dwNumNewObjText];

    memcpy(lpObjText,   trFile.lpObjectTextures,sizeof(TR2_OBJECT_TEXTURE)  * dwNumObjectTextures);

    TR2_TEXTILE16** lplp16          = NULL;
    
    lplp16 = new TR2_TEXTILE16*[dwNumTile + dwNumNewTile];
    memcpy(lplp16,      trFile.lplpTextiles16,  sizeof(TR2_TEXTILE16*)      * dwNumTile);

    // Get the new tiles / object textures from the material map
    for(theIter = MapMat.begin(); theIter != MapMat.end(); theIter++)
        {
        INFO_MAT* lpInfoMat = theIter->second;

        if(lpInfoMat->boUsed)
            {
            TR2_TEXTILE16* lp16 = new TR2_TEXTILE16;
            vConvertABGRToARGB(256,256,false,lpInfoMat->lpbData,(BYTE*)lp16);
            lplp16[dwNumTile] = lp16;

            for(LISTMESHMAT::iterator theIter2 = lpInfoMat->ListMeshMat.begin(); theIter2 != lpInfoMat->ListMeshMat.end(); theIter2++)
                {
                INFO_MESH_MAT* lpMeshMat = *theIter2;

                DWORD   dwCurrNumObjectTextures = dwNumObjectTextures;

                for(DWORD dwI = 0; dwI < lpMeshMat->dwNumObjText; dwI++)
                    {
                    lpObjText[dwNumObjectTextures]          = lpMeshMat->lpLstTextCoord[dwI];
                    lpObjText[dwNumObjectTextures].wTile    = dwNumTile;
                    dwNumObjectTextures++;
                    }

                for(dwI = 0; dwI < lpMeshMat->dwNumTris; dwI++)
                    {
                    lpMeshMat->lpMesh->lpTexturedTriangles[lpMeshMat->dwStartTriIndex + dwI].wTextureIndex += dwCurrNumObjectTextures;
                    }

                for(dwI = 0; dwI < lpMeshMat->dwNumQuads; dwI++)
                    {
                    lpMeshMat->lpMesh->lpTexturedRectangles[lpMeshMat->dwStartQuadIndex + dwI].wTextureIndex += dwCurrNumObjectTextures;
                    }

                delete [] lpMeshMat->lpLstTextCoord;
                delete lpMeshMat;
                }

            dwNumTile++;
            }

        delete [] lpInfoMat->lpbData;

        delete lpInfoMat;
        }

    // Set the new tiles / object textures
    trFile.dwNumTextiles        = dwNumTile;
    trFile.dwNumObjectTextures  = dwNumObjectTextures;

    delete [] trFile.lpObjectTextures;

    trFile.lpObjectTextures = lpObjText;

    delete [] trFile.lplpTextiles16;
    trFile.lplpTextiles16   = lplp16;
}


//----------------------------------------------------------------------------
//
//  Function:   vInitializeMesh
//
//  Synopsis:   Initialize the structure of a mesh
//
//  Arguments:  [lpMesh] -- mesh to initialize
//              [dwNumVert] -- num vertices in the mesh
//              [dwNumTrisMap] -- num textured triangles in the mesh
//              [dwNumRectMap] -- num textured rectangles in the mesh
//              [boInternalLighting] -- true if the mesh uses internal
//               lighting
//              [boTR3] -- true if this mesh is a mesh of a TR3 level
//
//  Return:     -
//
//----------------------------------------------------------------------------
void    CTRFile::vInitializeMesh(TR123_MESH* lpMesh,DWORD dwNumVert,DWORD dwNumTrisMap,DWORD dwNumRectMap,bool boInternalLighting,bool boTR3)
{
    DWORD   dwNumTris = 0;  // colored triangles
    DWORD   dwNumRect = 0;  // colored rectangles

    lpMesh->boIsDummy               = false;
    lpMesh->boIntensityAlpha        = false;
    lpMesh->sNumColoredRectangles   = dwNumRect;
    lpMesh->sNumColoredTriangles    = dwNumTris;
    lpMesh->Lighting.LightingType   = boInternalLighting ? MESH_LIGHTING_INTERNAL : MESH_LIGHTING_EXTERNAL;
    lpMesh->Lighting.sNumData       = (bit16) dwNumVert;
    lpMesh->sNumTexturedRectangles  = dwNumRectMap;
    lpMesh->sNumTexturedTriangles   = dwNumTrisMap;
    lpMesh->sNumVertices            = (bit16) dwNumVert;
    lpMesh->lpColoredRectangles     = dwNumRect ? new TR2_FACE4[dwNumRect] : NULL;
    lpMesh->lpColoredTriangles      = dwNumTris ? new TR2_FACE3[dwNumTris] : NULL;

    if(boInternalLighting)
        {
        lpMesh->Lighting.lpLights   = new bit16[dwNumVert];
        for(DWORD dwI = 0; dwI < dwNumVert; dwI++) lpMesh->Lighting.lpLights[dwI] = boTR3 ? 8192 : 0;
        }
    else
        {
        lpMesh->Lighting.lpNormals  = new TR2_VERTEX[dwNumVert];
        }

    lpMesh->lpTexturedRectangles    = dwNumRectMap ? new TR2_FACE4[dwNumRectMap] : NULL;
    lpMesh->lpTexturedTriangles     = dwNumTrisMap ? new TR2_FACE3[dwNumTrisMap] : NULL;
    lpMesh->lpVertices              = new TR2_VERTEX[dwNumVert];
}


//----------------------------------------------------------------------------
//
//  Function:   vUpdateMeshData
//
//  Synopsis:   Recalculate some mesh data
//
//  Arguments:  [lpMesh] -- mesh
//              [dwIdxTriMap] -- num textured triangles of the mesh. If
//               dwIdxTriMap is less than the actual num textured triangles in
//               the mesh, the triangles array is reallocated at the new size
//              [boInternalLighting] -- true if the mesh uses internal
//               lighting
//
//  Return:     -
//
//----------------------------------------------------------------------------
void    CTRFile::vUpdateMeshData(TR123_MESH* lpMesh,DWORD dwIdxTriMap,bool boInternalLighting)
{
    // Set the number of textured triangles and reallocate
    // arrays to the right size
    if(lpMesh->sNumTexturedTriangles > (bit16) dwIdxTriMap)
        {
        lpMesh->sNumTexturedTriangles = (bit16) dwIdxTriMap;

        TR2_FACE3*  lpTextured = NULL;
        
        if(dwIdxTriMap)
            {
            lpTextured = new TR2_FACE3 [dwIdxTriMap];
            memcpy(lpTextured, lpMesh->lpTexturedTriangles, sizeof(TR2_FACE3) * dwIdxTriMap);
            }

        delete [] lpMesh->lpTexturedTriangles;
        lpMesh->lpTexturedTriangles = lpTextured;
        }

    // Eliminate duplicated vertices
    vOptimizeMesh(lpMesh);

    // Calculate point normals if necessary
    if(!boInternalLighting) CTRFile::vComputeMeshNormals(lpMesh);

    // Compute an approximated collision sphere
    vComputeMeshCollisionSphere(lpMesh);
}


//----------------------------------------------------------------------------
//
//  Function:   vChangeMeshLightingType
//
//  Synopsis:   Change the mesh lighting type (reallocate memory / recalculate
//              data if necessary)
//
//  Arguments:  [lpMesh] -- mesh
//              [NewLightingType] -- new lighting type for this mesh
//              [boTR3] -- true if this mesh is a mesh of a TR3 level
//
//  Return:     -
//
//----------------------------------------------------------------------------
void    CTRFile::vChangeMeshLightingType(TR123_MESH* lpMesh,enumMeshLighting NewLightingType,bool boTR3)
{
    if(NewLightingType != lpMesh->Lighting.LightingType)
        {
        switch(NewLightingType)
            {
            case MESH_LIGHTING_EXTERNAL:
                {
                delete [] lpMesh->Lighting.lpLights;
                lpMesh->Lighting.lpNormals = new TR2_VERTEX[lpMesh->Lighting.sNumData];
                break;
                }
            case MESH_LIGHTING_INTERNAL:
                {
                delete [] lpMesh->Lighting.lpNormals;
                lpMesh->Lighting.lpLights = new bit16[lpMesh->Lighting.sNumData];
                break;
                }
            }
        lpMesh->Lighting.LightingType = NewLightingType;
        }

    switch(NewLightingType)
        {
        case MESH_LIGHTING_EXTERNAL:
            {
            vComputeMeshNormals(lpMesh);
            break;
            }
        case MESH_LIGHTING_INTERNAL:
            {
            for(bit16 sI = 0; sI < lpMesh->Lighting.sNumData; sI++) lpMesh->Lighting.lpLights[sI] = boTR3 ? 8192 : 0;
            break;
            }
        }
}


//----------------------------------------------------------------------------
//
//  Function:   vTranslateMesh
//
//  Synopsis:   Translate the mesh vertices
//
//  Arguments:  [lpMesh] -- mesh
//              [sX,sY,sZ] -- translation on X / Y / Z axis to apply
//              [boDontModifyMoveables] -- true if the translation of the mesh
//               should not modify the hierarchy of moveables containing this
//               mesh
//              [lpTRFile] -- pointer on the CTRFile where this mesh comes
//               from. Can be NULL if boDontModifyMoveables = false
//
//  Return:     -
//
//----------------------------------------------------------------------------
void    CTRFile::vTranslateMesh(TR123_MESH* lpMesh,bit16 sX,bit16 sY,bit16 sZ,bool boDontModifyMoveables,CTRFile* lpTRFile)
{
    for(bit16 sI = 0; sI < lpMesh->sNumVertices; sI++)
        {
        lpMesh->lpVertices[sI].sX += sX;
        lpMesh->lpVertices[sI].sY += sY;
        lpMesh->lpVertices[sI].sZ += sZ;
        }

    if(!boDontModifyMoveables  ||  !lpTRFile) return;

    TR_FILE&    trFile = lpTRFile->GetTRFile();
    DWORD       dwNumMoveables = trFile.tr123File.dwNumMoveables;

    for(DWORD dwI2 = 0; dwI2 < dwNumMoveables; dwI2++)
        {
        TR2_MOVEABLE* lpMoveable = &trFile.tr123File.lpMoveables[dwI2];
        TR2_MESHTREE* lpMeshTree = &trFile.tr123File.lpMeshTrees[lpMoveable->dwMeshTree];

        #define STACK_DEPTHMAX 10

        DWORD   Stack[STACK_DEPTHMAX];
        DWORD   dwStackPointer = 0;
        DWORD   dwLevel = 0, dwMeshLevel = 255;

        for(DWORD dwI = 0; dwI < lpMoveable->wNumMeshes; dwI++)
            {
            DWORD dwNumMesh     = trFile.tr123File.lpMeshPointers[lpMoveable->wStartingMesh + dwI];
            TR123_MESH* lpMesh2 = trFile.tr123File.lplpMeshes[dwNumMesh];

            bool boOk = (lpMesh == lpMesh2);

            bitu32  dwStack = 0;
            
            if(dwI) { dwStack = lpMeshTree->lCoord; lpMeshTree++; }
        
            if(dwStack & 1)
                {
                if(dwStackPointer > 0)
                    {
                    dwStackPointer--;
                    dwLevel = Stack[dwStackPointer];
                    if(dwLevel <= dwMeshLevel) dwMeshLevel = 255;
                    }
                }

            if(boOk) dwMeshLevel = dwLevel;

            if(dwI)
                {
                bit32   lTransX = lpMeshTree->lCoord;
                if(boOk) lpMeshTree->lCoord -= sX;
                if(dwLevel == dwMeshLevel+1) lpMeshTree->lCoord += sX;
                lpMeshTree++;

                bit32   lTransY = lpMeshTree->lCoord;
                if(boOk) lpMeshTree->lCoord -= sY;
                if(dwLevel == dwMeshLevel+1) lpMeshTree->lCoord += sY;
                lpMeshTree++;

                bit32   lTransZ = lpMeshTree->lCoord;
                if(boOk) lpMeshTree->lCoord -= sZ;
                if(dwLevel == dwMeshLevel+1) lpMeshTree->lCoord += sZ;
                lpMeshTree++;
                }

            if(dwStack & 2)
                {
                if(dwStackPointer + 1 <= STACK_DEPTHMAX)
                    {
                    Stack[dwStackPointer] = dwLevel;
                    dwStackPointer++;
                    }
                }

            dwLevel++;
            }
        }
}


//----------------------------------------------------------------------------
//
//  Function:   boIsMeshIntensityAlpha
//
//  Synopsis:   Loop through textured polys and check the attribute of each
//              OBJECT_TEXTURE to see if all the faces have the
//              "intensity alpha" flag set. If yes, return true, else false
//
//  Arguments:  [lpMesh] -- mesh to scan
//              [lpOText] -- list of OBJECT_TEXTURE for this mesh
//
//  Return:     true if all the faces have the "intensity alpha" flag set,
//              else false
//
//----------------------------------------------------------------------------
/*static*/ bool CTRFile::boIsMeshIntensityAlpha(TR123_MESH* lpMesh,TR2_OBJECT_TEXTURE* lpOText)
{
    for(DWORD dwI = 0; dwI < lpMesh->sNumTexturedRectangles; dwI++)
        {
        bitu16  wAttribute = lpOText[lpMesh->lpTexturedRectangles[dwI].wTextureIndex].wAttribute;

        if(!(wAttribute & 2)) return false;
        }

    for(dwI = 0; dwI < lpMesh->sNumTexturedTriangles; dwI++)
        {
        bitu16  wAttribute = lpOText[lpMesh->lpTexturedTriangles[dwI].wTextureIndex].wAttribute;

        if(!(wAttribute & 2)) return false;
        }

    return true;
}


//----------------------------------------------------------------------------
//
//  Function:   vSetMeshIntensityAlpha
//
//  Synopsis:   Set / unset the "intensity alpha" flags for all the
//              OBJECT_TEXTURE of the faces of the mesh
//
//  Arguments:  [lpMesh] -- mesh to process
//              [lpOText] -- list of OBJECT_TEXTURE for this mesh
//              [boSet] -- true to set the flag, false to unset
//
//  Return:     -
//
//----------------------------------------------------------------------------
/*static*/ void CTRFile::vSetMeshIntensityAlpha(TR123_MESH* lpMesh,TR2_OBJECT_TEXTURE* lpOText, bool boSet)
{
    WORD wFlag = boSet ? 2 : 0;

    for(DWORD dwI = 0; dwI < lpMesh->sNumTexturedRectangles; dwI++)
        {
        bitu16  wAttribute = lpOText[lpMesh->lpTexturedRectangles[dwI].wTextureIndex].wAttribute;

        lpOText[lpMesh->lpTexturedRectangles[dwI].wTextureIndex].wAttribute = (wAttribute & (~2)) | wFlag;
        }

    for(dwI = 0; dwI < lpMesh->sNumTexturedTriangles; dwI++)
        {
        bitu16  wAttribute = lpOText[lpMesh->lpTexturedTriangles[dwI].wTextureIndex].wAttribute;

        lpOText[lpMesh->lpTexturedTriangles[dwI].wTextureIndex].wAttribute = (wAttribute & (~2)) | wFlag;
        }
}


//----------------------------------------------------------------------------
//
//  Function:   vSetMeshIntensity
//
//  Synopsis:   Set all the intensities of the vertices to a given value
//
//  Arguments:  [lpMesh] -- mesh to process
//              [fIntensity] -- intensity value between 0.0 and 1.0
//              [boTR3] -- true if this mesh is a mesh from a TR3 level
//
//  Return:     -
//
//----------------------------------------------------------------------------
/*static*/ void CTRFile::vSetMeshIntensity(TR123_MESH* lpMesh, float fIntensity, bool boTR3)
{
    if(lpMesh->Lighting.LightingType == MESH_LIGHTING_EXTERNAL) return;

    bit16 sIntensity = boTR3 ? fIntensity * 8192 : (1.0f - fIntensity) * 8192;

    for(DWORD dwI = 0; dwI < lpMesh->Lighting.sNumData; dwI++)
        {
        lpMesh->Lighting.lpLights[dwI] = sIntensity;
        }
}


//----------------------------------------------------------------------------
//
//  Function:   fCalcMeshAvgIntensity
//
//  Synopsis:   Calculate the average intensity of all the vertices
//
//  Arguments:  [lpMesh] -- mesh to process
//              [boTR3] -- true if this mesh is a mesh from a TR3 level
//
//  Return:     The average intensity (between 0.0 and 1.0) of all the
//              vertices if the mesh is internally lighted, else 0.0
//
//----------------------------------------------------------------------------
/*static*/ float    CTRFile::fCalcMeshAvgIntensity(TR123_MESH* lpMesh,bool boTR3)
{
    if(lpMesh->Lighting.LightingType == MESH_LIGHTING_EXTERNAL  ||  lpMesh->Lighting.sNumData == 0) return 0.0f;

    float fIntens = 0.0f;

    for(DWORD dwI = 0; dwI < lpMesh->Lighting.sNumData; dwI++)
        {
        if(boTR3)
            fIntens += lpMesh->Lighting.lpLights[dwI] / 8192.0f;
        else
            fIntens += (8192 - lpMesh->Lighting.lpLights[dwI]) / 8192.0f;
        }

    fIntens /= lpMesh->Lighting.sNumData;

    return fIntens;
}


//----------------------------------------------------------------------------
//
//  Function:   vMeshReverseLighting
//
//  Synopsis:   Reverse the intensities of all the vertices of the mesh
//
//  Arguments:  [lpMesh] -- mesh to process
//
//  Return:     -
//
//----------------------------------------------------------------------------
/*static*/ void CTRFile::vMeshReverseLighting(TR123_MESH* lpMesh)
{
    if(lpMesh->Lighting.LightingType == MESH_LIGHTING_EXTERNAL  ||  lpMesh->Lighting.sNumData == 0) return;

    for(DWORD dwI = 0; dwI < lpMesh->Lighting.sNumData; dwI++)
        {
        lpMesh->Lighting.lpLights[dwI] = 8192 - lpMesh->Lighting.lpLights[dwI];
        }
}


//----------------------------------------------------------------------------
//
//  Function:   lpGetStaticMesh
//
//  Synopsis:   Look for a given static mesh in the level
//
//  Arguments:  [dwObjId] -- id of the static mesh
//
//  Return:     pointer on the static mesh if found, else NULL
//
//----------------------------------------------------------------------------
CTrStatMsh* CTRFile::lpGetStaticMesh(DWORD dwObjId)
{
    TRListStaticMesh*   lpListStaticMesh = m_trFile.tr123File.lpListStaticMesh;

    for(TRListStaticMesh::iterator theIterator = lpListStaticMesh->begin(); theIterator != lpListStaticMesh->end(); theIterator++)
        {
        CTrStatMsh* lpMsh = *theIterator;

        if(lpMsh->dwGetObjectId() == dwObjId) return lpMsh;
        }

    return NULL;
}


//----------------------------------------------------------------------------
//
//  Function:   lpGetSprite
//
//  Synopsis:   Look for a given sprite in the level
//
//  Arguments:  [dwObjId] -- id of the sprite
//
//  Return:     pointer on the sprite if found, else NULL
//
//----------------------------------------------------------------------------
CTrLstSpr* CTRFile::lpGetSprite(DWORD dwObjId)
{
    TRListSpriteSeq*   lpListSprite = m_trFile.tr123File.lpListSpriteSeq;

    for(TRListSpriteSeq::iterator theIterator = lpListSprite->begin(); theIterator != lpListSprite->end(); theIterator++)
        {
        CTrLstSpr* lpSpr = *theIterator;

        if(lpSpr->dwGetObjectId() == dwObjId) return lpSpr;
        }

    return NULL;
}
