#include "stdafx.h"

#include "Resource.h"
#include "3DSStuff.h"
#include "TGAFile.h"
#include "TextureTools.h"

#include "AnimData.h"
#include "AnimKey.h"
#include "AnimTrack.h"
#include "TRAnimTrack.h"
#include "AnimMgr.h"


//#define _ANIM_

//#define EXPORT_256x256_TEXTURE


//
//
// Import moveable from 3DS
//
//
DWORD CTRFile::dwCreateParentArray(database3ds* db,long** lplpLstParent,DWORD dwNumNodes,namelist3ds* lpLstNames)
{
    DWORD dwNumRootMeshes = 0;
    
    *lplpLstParent = new long[dwNumNodes];

    for(DWORD dw = 0; dw < dwNumNodes; dw++)
        {
        kfmesh3ds* objMotion = NULL;
        GetObjectMotionByIndex3ds(db,dw,&objMotion);

        long dwParent = -1;

        for(long dw2 = dw - 1; dw2 >= 0; dw2--)
            {
            if(stricmp(lpLstNames->list[dw2].name,objMotion->parent) == 0)
                {
                dwParent = dw2;
                break;
                }
            }

        if(dwParent == -1) dwNumRootMeshes++;

        (*lplpLstParent)[dw] = dwParent;

        ReleaseObjectMotion3ds(&objMotion);
        }

    return dwNumRootMeshes;
}


void CTRFile::vCreateMoveableHierarchy(TR2_MESHTREE** lplpMeshTree,DWORD dwNumNodes,long* lpLstParent)
{
    // Create the hierarchy as understood by the TR engine
    *lplpMeshTree = new TR2_MESHTREE[(dwNumNodes - 1) * 4];

    ZeroMemory(*lplpMeshTree,(dwNumNodes - 1) * 4 * sizeof(TR2_MESHTREE));

    for(DWORD dw = 1; dw < dwNumNodes; dw++)
        {
        DWORD dwStack = 0;

        // Get the index of the parent mesh
        long dwParent = lpLstParent[dw];

        // For each mesh following the current mesh which has the parent dwParent, change
        // to say "your parent mesh is now dw (the current mesh)"
        for(long dw3 = (long)dw + 1; dw3 < (long)dwNumNodes; dw3++)
            {
            if(lpLstParent[dw3] == dwParent)
                {
                lpLstParent[dw3] = dw-1;    // substract 1 because the first mesh is not taken into account
                }
            }

        // If the parent is not the previous mesh, make some stuff with the stack...
        if(dwParent < (long)dw - 1)
            {
            // Get the stack state of the parent mesh
            dwStack = (*lplpMeshTree)[dwParent * 4 + 0].lCoord;

            // Change and save the stack state (2 means "push")
            (*lplpMeshTree)[dwParent * 4 + 0].lCoord = dwStack | 2;

            // The stack state of the current mesh is "pop"
            dwStack = 1;
            }

        (*lplpMeshTree)[(dw-1) * 4 + 0].lCoord = dwStack;
        (*lplpMeshTree)[(dw-1) * 4 + 1].lCoord = 0;
        (*lplpMeshTree)[(dw-1) * 4 + 2].lCoord = 0;
        (*lplpMeshTree)[(dw-1) * 4 + 3].lCoord = 0;
        }
}


void    vQuat2Euler( CQuaternion& Quat, FLOAT* lpfX, FLOAT* lpfY, FLOAT* lpfZ )
{
    CVect3D&    vec3 = Quat.v3GetVector();
    FLOAT       fX = vec3[0], fY = vec3[1], fZ = vec3[2];
    FLOAT       fW = Quat.fGetScalar();

    FLOAT       fR00 = 1.0 - 2.0 * fY * fY - 2.0 * fZ * fZ;
    FLOAT       fR01 = 2.0 * fX * fY + 2.0 * fW * fZ;
    FLOAT       fR02 = 2.0 * fX * fZ - 2.0 * fW * fY;
    FLOAT       fR10 = 2.0 * fX * fY - 2.0 * fW * fZ;
    FLOAT       fR11 = 1.0 - 2.0 * fX * fX - 2.0 * fZ * fZ;
    FLOAT       fR12 = 2.0 * fY * fZ + 2.0 * fW * fX;
    FLOAT       fR20 = 2.0 * fX * fZ + 2.0 * fW * fY;
    FLOAT       fR21 = 2.0 * fY * fZ - 2.0 * fW * fX;
    FLOAT       fR22 = 1.0 - 2.0 * fX * fX - 2.0 * fY * fY;

    *lpfX = -asin( fR21 );

    if( *lpfX < _PI_ / 2.0 )
        {
        if( *lpfX > -_PI_ / 2.0)
            {
            *lpfZ = atan2( fR01, fR11 );
            *lpfY = atan2( fR20, fR22 );
            }
        else
            {
            *lpfZ = -atan2( -fR02, fR00 );
            *lpfY = 0.0;
            }
        }
    else
        {
        *lpfZ = atan2( fR02, fR00 );
        *lpfY = 0.0;
        }
}


void CTRFile::vCreateMoveable(database3ds* db,CTrMvb* lpMvb,DWORD dwNumNodes,bool boInternalLighting,MATNAME2INFO* lpMapMat)
{
    #define STACK_DEPTHMAX 10

    float   Stack[3 * STACK_DEPTHMAX];
    DWORD   dwStackPointer = 0;
    float   fTransX = 0, fTransY = 0, fTransZ = 0;

    TR2_MESHTREE*   lpMeshTree  = lpMvb->lpGetMeshTree();
    TR123_MESH*     lpLstMeshes = new TR123_MESH [ dwNumNodes ];

    kfsets3ds* key = NULL;
    GetKfSets3ds(db,&key);

    DWORD    dwNumFrames = key->anim.length + 1;

    ReleaseKfSets3ds(&key);

    kfmesh3ds* objMotion = NULL;
    GetObjectMotionByIndex3ds(db,0,&objMotion);

    DWORD   dwAnimNumKeys = objMotion->nrkeys;
    DWORD   dwAnimFramerate = dwAnimNumKeys > 1 ? objMotion->rkeys[1].time - objMotion->rkeys[0].time : 1;

    ReleaseObjectMotion3ds(&objMotion);

    CVect3D*        lpTransRootMesh = new CVect3D [ dwAnimNumKeys + 1 ];
    CQuaternion*    lpRotMeshes     = new CQuaternion [ dwNumNodes * (dwAnimNumKeys + 1) ];

    CTRAnimTrack*   lpTrack = new CTRAnimTrack( dwAnimNumKeys, dwNumFrames, dwAnimFramerate, (FLOAT) 30.0f );

    // Scan through the meshes in the 3DS file
    for(DWORD dwI = 0; dwI < dwNumNodes; dwI++)
        {
        // Get the 3DS mesh
        kfmesh3ds* objMotion = NULL;
        GetObjectMotionByIndex3ds(db,dwI,&objMotion);

        mesh3ds* lpMesh3ds = NULL;
        GetMeshByName3ds(db,objMotion->name,&lpMesh3ds);

        bool boIsDummy = (strnicmp(objMotion->name,"dummy",5) == 0);

        if(lpMesh3ds)
            {
            DWORD dwIdxVert     = 0;
            DWORD dwIdxTriMap   = 0;
            DWORD dwNumTris     = lpMesh3ds->nfaces;
            DWORD dwNumVert     = lpMesh3ds->nvertices;

            TR123_MESH* lpMesh  = &lpLstMeshes[dwI];

            // Convert the 3DS mesh to a TR mesh
            if(!boIsDummy)
                {
                vInitializeMesh(lpMesh,dwNumVert,dwNumTris,0,boInternalLighting,true);

                vCopy3DSMeshIntoTRMesh(lpMesh,db,lpMesh3ds,dwIdxVert,dwIdxTriMap,lpMapMat);
                }
            else
                {
                CTRFile::vInitializeMesh(lpMesh,0,0,0,false,true);
                CTRFile::vComputeMeshCollisionSphere(lpMesh);
                lpMesh->boIsDummy = true;
                }

            // Set the translation of the mesh
            if(dwI != 0)
                {
                bitu32  dwStack = lpMeshTree->lCoord; lpMeshTree++;
                
                if(dwStack & 1)
                    {
                    dwStackPointer -= 3;
                    fTransX = Stack[dwStackPointer + 0];
                    fTransY = Stack[dwStackPointer + 1];
                    fTransZ = Stack[dwStackPointer + 2];
                    }

                if(dwStack & 2)
                    {
                    Stack[dwStackPointer + 0] = fTransX;
                    Stack[dwStackPointer + 1] = fTransY;
                    Stack[dwStackPointer + 2] = fTransZ;
                    dwStackPointer += 3;
                    }

                lpMeshTree->lCoord = (bit32)(objMotion->pos[0].x);  lpMeshTree++;
                lpMeshTree->lCoord = (bit32)(-objMotion->pos[0].z); lpMeshTree++;
                lpMeshTree->lCoord = (bit32)(objMotion->pos[0].y);  lpMeshTree++;

                fTransX +=   objMotion->pos[0].x;
                fTransY += - objMotion->pos[0].z;
                fTransZ +=   objMotion->pos[0].y;
                }
            else
                {
                for(DWORD dwKey = 0; dwKey < dwAnimNumKeys; dwKey++)
                    {
                    lpTransRootMesh[dwKey][0] =   objMotion->pos[dwKey].x;
                    lpTransRootMesh[dwKey][2] = - objMotion->pos[dwKey].y;
                    lpTransRootMesh[dwKey][1] =   objMotion->pos[dwKey].z;
                    }
                }

            CQuaternion Quat(0.0,0.0,0.0,1.0);

            for(DWORD dwKey = 0; dwKey < dwAnimNumKeys; dwKey++)
                {
                CVect3D         vecAxis;
                FLOAT           fAngle;

                fAngle      = - objMotion->rot[dwKey].angle;
                vecAxis[0]  =   objMotion->rot[dwKey].x;
                vecAxis[2]  = - objMotion->rot[dwKey].y;
                vecAxis[1]  =   objMotion->rot[dwKey].z;

                CQuaternion     QuatL;
            
                QuatL.vFromAxisAngle(vecAxis, fAngle);

                Quat = QuatL * Quat;

                lpRotMeshes[dwKey * dwNumNodes + dwI] = Quat;
                }

            if(!boIsDummy)
                {
                // Correct the vertices coordinates to take into account the translation
                // of the mesh with respect to the root mesh
                while(dwIdxVert--)
                    {
                    lpMesh->lpVertices[dwIdxVert].sX -= (bit16)fTransX;
                    lpMesh->lpVertices[dwIdxVert].sY -= (bit16)fTransY;
                    lpMesh->lpVertices[dwIdxVert].sZ -= (bit16)fTransZ;
                    }

                // Update data of the mesh (reallocate triangles array / compute point normals & collision sphere)
                vUpdateMeshData(lpMesh,dwIdxTriMap,boInternalLighting);
                }
            }

        RelMeshObj3ds(&lpMesh3ds);
        ReleaseObjectMotion3ds(&objMotion);
        }

    // Loop through all the keys of the track
    for(DWORD dwKey = 0; dwKey < dwAnimNumKeys; dwKey++)
        {
        // Build the key
        CAnimKey Key(dwKey * dwAnimFramerate, dwNumNodes, CBoundingBox(CVect3D(0,0,0),CVect3D(0,0,0)));

        // Loop through all the meshes of the key
        for(DWORD dwMesh = 0; dwMesh < dwNumNodes; dwMesh++)
            {
            CVect3D vecTrans(0,0,0);
            
            if(dwMesh == 0)
                {
                vecTrans = lpTransRootMesh[dwKey];
                }

            CQuaternion quat = lpRotMeshes[dwKey * dwNumNodes + dwMesh];

            FLOAT   fX, fY, fZ;

            vQuat2Euler( quat, &fX, &fY, &fZ );

            fY = -fY;   // OGL -> TR
            fZ = -fZ;

            WORD    wX = (WORD) (fX >= 0.0 ? fX * 512.0 / M_PI : fX * 512.0 / M_PI + 1024.0), 
                    wY = (WORD) (fY >= 0.0 ? fY * 512.0 / M_PI : fY * 512.0 / M_PI + 1024.0), 
                    wZ = (WORD) (fZ >= 0.0 ? fZ * 512.0 / M_PI : fZ * 512.0 / M_PI + 1024.0);

            WORD    wTRAngle1 = (wX << 4) + ((wY & 0x03C0) >> 6);
            WORD    wTRAngle2 = ((wY & 0x003F) << 10) + (wZ & 0x03FF);

            Key.vPushData(CAnimData(quat,vecTrans,wTRAngle1,wTRAngle2));
            }

        // Add the key to the track
        lpTrack->vPushKey(Key);
        }

    // Add the track to the anim manager
    lpMvb->lpGetAnimMgr()->vPushTrack(lpTrack);

    delete [] lpTransRootMesh;
    delete [] lpRotMeshes;

    // Convert the 32 bits textures to 16 bits one
    for(MATNAME2INFO::iterator theIter = lpMapMat->begin(); theIter != lpMapMat->end(); theIter++)
        {
        INFO_MAT* lpInfoMat = theIter->second;

        if(lpInfoMat->boUsed)
            {
            BYTE*   lpb16 = new BYTE [ lpInfoMat->dwWidth * lpInfoMat->dwHeight * 2 ];
            
            vConvertABGRToARGB(lpInfoMat->dwWidth,lpInfoMat->dwHeight,false,lpInfoMat->lpbData,lpb16);

            delete [] lpInfoMat->lpbData;

            lpInfoMat->lpbData = lpb16;
            }
        }

    lpMvb->vSetMeshes(dwNumNodes,lpLstMeshes,lpMapMat);

    // Compute the bounding boxes for each key
    CVect3D vecFirstTrans;

    for(dwKey = 0; dwKey < dwAnimNumKeys; dwKey++)
        {
        CBoundingBox    BoundingBox;
        CAnimKey*       lpKey = lpTrack->lpGetKey( dwKey );

        BoundingBox.vSetBB(CVect3D(1e20,1e20,1e20),CVect3D(-1e20,-1e20,-1e20));

        #define STACK_DEPTHMAX 10

        CMat4x4 StackM[ STACK_DEPTHMAX ];
        DWORD   dwStackPointer = 0;
        CMat4x4 matTransf;

        TR2_MESHTREE*   lpMeshTree  = lpMvb->lpGetMeshTree();
        TR123_MESH*     lpMesh = lpLstMeshes;
    
        // Get & transform the BB of the root mesh
        const CAnimData&    AnimData    = lpKey->GetData( 0 );
        const CVect3D&      vecTrans    = AnimData.vecGetTrans();
        const CQuaternion&  Quat        = AnimData.quatGetQuaternion();

        if( dwKey == 0 ) vecFirstTrans = vecTrans;

        matTransf = m4Translation3D( -vecTrans );

        CMat4x4 matRot;

        Quat.vToRotation(&matRot);

        matTransf = matTransf * matRot;

        CBoundingBox    BBMesh;

        lpMvb->CTrLstMsh::vComputeBoundingBox( BBMesh, 0, &matTransf );

        // Update the BB of the moveable
        BoundingBox.vEnlarge( BBMesh );

        for(DWORD dwI = 1; dwI < dwNumNodes; dwI++)
            {
            lpMesh = &lpLstMeshes[dwI];

            bitu32  dwStack = lpMeshTree->lCoord; lpMeshTree++;
            bit32   lTransX = lpMeshTree->lCoord; lpMeshTree++;
            bit32   lTransY = lpMeshTree->lCoord; lpMeshTree++;
            bit32   lTransZ = lpMeshTree->lCoord; lpMeshTree++;

            if(dwStack & 1)
                {
                if(dwStackPointer == 0)
                    {
                    AfxMessageBox("Moveable bounding box calculation: stack pointer = 0 with pop operation !",MB_OK | MB_ICONSTOP | MB_TOPMOST);
                    }
                else
                    {
                    dwStackPointer -= 1;
                    matTransf  = StackM[dwStackPointer];
                    }
                }

            if(dwStack & 2)
                {
                if(dwStackPointer + 3 > STACK_DEPTHMAX)
                    {
                    AfxMessageBox("Moveable bounding box calculation: stack overflow !",MB_OK | MB_ICONSTOP | MB_TOPMOST);
                    }
                else
                    {
                    StackM[dwStackPointer]    = matTransf;
                    dwStackPointer += 1;
                    }
                }

            const CAnimData& AnimData = lpKey->GetData( dwI );
            const CQuaternion& Quat = AnimData.quatGetQuaternion();

            CMat4x4 matRot;

            Quat.vToRotation(&matRot);

            CMat4x4 matTrans = m4Translation3D(CVect3D(-lTransX, lTransY, lTransZ));

            matTransf = matTransf * matTrans;
            matTransf = matTransf * matRot;

            CBoundingBox    BBMesh;

            lpMvb->CTrLstMsh::vComputeBoundingBox(BBMesh, dwI, &matTransf);

            BoundingBox.vEnlarge( BBMesh );
            }

        lpKey->vSetBoundingBox( BoundingBox );
        }
}


bool CTRFile::boImportMoveableFrom3DS(CTrMvb* lpMvb,TCHAR* lpsz3DSName,bool boInternalLighting)
{
    bool    boOk = false;

    // Get the path to the textures
    TCHAR szDrive[_MAX_DRIVE];
    TCHAR szDir[_MAX_DIR];
    TCHAR szPath[_MAX_PATH];

    _tsplitpath(lpsz3DSName,szDrive,szDir,NULL,NULL);

    lstrcpy(szPath,szDrive);
    lstrcat(szPath,szDir);
    
    // Read the 3DS file
    file3ds*        file= NULL;
    database3ds*    db = NULL;

    file = OpenFile3ds(lpsz3DSName, "rw");
    ON_ERROR_PRINT

    InitDatabase3ds(&db);
    ON_ERROR_PRINT

    CreateDatabase3ds(file, db);
    ON_ERROR_PRINT

    // Create the material list
    MATNAME2INFO    MapMat;

    vCreateMapMaterial(db,MapMat,szPath);

    // Import only 3ds files which have the same mesh number than the replaced moveable
    DWORD dwNumNodes = GetObjectNodeCount3ds(db);

    // Create an array giving for each mesh its parent index mesh
    namelist3ds* lpLstNames = NULL;

    GetObjectNodeNameList3ds(db,&lpLstNames);

    long*   lpLstParent;
    DWORD   dwNumRootMeshes = dwCreateParentArray(db,&lpLstParent,dwNumNodes,lpLstNames);

    ReleaseNameList3ds(&lpLstNames);

    // The TR engine allows only one root mesh, not several
    if(dwNumRootMeshes <= 1)
        {
        TR2_MESHTREE* lpNewMeshTree = NULL;

        vCreateMoveableHierarchy(&lpNewMeshTree,dwNumNodes,lpLstParent);

        delete [] lpLstParent;

        lpMvb->vSetMeshTree(lpNewMeshTree, dwNumNodes);

        vCreateMoveable(db,lpMvb,dwNumNodes,boInternalLighting,&MapMat);

        boOk = true;

        delete [] lpNewMeshTree;
        }
    else
        {
        delete [] lpLstParent;

        CString s;

        s.Format(_T("You can have only one mesh at the root of the hierarchy, not %ld !"),dwNumRootMeshes);

        AfxMessageBox(s,MB_OK | MB_ICONSTOP);
        }

    // Release the memory used by the material map
    for(MATNAME2INFO::iterator theIter = MapMat.begin(); theIter != MapMat.end(); theIter++)
        {
        INFO_MAT* lpInfoMat = theIter->second;

        if(lpInfoMat->boUsed)
            {
            for(LISTMESHMAT::iterator theIter2 = lpInfoMat->ListMeshMat.begin(); theIter2 != lpInfoMat->ListMeshMat.end(); theIter2++)
                {
                INFO_MESH_MAT* lpMeshMat = *theIter2;

                delete [] lpMeshMat->lpLstTextCoord;
                delete lpMeshMat;
                }
            }

        delete [] lpInfoMat->lpbData;
        delete lpInfoMat;
        }

    ReleaseDatabase3ds(&db);
    ON_ERROR_PRINT

    CloseFile3ds(file);
    ON_ERROR_PRINT

    CloseAllFiles3ds();

    return boOk;
}


bool    boGetAnimsData( CTrMvb& trMVB, DWORD dwAnimId, 
                        DWORD& dwNumMeshes, DWORD& dwNumFrames, DWORD& dwAnimNumKeys, DWORD& dwAnimFrameRate, 
                        CVect3D** lplpTransRootMesh, CQuaternion** lplpRotMeshes )
{
    *lplpTransRootMesh  = NULL;
    *lplpRotMeshes      = NULL;

    CAnimTrack*     lpTrack     = trMVB.lpGetAnimMgr()->lpGetTrack( dwAnimId );

    _ASSERTE( lpTrack );

    if( !lpTrack ) return false;

    dwNumMeshes         = trMVB.dwGetNumMeshes();
    dwNumFrames         = lpTrack->dwGetNumFrames();
    dwAnimNumKeys       = lpTrack->dwGetNumKeys();
    dwAnimFrameRate     = lpTrack->dwGetFrameRate();
    *lplpTransRootMesh  = new CVect3D [ dwAnimNumKeys + 1 ];
    *lplpRotMeshes      = new CQuaternion [ dwNumMeshes * (dwAnimNumKeys + 1) ];

    for( DWORD dwKey = 0; dwKey < dwAnimNumKeys; dwKey++ )
        {
        CAnimKey*   lpKey = lpTrack->lpGetKey( dwKey );

        if( lpKey->dwGetNumData() != dwNumMeshes )
            {
            _ASSERTE( !"??!!??" );
            }

        // Translation of the root mesh
        const CAnimData& Data = lpKey->GetData( 0 );

        (*lplpTransRootMesh)[ dwKey ].vSet( Data.vecGetTrans() );
        
        // Rotations
        for( DWORD dwMesh = 0; dwMesh < dwNumMeshes; dwMesh++ )
            {
            const CAnimData& Data = lpKey->GetData( dwMesh );
            (*lplpRotMeshes)[ dwKey * dwNumMeshes + dwMesh ] = Data.quatGetQuaternion();
            }
        }

    for( DWORD dwMesh = 0; dwMesh < dwNumMeshes; dwMesh++ )
        {
        // Copy the first key at the end of the anim
        (*lplpTransRootMesh)[dwAnimNumKeys] = (*lplpTransRootMesh)[0];
        (*lplpRotMeshes)[ dwAnimNumKeys * dwNumMeshes + dwMesh ] = (*lplpRotMeshes)[ 0 * dwNumMeshes + dwMesh ];

        for( DWORD dwKey = 0; dwKey < dwAnimNumKeys; dwKey++ )
            {
            CQuaternion* lpQ1 = &((*lplpRotMeshes)[ dwKey * dwNumMeshes + dwMesh ]);
            CQuaternion* lpQ2 = &((*lplpRotMeshes)[ (dwKey + 1) * dwNumMeshes + dwMesh ]);

            if( ((*lpQ1) - (*lpQ2)).fLength2() > ((*lpQ1) + (*lpQ2)).fLength2() ) *lpQ2 = -(*lpQ2);
            }
        }

    return true;
}


//
//
// Export moveable to 3DS
//
//
bool    boExportMoveableTo3DS(CTrMvb& trMVB,TCHAR* lpsz3DSName,TCHAR* lpszTextureName,CTRFile* lpTRFile,DWORD dwNumTrack)
{
    // Extract the name of the moveable from the .3DS path
    TCHAR lpszObjName[_MAX_FNAME];
    TCHAR szDrive[_MAX_DRIVE];
    TCHAR szDir[_MAX_DIR];

    _tsplitpath(lpsz3DSName,szDrive,szDir,lpszObjName,NULL);

    // Construct the path to the texture file
    TCHAR szTexturePath[MAX_PATH];
    CString s;

    _tmakepath(szTexturePath,szDrive,szDir,lpszTextureName,_T(".tga"));

    // Init the 3DS database
    database3ds* Db3ds = NULL;

    Init3DSForSave(&Db3ds);

    float fMinX = 1e20f, fMaxX = -1e20f;
    float fMinY = 1e20f, fMaxY = -1e20f;
    float fMinZ = 1e20f, fMaxZ = -1e20f;

    // Create the material for the moveable texture
    CREATE_MATERIAL_FOR_TILE()

    // Copy the meshes into the 3DS database
    bool boOk = true;

    #define STACK_DEPTHMAX 10
    bool	StackIsDummy[STACK_DEPTHMAX];
    DWORD   Stack[STACK_DEPTHMAX];
    float   StackTrans[3 * STACK_DEPTHMAX];
    float   fTransX = 0, fTransY = 0, fTransZ = 0;
    DWORD   dwStackPointer = 0;

    TR2_MESHTREE* lpMeshTree = trMVB.m_lpMeshTree;
    DWORD   dwParent = 0;
	bool	boParentIsDummy = trMVB.m_lpLstMeshes[0].boIsDummy;

    DWORD   dwNumAnimsId = 1;
    DWORD*  lpLstAnimsId = new DWORD [ dwNumAnimsId ];

    lpLstAnimsId[0] = dwNumTrack;

    DWORD dwNumMeshes     = trMVB.m_dwNumMeshes;

    DWORD   dwNumTotFrames          = 0;
    DWORD   dwNumTotKeys            = 0;
    DWORD   dwNumFramesAnimFirst    = 0;
    DWORD   dwNumFramesAnimLast     = 0;

    if( dwNumTrack != -1 )
        {
        for( DWORD dwTrack = 0; dwTrack < dwNumAnimsId; dwTrack++ )
            {
            CAnimTrack* lpTrack = trMVB.lpGetAnimMgr()->lpGetTrack( lpLstAnimsId[ dwTrack ] );

            _ASSERTE( lpTrack );

            dwNumTotFrames  += lpTrack->dwGetNumFrames();
            dwNumTotKeys    += lpTrack->dwGetNumKeys();

            if( dwTrack == 0 ) dwNumFramesAnimFirst = lpTrack->dwGetNumFrames();
            if( dwTrack == dwNumAnimsId - 1 ) dwNumFramesAnimLast = lpTrack->dwGetNumFrames();
            }

        dwNumTotKeys += dwNumAnimsId - 1;

        kfsets3ds* key = NULL;
        GetKfSets3ds(Db3ds,&key);

        key->anim.length = (dwNumAnimsId - 1) * 1000 + dwNumFramesAnimLast - 1;
        key->seg.use = True3ds;
        key->seg.begin = 0;
        key->seg.end = dwNumFramesAnimFirst - 1;

        PutKfSets3ds(Db3ds,key);

        ReleaseKfSets3ds(&key);
        }

    for(DWORD dwI = 0; dwI < dwNumMeshes; dwI++)
        {
		bool boIsDummy = trMVB.m_lpLstMeshes[dwI].boIsDummy;

		if( !boIsDummy )
			wsprintf(lpszObjName,__TEXT("M%.3d"),dwI);
		else
			wsprintf(lpszObjName,__TEXT("dummy%.3d"),dwI);

        kfmesh3ds* objMotion = NULL;

        if( dwNumTrack != -1 )
            {
            InitObjectMotion3ds(&objMotion,dwI ? 1 : dwNumTotKeys,dwNumTotKeys,1,0,0);
            }
        else
            {
            InitObjectMotion3ds(&objMotion,1,1,1,0,0);
            }

		if( !boIsDummy )
			sprintf(objMotion->name,"M%.3d",dwI);
		else
			sprintf(objMotion->name,"dummy%.3d",dwI);

        if(dwI)
            {
            bitu32  dwStack = lpMeshTree->lCoord; lpMeshTree++;
            bit32   lTransX = lpMeshTree->lCoord; lpMeshTree++;
            bit32   lTransY = lpMeshTree->lCoord; lpMeshTree++;
            bit32   lTransZ = lpMeshTree->lCoord; lpMeshTree++;

            if(dwStack & 1)
                {
                dwStackPointer--;
                dwParent = Stack[dwStackPointer];
				boParentIsDummy = StackIsDummy[ dwStackPointer ];
                fTransX = StackTrans[dwStackPointer*3 + 0];
                fTransY = StackTrans[dwStackPointer*3 + 1];
                fTransZ = StackTrans[dwStackPointer*3 + 2];
                }

            if(dwStack & 2)
                {
                StackIsDummy[dwStackPointer] = boParentIsDummy;
                Stack[dwStackPointer] = dwParent;
                StackTrans[dwStackPointer*3 + 0] = fTransX;
                StackTrans[dwStackPointer*3 + 1] = fTransY;
                StackTrans[dwStackPointer*3 + 2] = fTransZ;
                dwStackPointer++;
                }

            fTransX +=  (float)lTransX;
            fTransY +=  (float)lTransY;
            fTransZ +=  (float)lTransZ;

            if(dwParent != -1)
				{
				if( !boParentIsDummy )
					sprintf(objMotion->parent,"M%.3d",dwParent);
				else
					sprintf(objMotion->parent,"dummy%.3d",dwParent);
				}

            objMotion->pos[0].x = (float)lTransX;   // Convert TR coordinates to 3DS (file) coordinates
            objMotion->pos[0].y = (float)lTransZ;
            objMotion->pos[0].z = (float)-lTransY;
            }

        if( dwNumTrack != -1 )
            {
            DWORD           dwBaseTime  = 0;
            DWORD           dwBaseKey   = 0;

            for( DWORD dwAnimId = 0; dwAnimId < dwNumAnimsId; dwAnimId++ )
                {
                DWORD           dwDummy;
                DWORD           dwNumFrames;
                DWORD           dwAnimNumKeys;
                DWORD           dwAnimFrameRate;
                CVect3D*        lpTransRootMesh;
                CQuaternion*    lpRotMeshes;

                boGetAnimsData( trMVB, lpLstAnimsId[ dwAnimId ], dwDummy, dwNumFrames, dwAnimNumKeys, dwAnimFrameRate, &lpTransRootMesh, &lpRotMeshes );
        
                if( dwI == 0 )
                    {
                    for(DWORD dwKey = 0; dwKey < dwAnimNumKeys; dwKey++)
                        {
                        objMotion->pos[dwBaseKey + dwKey].x = lpTransRootMesh[dwKey][0];    // Convert OGL coordinates to 3DS (file) coordinates
                        objMotion->pos[dwBaseKey + dwKey].y = -lpTransRootMesh[dwKey][2];
                        objMotion->pos[dwBaseKey + dwKey].z = lpTransRootMesh[dwKey][1];
                        objMotion->pkeys[dwBaseKey + dwKey].time = dwBaseTime + dwKey * dwAnimFrameRate;
                        objMotion->pkeys[dwBaseKey + dwKey].rflags = 0x1F;
                        objMotion->pkeys[dwBaseKey + dwKey].tension = 0.0f;
                        objMotion->pkeys[dwBaseKey + dwKey].continuity = 0.0f;
                        objMotion->pkeys[dwBaseKey + dwKey].bias = 0.0f;
                        }

                    objMotion->pkeys[dwBaseKey + dwAnimNumKeys - 1].time = dwBaseTime + dwNumFrames - 1;

                    if( dwAnimId != dwNumAnimsId - 1 )
                        {
                        objMotion->pos[dwBaseKey + dwAnimNumKeys].x = 0;
                        objMotion->pos[dwBaseKey + dwAnimNumKeys].y = 0;
                        objMotion->pos[dwBaseKey + dwAnimNumKeys].z = 0;
                        objMotion->pkeys[dwBaseKey + dwAnimNumKeys].time = dwBaseTime + 999;
                        objMotion->pkeys[dwBaseKey + dwAnimNumKeys].rflags = 0x1F;
                        objMotion->pkeys[dwBaseKey + dwAnimNumKeys].tension = 0.0f;
                        objMotion->pkeys[dwBaseKey + dwAnimNumKeys].continuity = 0.0f;
                        objMotion->pkeys[dwBaseKey + dwAnimNumKeys].bias = 0.0f;
                        }
                    }

                CQuaternion QuatPrev(0.0,0.0,0.0,1.0);

                CQuaternion*    lpQuat;
                CVect3D         vecAxis;
                FLOAT           fAngle;
                CQuaternion     Quat;

                for(DWORD dwKey = 0; dwKey < dwAnimNumKeys; dwKey++)
                    {
                    lpQuat = &lpRotMeshes[dwKey * dwNumMeshes + dwI];

                    Quat = (*lpQuat) * QuatPrev.quatConjugate();
                    Quat.vToAxisAngle(&vecAxis,&fAngle);
                    vecAxis.boNormalize1();

                    QuatPrev = *lpQuat;

                    objMotion->rot[dwBaseKey + dwKey].angle = -fAngle;
                    objMotion->rot[dwBaseKey + dwKey].x = vecAxis[0];   // Convert OGL coordinates to 3DS (file) coordinates
                    objMotion->rot[dwBaseKey + dwKey].y = -vecAxis[2];
                    objMotion->rot[dwBaseKey + dwKey].z = vecAxis[1];
                    objMotion->rkeys[dwBaseKey + dwKey].time = dwBaseTime + dwKey * dwAnimFrameRate;
                    objMotion->rkeys[dwBaseKey + dwKey].rflags = 0x1F;
                    objMotion->rkeys[dwBaseKey + dwKey].tension = 0.0f;
                    objMotion->rkeys[dwBaseKey + dwKey].continuity = 0.0f;
                    objMotion->rkeys[dwBaseKey + dwKey].bias = 0.0f;
                    }

                objMotion->rkeys[dwBaseKey + dwAnimNumKeys - 1].time = dwBaseTime + dwNumFrames - 1;

                if( dwAnimId != dwNumAnimsId - 1 )
                    {
                    Quat = (*lpQuat).quatConjugate();
                    Quat.vToAxisAngle(&vecAxis,&fAngle);
                    vecAxis.boNormalize1();

                    objMotion->rot[dwBaseKey + dwAnimNumKeys].angle = -fAngle;
                    objMotion->rot[dwBaseKey + dwAnimNumKeys].x = vecAxis[0];   // Convert OGL coordinates to 3DS (file) coordinates
                    objMotion->rot[dwBaseKey + dwAnimNumKeys].y = -vecAxis[2];
                    objMotion->rot[dwBaseKey + dwAnimNumKeys].z = vecAxis[1];
                    objMotion->rkeys[dwBaseKey + dwAnimNumKeys].time = dwBaseTime + 999;
                    objMotion->rkeys[dwBaseKey + dwAnimNumKeys].rflags = 0x1F;
                    objMotion->rkeys[dwBaseKey + dwAnimNumKeys].tension = 0.0f;
                    objMotion->rkeys[dwBaseKey + dwAnimNumKeys].continuity = 0.0f;
                    objMotion->rkeys[dwBaseKey + dwAnimNumKeys].bias = 0.0f;
                    }

                dwBaseTime += 1000;
                dwBaseKey += dwAnimNumKeys + 1;

                delete [] lpTransRootMesh;
                delete [] lpRotMeshes;
                }
            }

        PutObjectMotion3ds(Db3ds,objMotion);
        ReleaseObjectMotion3ds(&objMotion);

        boOk &= boCopyTRMeshInto3DSMesh(Db3ds,lpszObjName,trMVB,&trMVB.m_lpLstMeshes[dwI],lpszTextureName,fMinX,fMaxX,fMinY,fMaxY,fMinZ,fMaxZ,fTransX,fTransY,fTransZ);

        dwParent = dwI;
		boParentIsDummy = boIsDummy;
        }

    if(boOk)
        {
        // Lights
        light3ds*       lpSlite = NULL;

        PUT_LIGHT(fMinX,fMinY,fMaxZ,0.5f,"Light1")
        PUT_LIGHT(fMaxX,fMinY,fMaxZ,0.5f,"Light2")
        PUT_LIGHT(fMaxX,fMaxY,fMaxZ,0.5f,"Light3")
        PUT_LIGHT(fMinX,fMaxY,fMaxZ,0.5f,"Light4")

        // Texture
#ifdef EXPORT_256x256_TEXTURE
        trMVB.vSaveTextureAsTGA(szTexturePath, true);
#else
        trMVB.vSaveTextureAsTGA(szTexturePath);
#endif
        }

    Save3DS(&Db3ds,lpsz3DSName);

    return true;
}
