#pragma once

#pragma pack(push,1)


typedef enum { TrObj_NoObj = 0, TrObj_Type_Mesh = 1, TrObj_Type_StaticMesh = 2, TrObj_Type_Moveable = 3, TrObj_Type_Sound = 4, TrObj_Type_AnimTexture = 5, TrObj_Type_Sprite = 6 } enumTrObjType;
typedef enum { Tr_UnknownVersion = 0, Tr1_File = 1, Tr2_File = 2, Tr3_File = 3, Tr4_File = 4 } enumTrVersion;


#define VERSION_TRMESH                  1
#define VERSION_TRSTATICMESH            1
#define VERSION_TRMOVEABLE              2
#define VERSION_TRSPRITE                1

#define __LOADERR_MISMATCHVERSION__     -1

class CTrLstMsh;
class CTrMvb;

//
// CTrObjBase
//
// -> define the base operations for any TR object
// -> handle the header of the object
//
class CTrObjBase
{
public:

    CTrObjBase() { vInitialize(); }
    CTrObjBase(BYTE bVersion, enumTrObjType trType, enumTrVersion trVersion) :
        m_bVersion(bVersion), m_trType(trType), m_trVersion(trVersion),
        m_bDummy(0), m_dwDirectory(0)
    {
    }
    CTrObjBase(BYTE* lpObj);
    virtual ~CTrObjBase() { vRelease(false); }

    virtual DWORD                   dwGetDataSize();
    virtual BYTE*                   lpbGetData(DWORD& dwDataSize);
    virtual void                    vRelease(bool boReleaseParent = true);

    bool                            boSave(CString& cstrFileName);

    BYTE                            GetVersion()    { return m_bVersion; }
    enumTrObjType                   GetTrType()     { return m_trType; }
    enumTrVersion                   GetTrVersion()  { return m_trVersion; }

    void                            vSetVersion(BYTE bVersion)          { m_bVersion = bVersion; }
    void                            vSetTRType(enumTrObjType ObjType)   { m_trType = ObjType; }
    void                            vSetTRVersion(enumTrVersion trVer)  { m_trVersion = trVer; }
    void                            vStealData(CTrObjBase& obj);

private:

    void                            vInitialize();
    void                            vConstruct(BYTE* lpObjBase);
    static BYTE*                    lpbLoad(CString& cstrFileName);

    BYTE                            m_bVersion;
    enumTrObjType                   m_trType;
    enumTrVersion                   m_trVersion;
    BYTE                            m_bDummy;
    DWORD                           m_dwDirectory;
};


//
// CTrObjGfx
//
// -> define the base class for a TR graphic object
// -> handle a texture fragments list
// -> handle a TR animated textures list
//
class CTrObjGfx : public CTrObjBase
{
public:

    CTrObjGfx(BYTE bVersion, enumTrObjType trType, enumTrVersion trVersion) : CTrObjBase(bVersion,trType,trVersion) { }
    virtual ~CTrObjGfx() { vRelease(false); }

    CTrObjGfx(CTRFile* lpTRFile, bitu16* lpAnimatedText, DWORD dwDesiredTextureWidth = 0,DWORD dwDesiredTextureHeight = 0);
    CTrObjGfx(BYTE* lpObjGfx = NULL);

    virtual DWORD                   dwGetDataSize();
    virtual BYTE*                   lpbGetData(DWORD& dwDataSize);
    virtual void                    vRelease(bool boReleaseParent = true);

    void                            vSaveTextureAsTGA(const TCHAR* lpszFilename, bool bo256x256 = false);
    bool                            boLoadTextureFromTGA(const TCHAR* lpszFilename);

    DWORD                           dwGetObjectId()         const { return m_dwObjectId; }
    DWORD                           dwGetTextureWidth()     const { return m_dwTextureWidth; }
    DWORD                           dwGetTextureHeight()    const { return m_dwTextureHeight; }
    DWORD                           dwGetTextureDepth()     const { return m_dwTextureDepth; }
    const BYTE*                     lpGetTextureData()      const { return m_lpbTextureData; }
    const TR2_OBJECT_TEXTURE*       lpGetFragText()         const { return m_lpFragText; }
    DWORD                           dwGetNumFragments()     const { return m_dwNumFragText; }

    void                            vSetObjectId(DWORD dwObjectId) { m_dwObjectId = dwObjectId; }

    void                            vStealData(CTrObjGfx& obj);

private:

    void                            vInitialize();
    void                            vConstruct(BYTE* lpObjGfx);

    friend bool boCopyTRMeshInto3DSMesh(database3ds* Db3ds,char* lpszObjName,
                    CTrLstMsh& trObj,TR123_MESH* lpMesh,char* lpszTextureName,
                    float& fMinX,float& fMaxX,float& fMinY,float& fMaxY,float& fMinZ,float& fMaxZ,
                    float fTransX, float fTransY, float fTransZ);

protected:

    void                            vConstruct(CFragment* lpLstFragments, DWORD dwNumFragments, DWORD dwDesiredTextureWidth = 0,DWORD dwDesiredTextureHeight = 0, bool boRemoveDup = true);

    DWORD                           m_dwObjectId;
    DWORD                           m_dwNumFragText;
    TR2_OBJECT_TEXTURE*             m_lpFragText;
    DWORD                           m_dwTextureWidth;
    DWORD                           m_dwTextureHeight;
    DWORD                           m_dwTextureDepth;
    BYTE*                           m_lpbTextureData;
};


//
// CTrLstSpr
//
// -> handle a list of sprites
// -> handle a TR sprite sequence
//
typedef struct
{
    bit16                   sLeftSide;
    bit16                   sTopSide;
    bit16                   sRightSide;
    bit16                   sBottomSide;
    bitu16                  wFragIndex;
} WORLD_POSITIONNING;


class CTrLstSpr : public CTrObjGfx
{
public:

    CTrLstSpr(TR2_SPRITE_SEQUENCE* lpSprSeq, CTRFile* lpTRFile);
    CTrLstSpr(BYTE* lpSpr = NULL,bool boDeleteData = false) throw(DWORD);
    virtual ~CTrLstSpr() { vRelease(false); }

    virtual DWORD                   dwGetDataSize();
    virtual BYTE*                   lpbGetData(DWORD& dwDataSize);
    virtual void                    vRelease(bool boReleaseParent = true);
    static CTrLstSpr*               objLoad(CString& cstrFileName);

    const DWORD                     dwGetNumSprites()           const { return m_dwNumFragText; }
    const WORLD_POSITIONNING*       lpGetWorldPositionning()    const { return m_lpFragWorldPos; }

    void                            vStealData(CTrLstSpr& obj);

private:

    void                            vInitialize();
    void                            vConstruct(BYTE* lpObjSpr);

protected:

    void                            vConstruct(CTRFile* lpTRFile, TR2_SPRITE_SEQUENCE* lpSprSeq);

    WORLD_POSITIONNING*             m_lpFragWorldPos;
};


//
// CTrLstMsh
//
// -> define the base class for a graphic objects made of meshes
// -> handle a list of meshes
// -> handle a TR mesh (which is a list with only one mesh)
//
class CTrLstMsh : public CTrObjGfx
{
public:

    CTrLstMsh(BYTE bVersion, enumTrObjType trType, enumTrVersion trVersion) : CTrObjGfx(bVersion,trType,trVersion) { }

    CTrLstMsh(TR123_MESH* lpMesh, CTRFile* lpTRFile, DWORD dwDesiredTextureWidth = 0, DWORD dwDesiredTextureHeight = 0);
    CTrLstMsh(TR123_MESH** lplpMeshes, DWORD dwNumMeshes, CTRFile* lpTRFile, DWORD dwDesiredTextureWidth = 0, DWORD dwDesiredTextureHeight = 0);
    CTrLstMsh(BYTE* lpMSH = NULL,bool boDeleteData = false);

    virtual ~CTrLstMsh() { vRelease(false); }

    virtual DWORD                   dwGetDataSize();
    virtual BYTE*                   lpbGetData(DWORD& dwDataSize);
    virtual void                    vRelease(bool boReleaseParent = true);
    static CTrLstMsh                objLoad(CString& cstrFileName);
    virtual void                    vComputeBoundingBox(CBoundingBox& BoundingBox,DWORD dwNumMesh = -1, CMat4x4* lpMat = NULL)
                                                                  { vComputeBoundingBox(dwNumMesh,BoundingBox,lpMat); }
    void                            vClearMesh(DWORD dwNumMesh);
    void                            vSetMeshes(DWORD dwNumMeshes, TR123_MESH* lpLstMeshes,MATNAME2INFO* lpMapMat);

    DWORD                           dwGetNumMeshes()        const { return m_dwNumMeshes; }
    const TR123_MESH*               lpGetMesh(DWORD dwNum)  const { ASSERT(dwNum < m_dwNumMeshes); return &m_lpLstMeshes[dwNum]; }

    void                            vStealData(CTrLstMsh& obj);

private:

    void                            vInitialize();
    void                            vConstruct(BYTE* lpObjMsh);

    friend bool boExportMSHTo3DS(CTrLstMsh& trMSH,DWORD dwNumMesh,TCHAR* lpsz3DSName,TCHAR* lpszTextureName);
    friend bool boCopyTRMeshInto3DSMesh(database3ds* Db3ds,char* lpszObjName,
                    CTrLstMsh& trObj,TR123_MESH* lpMesh,char* lpszTextureName,
                    float& fMinX,float& fMaxX,float& fMinY,float& fMaxY,float& fMinZ,float& fMaxZ,
                    float fTransX, float fTransY, float fTransZ);
    friend bool boExportMoveableTo3DS(CTrMvb& trMVB,TCHAR* lpsz3DSName,TCHAR* lpszTextureName,CTRFile* lpTRFile,DWORD dwNumTrack);

protected:

    void                            vComputeBoundingBox(DWORD dwNumMesh,CBoundingBox& BoundingBox, CMat4x4* lpMat = NULL);
    void                            vConstruct(CTRFile* lpTRFile,DWORD dwNumMeshes,TR123_MESH** lplpLstMeshes, DWORD dwDesiredTextureWidth = 0, DWORD dwDesiredTextureHeight = 0);
    void                            vConstruct(MATNAME2INFO* lpMapMat,DWORD dwNumMeshes,TR123_MESH* lpLstMeshes, DWORD dwDesiredTextureWidth = 0, DWORD dwDesiredTextureHeight = 0);

    DWORD                           m_dwNumMeshes;
    TR123_MESH*                     m_lpLstMeshes;
};


//
// CTrStatMsh
//
// -> handle a TR static mesh
//
#define TRSTATMSH_BB_VISIBILITY             0
#define TRSTATMSH_BB_COLLISION              1
#define TRSTATMSH_FLAGS_NONCOLLISIONABLE    0x0001
#define TRSTATMSH_FLAGS_VISIBLE             0x0002


class CTrStatMsh : public CTrLstMsh
{
public:

    CTrStatMsh(TR2_STATICMESH* lpStaticMesh, CTRFile* lpTRFile);
    CTrStatMsh(BYTE* lpSMsh = NULL,bool boDeleteData = false) throw(DWORD);
    virtual ~CTrStatMsh() { vRelease(false); }

    virtual DWORD                   dwGetDataSize();
    virtual BYTE*                   lpbGetData(DWORD& dwDataSize);
    virtual void                    vRelease(bool boReleaseParent = true);
    static CTrStatMsh*              objLoad(CString& cstrFileName);

    void                            vSetFlags(WORD wFlags)        { m_wFlags = wFlags; }
    CBoundingBox&                   GetBBVisibility()             { return m_BBVisibility; }
    CBoundingBox&                   GetBBCollision()              { return m_BBCollision; }
    WORD                            wGetFlags()             const { return m_wFlags; }

    void                            vSetBBVisibility(CBoundingBox& BB) { m_BBVisibility = BB; }
    void                            vSetBBCollision(CBoundingBox& BB)  { m_BBCollision  = BB; }
    void                            vStealData(CTrStatMsh& obj);

private:

    void                            vInitialize();
    void                            vConstruct(BYTE* lpObjSMsh);

protected:

    void                            vConstruct(CTRFile* lpTRFile, TR2_STATICMESH* lpStaticMesh);

    CBoundingBox                    m_BBVisibility;
    CBoundingBox                    m_BBCollision;
    WORD                            m_wFlags;
};


class CAnimMgr;

//
// CTrMvb
//
// -> handle a TR moveable
//
class CTrMvb : public CTrLstMsh
{
public:

    CTrMvb(TR2_MOVEABLE* lpMoveable, CTRFile* lpTRFile);
    CTrMvb(BYTE* lpMVB = NULL,bool boDeleteData = false) throw(DWORD);
    virtual ~CTrMvb() { vRelease(false); }

    virtual DWORD                   dwGetDataSize();
    virtual BYTE*                   lpbGetData(DWORD& dwDataSize);
    virtual void                    vRelease(bool boReleaseParent = true);
    void                            vRelease2(bool boReleaseParent = true, bool boReleaseMeshes = true, bool boReleaseAnimations = true);
    static CTrMvb*                  objLoad(CString& cstrFileName);
    virtual void                    vComputeBoundingBox(CBoundingBox& BoundingBox,DWORD dwNumMesh = -1, CMat4x4* lpMat = NULL);

    TR2_MESHTREE*                   lpGetMeshTree()         { return m_lpMeshTree; }
    CAnimMgr*                       lpGetAnimMgr()          { return m_lpAnimMgr; }
    CTrSndMgr*                      lpGetSndMgr()           { return m_lpSndMgr; }
    bool                            boIsAnimatedByEngine()  { return m_boAnimatedByEngine; }

    void                            vSetMeshTree(TR2_MESHTREE* lpMeshTree, DWORD dwNumMeshes = 0);
    void                            vReleaseAllAnims();

    void                            vStealData(CTrMvb& obj,bool boStealMeshes = true, bool boStealAnimations = true);

private:

    void                            vInitialize(bool boInitAnimations = true);
    void                            vConstruct(BYTE* lpObjMvb);

    friend bool boExportMoveableTo3DS(CTrMvb& trMVB,TCHAR* lpsz3DSName,TCHAR* lpszTextureName,CTRFile* lpTRFile,DWORD dwNumTrack);


protected:

    void                            vConstruct(CTRFile* lpTRFile, TR2_MOVEABLE* lpMoveable);

    TR2_MESHTREE*                   m_lpMeshTree;
    CAnimMgr*                       m_lpAnimMgr;
    CTrSndMgr*                      m_lpSndMgr;
    bool                            m_boAnimatedByEngine;
};

#pragma pack(pop)

