/**********************************************************************************************************************/
/*                                                VECTORS & MATRICES CLASSES                                          */
/* 15/12/98, M : inl                                                                                                  */
/* 14/10/99, P : Set fct added for optims                                                                             */
/* 25/11/99, M : rewriting for operators speed optims                                                                 */
/* 09/06/00, M : inline optim                                                                                         */
/**********************************************************************************************************************/

#ifndef   _CVECTMAT_INL_
#define   _CVECTMAT_INL_

/************************************************ INCLUDE *************************************************************/

#include  "CVectMat.h"

#include  <math.h>                                          // sqrt

/**********************************************************************************************************************/
/*                                                2D VECTOR                                                           */
/**********************************************************************************************************************/

/************************************************ constructors ********************************************************/
// initialize 2D vector
// 17/09/97, M
/**********************************************************************************************************************/

inline CVect2D::CVect2D()                                             // do nothing
  {}

inline CVect2D::CVect2D(const FLOAT fD)                               // x=y=fD
  { m_fV[_X_] = m_fV[_Y_] = fD; }

inline CVect2D::CVect2D(const FLOAT fX,const FLOAT fY)                // x=fX; y=fY
  { m_fV[_X_] = fX; m_fV[_Y_] = fY; }

inline CVect2D::CVect2D(const CVect2D& v2V)                           // copy
  { memcpy(m_fV,v2V.m_fV,sizeof(m_fV)); }

inline CVect2D::CVect2D(const CVect3D& v3V,const BYTE bDropAxis)      // cast v3 to v2 (drop an axis)
  { 
  switch (bDropAxis)
    { 
    case _X_ : m_fV[_X_] = v3V.m_fV[_Y_]; m_fV[_Y_] = v3V.m_fV[_Z_]; break;
    case _Y_ : m_fV[_X_] = v3V.m_fV[_X_]; m_fV[_Y_] = v3V.m_fV[_Z_]; break;
    default  : m_fV[_X_] = v3V.m_fV[_X_]; m_fV[_Y_] = v3V.m_fV[_Y_]; break; 
    } 
  }

#ifdef _DANGEROUS_CAST_
inline CVect2D::CVect2D(const CVect3D& v3V)                           // cast v3 to v2 (perspective)
  {
  FLOAT fZinv = 1.f/v3V[_Z_]; m_fV[_X_] = v3V[_X_]*fZinv; m_fV[_Y_] = v3V[_Y_]*fZinv; 
  }
#endif

/************************************************ assignment operators ************************************************/
// assign or modify 2D vector
// 17/09/97, M
/**********************************************************************************************************************/

inline CVect2D& CVect2D::operator  = (const CVect2D& v2V)             // assignment
  { memcpy(m_fV,v2V.m_fV,sizeof(m_fV)); return *this; }

inline CVect2D& CVect2D::operator += (const CVect2D& v2V)             // incrementation
  { vAdd(this,v2V); return *this; }

inline CVect2D& CVect2D::operator -= (const CVect2D& v2V)             // decrementation
  { vSub(this,v2V); return *this; }

inline CVect2D& CVect2D::operator *= (const FLOAT fK)                 // multiplication by a constant
  { vMul(this,fK);  return *this; }

inline CVect2D& CVect2D::operator *= (const CVect2D& v2V)             // term by term *
  { vMul(this,v2V); return *this; }

inline CVect2D& CVect2D::operator /= (const FLOAT fK)                 // division by a constant
  { vDiv(this,fK);  return *this; }

inline CVect2D& CVect2D::operator /= (const CVect2D& v2V)             // term by term /
  { vDiv(this,v2V); return *this; }

inline FLOAT&   CVect2D::operator [] (const DWORD dwN)                // indexing
  {
  return m_fV[dwN]; 
  }

inline const FLOAT CVect2D::operator [] (const DWORD dwN) const       // indexing
  {
  return m_fV[dwN]; 
  }

/************************************************ friends operators ***************************************************/
// 2D vector maths, test
// 17/09/97, M
/**********************************************************************************************************************/

inline CVect2D  operator  - (const CVect2D& v2V)                      // -V
  { CVect2D v2Tmp; vNeg(&v2Tmp,v2V);    return(v2Tmp); }

inline CVect2D  operator  * (const FLOAT fK,const CVect2D& v2V)       // k*V
  { CVect2D v2Tmp; vMul(&v2Tmp,v2V,fK); return(v2Tmp); }

inline CVect2D  operator  * (const CVect2D& v2V,const FLOAT fK)       // V*k
  { CVect2D v2Tmp; vMul(&v2Tmp,v2V,fK); return(v2Tmp); }

inline CVect2D  operator  / (const CVect2D& v2V,const FLOAT fK)       // V/k
  { CVect2D v2Tmp; vDiv(&v2Tmp,v2V,fK); return(v2Tmp); }

inline CVect2D  operator  - (const CVect2D& v2V1,const CVect2D& v2V2) // V1-V2
  { CVect2D v2Tmp; vSub(&v2Tmp,v2V1,v2V2); return(v2Tmp); }

inline CVect2D  operator  + (const CVect2D& v2V1,const CVect2D& v2V2) // V1+V2
  { CVect2D v2Tmp; vAdd(&v2Tmp,v2V1,v2V2); return(v2Tmp); }

inline FLOAT    operator  * (const CVect2D& v2V1,const CVect2D& v2V2) // V1*V2 (dot product)
  { return (v2V1.m_fV[_X_]*v2V2.m_fV[_X_] + v2V1.m_fV[_Y_]*v2V2.m_fV[_Y_]); }

inline FLOAT    operator  ^ (const CVect2D& v2V1,const CVect2D& v2V2) // V1^V2 (cross product)
  { return (v2V1.m_fV[_X_]*v2V2.m_fV[_Y_] - v2V1.m_fV[_Y_]*v2V2.m_fV[_X_]); }

inline bool     operator == (const CVect2D& v2V1,const CVect2D& v2V2) // V1==V2
  { return ((v2V1.m_fV[_X_] == v2V2.m_fV[_X_]) && (v2V1.m_fV[_Y_] == v2V2.m_fV[_Y_])); }

inline bool     operator != (const CVect2D& v2V1,const CVect2D& v2V2) // V1!=V2
  { return ((v2V1.m_fV[_X_] != v2V2.m_fV[_X_]) || (v2V1.m_fV[_Y_] != v2V2.m_fV[_Y_])); }

inline ostream& operator << (ostream& s,CVect2D& v2V)                 // output
  { return s << '(' << v2V[_X_] << ';' << v2V[_Y_] << ')'; }

inline ostream& operator << (ostream& s,const CVect2D& v2V)           // output
  { 
  CVect2D v2Tmp(v2V);
  return s << '(' << v2Tmp[_X_] << ';' << v2Tmp[_Y_] << ')'; 
  }

/************************************************ friends functions ***************************************************/
// 2D vector maths, utilities
// 17/09/97, M
/**********************************************************************************************************************/

inline void    vSwap   (CVect2D& v2V1,CVect2D& v2V2)                  // swap
  { CVect2D v2Tmp(v2V1); v2V1 = v2V2; v2V2 = v2Tmp; }

inline CVect2D v2Mini  (const CVect2D& v2V1,const CVect2D& v2V2)      // mini(V1,V2) (smallest vector)
  { 
  FLOAT f1 = v2V1*v2V1; FLOAT f2 = v2V2*v2V2;
  if(f1 < f2) return v2V1; else return v2V2; 
  }

inline CVect2D v2Maxi  (const CVect2D& v2V1,const CVect2D& v2V2)      // mini(V1,V2) (longest  vector) 
  { 
  FLOAT f1 = v2V1*v2V1; FLOAT f2 = v2V2*v2V2;
  if(f1 > f2) return v2V1; else return v2V2; 
  }

inline CVect2D v2Prod  (const CVect2D& v2V1,const CVect2D& v2V2)      // term by term *
  { CVect2D v2Tmp; vMul(&v2Tmp,v2V1,v2V2); return(v2Tmp); }

inline CVect2D v2Div   (const CVect2D& v2V1,const CVect2D& v2V2)      // term by term /
  { CVect2D v2Tmp; vDiv(&v2Tmp,v2V1,v2V2); return(v2Tmp); }

/************************************************ special functions ***************************************************/
// 2D vector normalize
// 17/09/97, M
/**********************************************************************************************************************/

inline FLOAT CVect2D::fLength (void) const                            // vector length d
  { return(FLOAT)sqrt(fLength2()); }

inline FLOAT CVect2D::fLength2(void) const                            // vector squared length d^2
  { return(m_fV[_X_]*m_fV[_X_] + m_fV[_Y_]*m_fV[_Y_]); }

inline bool  CVect2D::boNormalize(const FLOAT fNorm)                  // in:desired norm ; out:-
  {
  if(fNorm == 1.0f) return(boNormalize1());

  FLOAT fL = fLength();
  if(fL == 0.f) 
    {
    return(false);
    }
  *this *= fNorm/fL;
  return(true); 
  }

inline bool  CVect2D::boNormalize1(void)                              // norm to 1.0f (dedicated version)
  {
  FLOAT fL = fLength2();
  if((fL >= 1.0f-_EPS_NORMALIZE_) && (fL <= 1.0f+_EPS_NORMALIZE_))
    return(true);                                                     // already normalized

  if(fL == 0.f) 
    {
    return(false);
    }

  fL = (FLOAT)sqrt(fL);
  *this /= fL;
  return(true); 
  }

/************************************************ set  ****************************************************************/
// 2D vector assignment
// 25/11/99, M
/**********************************************************************************************************************/

inline void CVect2D::vSet(const FLOAT fD)
  { m_fV[_X_] = m_fV[_Y_] = fD; }

inline void CVect2D::vSet(const FLOAT fX,const FLOAT fY)
  { m_fV[_X_] = fX; m_fV[_Y_] = fY; }

inline void CVect2D::vSet(const CVect2D& v2V)
  { memcpy(m_fV,v2V.m_fV,sizeof(m_fV)); }

inline void CVect2D::vSet(const CVect3D& v3V,const BYTE bDropAxis)
  { 
  switch (bDropAxis)
    { 
    case _X_ : m_fV[_X_] = v3V.m_fV[_Y_]; m_fV[_Y_] = v3V.m_fV[_Z_]; break;
    case _Y_ : m_fV[_X_] = v3V.m_fV[_X_]; m_fV[_Y_] = v3V.m_fV[_Z_]; break;
    default  : m_fV[_X_] = v3V.m_fV[_X_]; m_fV[_Y_] = v3V.m_fV[_Y_]; break; 
    } 
  }

/**********************************************************************************************************************/
/*                                                3D VECTOR                                                           */
/**********************************************************************************************************************/

/************************************************ constructors ********************************************************/
// initialize 3D vector
// 17/09/97, M
/**********************************************************************************************************************/

inline CVect3D::CVect3D()                                             // do nothing
  {}

inline CVect3D::CVect3D(const FLOAT fD)                               // x=y=z=fD
  { m_fV[_X_] = m_fV[_Y_] = m_fV[_Z_] = fD; }

inline CVect3D::CVect3D(const FLOAT fX,const FLOAT fY,const FLOAT fZ) // x=fX; y=fY; z=fZ
  { m_fV[_X_] = fX; m_fV[_Y_] = fY; m_fV[_Z_] = fZ; }

inline CVect3D::CVect3D(const CVect3D& v3V)                           // copy
  { memcpy(m_fV,v3V.m_fV,sizeof(m_fV)); }

inline CVect3D::CVect3D(const CVect4D& v4V,const BYTE bDropAxis)      // cast v4 to v3 (drop an axis)
  { 
  switch (bDropAxis)
    { 
    case _X_ : m_fV[_X_] = v4V.m_fV[_Y_]; m_fV[_Y_] = v4V.m_fV[_Z_]; m_fV[_Z_] = v4V.m_fV[_W_]; break;
    case _Y_ : m_fV[_X_] = v4V.m_fV[_X_]; m_fV[_Y_] = v4V.m_fV[_Z_]; m_fV[_Z_] = v4V.m_fV[_W_]; break;
    case _Z_ : m_fV[_X_] = v4V.m_fV[_X_]; m_fV[_Y_] = v4V.m_fV[_Y_]; m_fV[_Z_] = v4V.m_fV[_W_]; break;
    default  : m_fV[_X_] = v4V.m_fV[_X_]; m_fV[_Y_] = v4V.m_fV[_Y_]; m_fV[_Z_] = v4V.m_fV[_Z_]; break; 
    } 
  }

inline CVect3D::CVect3D(const CVect2D& v2V,const FLOAT fZ)            // cast v2 to v3; z=fZ
  { memcpy(m_fV,v2V.m_fV,sizeof(v2V.m_fV)); m_fV[_Z_] = fZ; }

#ifdef _DANGEROUS_CAST_
inline CVect3D::CVect3D(const CVect4D& v4V)                           // cast v4 to v3 (perspective)
  {
  FLOAT fWinv = 1.f/v4V[_W_]; m_fV[_X_] = v4V[_X_]*fWinv; m_fV[_Y_] = v4V[_Y_]*fWinv; m_fV[_Z_] = v4V[_Z_]*fWinv;
  }

inline CVect3D::CVect3D(const CVect2D& v2V)                           // cast v2 to v3; z=1.f
  { memcpy(m_fV,v2V.m_fV,sizeof(v2V.m_fV)); m_fV[_Z_] = 1.f; }
#endif

/************************************************ assignment operators ************************************************/
// assign or modify 3D vector
// 17/09/97, M
/**********************************************************************************************************************/

inline CVect3D& CVect3D::operator  = (const CVect3D& v3V)             // assignment
  { memcpy(m_fV,v3V.m_fV,sizeof(m_fV)); return *this; }

inline CVect3D& CVect3D::operator += (const CVect3D& v3V)             // incrementation
  { vAdd(this,v3V); return *this; }

inline CVect3D& CVect3D::operator -= (const CVect3D& v3V)             // decrementation
  { vSub(this,v3V); return *this; }

inline CVect3D& CVect3D::operator *= (const FLOAT fK)                 // multiplication by a constant
  { vMul(this,fK);  return *this; }

inline CVect3D& CVect3D::operator *= (const CVect3D& v3V)             // term by term *
  { vMul(this,v3V); return *this; }

inline CVect3D& CVect3D::operator /= (const FLOAT fK)                 // division by a constant
  { vDiv(this,fK);  return *this; }

inline CVect3D& CVect3D::operator /= (const CVect3D& v3V)             // term by term /
  { vDiv(this,v3V); return *this; }

inline FLOAT&   CVect3D::operator [] (const DWORD dwN)                // indexing
  {
  return m_fV[dwN]; 
  }

inline const FLOAT CVect3D::operator [] (const DWORD dwN) const       // indexing
  {
  return m_fV[dwN]; 
  }

/************************************************ friends operators ***************************************************/
// 3D vector maths, test
// 17/09/97, M
/**********************************************************************************************************************/

inline CVect3D  operator  - (const CVect3D& v3V)                      // -V
  { CVect3D v3Tmp; vNeg(&v3Tmp,v3V);    return(v3Tmp); }

inline CVect3D  operator  * (const FLOAT fK,const CVect3D& v3V)       // k*V
  { CVect3D v3Tmp; vMul(&v3Tmp,v3V,fK); return(v3Tmp); }

inline CVect3D  operator  * (const CVect3D& v3V,const FLOAT fK)       // V*k
  { CVect3D v3Tmp; vMul(&v3Tmp,v3V,fK); return(v3Tmp); }

inline CVect3D  operator  / (const CVect3D& v3V,const FLOAT fK)       // V/k
  { CVect3D v3Tmp; vDiv(&v3Tmp,v3V,fK); return(v3Tmp); }

inline CVect3D  operator  - (const CVect3D& v3V1,const CVect3D& v3V2) // V1-V2
  { CVect3D v3Tmp; vSub(&v3Tmp,v3V1,v3V2); return(v3Tmp); }

inline CVect3D  operator  + (const CVect3D& v3V1,const CVect3D& v3V2) // V1+V2
  { CVect3D v3Tmp; vAdd(&v3Tmp,v3V1,v3V2); return(v3Tmp); }

inline FLOAT    operator  * (const CVect3D& v3V1,const CVect3D& v3V2) // V1*V2 (dot product)
  { return (v3V1.m_fV[_X_]*v3V2.m_fV[_X_] + v3V1.m_fV[_Y_]*v3V2.m_fV[_Y_] + v3V1.m_fV[_Z_]*v3V2.m_fV[_Z_]); }

inline CVect3D  operator  ^ (const CVect3D& v3V1,const CVect3D& v3V2) // V1^V2 (cross product)
  { CVect3D v3Tmp; vCrossProd(&v3Tmp,v3V1,v3V2); return(v3Tmp); }

inline bool     operator == (const CVect3D& v3V1,const CVect3D& v3V2) // V1==V2
  { return ((v3V1.m_fV[_X_] == v3V2.m_fV[_X_]) && (v3V1.m_fV[_Y_] == v3V2.m_fV[_Y_]) && (v3V1.m_fV[_Z_] == v3V2.m_fV[_Z_])); }

inline bool     operator != (const CVect3D& v3V1,const CVect3D& v3V2) // V1!=V2
  { return ((v3V1.m_fV[_X_] != v3V2.m_fV[_X_]) || (v3V1.m_fV[_Y_] != v3V2.m_fV[_Y_]) || (v3V1.m_fV[_Z_] != v3V2.m_fV[_Z_])); }

inline CVect3D  operator  * (const CMat3x3& m3M, const CVect3D& v3V)  // M3*V3
  { CVect3D v3Tmp; vMul(&v3Tmp,m3M,v3V); return(v3Tmp); }

inline CVect3D  operator  * (const CVect3D& v3V, const CMat3x3& m3M)  // V3*M3
  { CVect3D v3Tmp; vMul(&v3Tmp,v3V,m3M); return(v3Tmp); }

inline CVect4D  operator  * (const CMat4x4& m4M, const CVect3D& v3V)  // M4*V3
  { CVect4D v4Tmp; vMul(&v4Tmp,m4M,v3V); return(v4Tmp); }

inline CVect4D  operator  * (const CVect3D& v3V, const CMat4x4& m4M)  // V3*M4
  { CVect4D v4Tmp; vMul(&v4Tmp,v3V,m4M); return(v4Tmp); }

inline FLOAT    operator  * (const CVect4D& v4V, const CVect3D& v3V)  // V4*V3
  { return (v3V.m_fV[_X_]*v4V.m_fV[_X_] + v3V.m_fV[_Y_]*v4V.m_fV[_Y_] + v3V.m_fV[_Z_]*v4V.m_fV[_Z_] + v4V.m_fV[_W_]); }

inline FLOAT    operator  * (const CVect3D& v3V, const CVect4D& v4V)  // V3*V4
  { return (v3V.m_fV[_X_]*v4V.m_fV[_X_] + v3V.m_fV[_Y_]*v4V.m_fV[_Y_] + v3V.m_fV[_Z_]*v4V.m_fV[_Z_] + v4V.m_fV[_W_]); }

inline ostream& operator << (ostream& s,CVect3D& v3V)                 // output
  { return s << '(' << v3V[_X_] << ';' << v3V[_Y_] << ';' << v3V[_Z_] << ')'; }

inline ostream& operator << (ostream& s,const CVect3D& v3V)           // output
  { CVect3D v3Tmp(v3V);
    return s << '(' << v3Tmp[_X_] << ';' << v3Tmp[_Y_] << ';' << v3Tmp[_Z_] << ')'; }

/************************************************ friends functions ***************************************************/
// 3D vector maths, utilities
// 17/09/97, M
/**********************************************************************************************************************/

inline void    vSwap(CVect3D& v3V1,CVect3D& v3V2)                     // swap
  { CVect3D v3Tmp(v3V1); v3V1 = v3V2; v3V2 = v3Tmp; }

inline CVect3D v3Mini(const CVect3D& v3V1,const CVect3D& v3V2)        // mini(V1,V2) (smallest vector) 
  { 
  FLOAT f1 = v3V1*v3V1; FLOAT f2 = v3V2*v3V2;
  if(f1 < f2) return v3V1; else return v3V2; 
  }

inline CVect3D v3Maxi(const CVect3D& v3V1,const CVect3D& v3V2)        // mini(V1,V2) (longest  vector) 
  { 
  FLOAT f1 = v3V1*v3V1; FLOAT f2 = v3V2*v3V2;
  if(f1 > f2) return v3V1; else return v3V2; 
  }

inline CVect3D v3Prod(const CVect3D& v3V1,const CVect3D& v3V2)        // term by term *
  { CVect3D v3Tmp; vMul(&v3Tmp,v3V1,v3V2); return(v3Tmp); }

inline CVect3D v3Div (const CVect3D& v3V1,const CVect3D& v3V2)        // term by term /
  { CVect3D v3Tmp; vDiv(&v3Tmp,v3V1,v3V2); return(v3Tmp); }

inline CVect3D v3ModifyAxis(CVect3D& v3V,const SBYTE bX,const SBYTE bY, const SBYTE bZ)
  { 
  FLOAT fX,fY,fZ;                                                     // exchange some axis (3D vector)
  if(bX >= 0) fX = +v3V[bX]; else fX = -v3V[BYTE(-bX)];
  if(bY >= 0) fY = +v3V[bY]; else fY = -v3V[BYTE(-bY)];
  if(bZ >= 0) fZ = +v3V[bZ]; else fZ = -v3V[BYTE(-bZ)];
  return CVect3D(fX,fY,fZ); 
  }
  
/************************************************ special functions ***************************************************/
// 3D vector normalize
// 17/09/97, M
/**********************************************************************************************************************/

inline FLOAT CVect3D::fLength (void) const                  // vector length d
  { return(FLOAT)sqrt(fLength2()); }

inline FLOAT CVect3D::fLength2(void) const                  // vector squared length d^2
  { return(m_fV[_X_]*m_fV[_X_] + m_fV[_Y_]*m_fV[_Y_] + m_fV[_Z_]*m_fV[_Z_]); }

inline bool  CVect3D::boNormalize(FLOAT fNorm)              // in:desired norm ; out:OK? TRUE/FALSE
  {
  if(fNorm == 1.0f) return(boNormalize1());

  FLOAT fL = fLength();
  if(fL == 0) 
    {
    return(false);
    }
  *this *= fNorm/fL; 
  return(true); 
  }

inline bool  CVect3D::boNormalize1(void)                    // norm to 1.0f (dedicated version)
  {
  FLOAT fL = fLength2();
  if((fL >= 1.0f-_EPS_NORMALIZE_) && (fL <= 1.0f+_EPS_NORMALIZE_))
    return(true);                                           // already normalized

  if(fL == 0.f) 
    {
    return(false);
    }

  fL = (FLOAT)sqrt(fL);
  *this /= fL;
  return(true); 
  }

/************************************************ set  ****************************************************************/
// 3D vector assignment
// 14/10/99, P
/**********************************************************************************************************************/

inline void CVect3D::vSet(const FLOAT fD)
  { m_fV[_X_] = m_fV[_Y_] = m_fV[_Z_] = fD; }

inline void CVect3D::vSet(const FLOAT fX,const FLOAT fY,const FLOAT fZ)
  { m_fV[_X_] = fX; m_fV[_Y_] = fY; m_fV[_Z_] = fZ; }

inline void CVect3D::vSet(const CVect3D& v3V)
  { memcpy(m_fV,v3V.m_fV,sizeof(m_fV)); }

inline void CVect3D::vSet(const CVect4D& v4V,const BYTE bDropAxis)
  { 
  switch (bDropAxis)
    { 
    case _X_ : m_fV[_X_] = v4V.m_fV[_Y_]; m_fV[_Y_] = v4V.m_fV[_Z_]; m_fV[_Z_] = v4V.m_fV[_W_]; break;
    case _Y_ : m_fV[_X_] = v4V.m_fV[_X_]; m_fV[_Y_] = v4V.m_fV[_Z_]; m_fV[_Z_] = v4V.m_fV[_W_]; break;
    case _Z_ : m_fV[_X_] = v4V.m_fV[_X_]; m_fV[_Y_] = v4V.m_fV[_Y_]; m_fV[_Z_] = v4V.m_fV[_W_]; break;
    default  : m_fV[_X_] = v4V.m_fV[_X_]; m_fV[_Y_] = v4V.m_fV[_Y_]; m_fV[_Z_] = v4V.m_fV[_Z_]; break; 
    } 
  }

inline void CVect3D::vSet(const CVect2D& v2V,const FLOAT fZ)
  { memcpy(m_fV,v2V.m_fV,sizeof(v2V.m_fV)); m_fV[_Z_] = fZ; }

/**********************************************************************************************************************/
/*                                                4D VECTOR                                                           */
/**********************************************************************************************************************/

/************************************************ constructors ********************************************************/
// initialize 4D vector
// 17/09/97, M
/**********************************************************************************************************************/

inline CVect4D::CVect4D()                                                                 // do nothing
  {}

inline CVect4D::CVect4D(const FLOAT fD)                                                   // x=y=z=w=fD
  { m_fV[_X_] = m_fV[_Y_] = m_fV[_Z_] = m_fV[_W_] = fD; }

inline CVect4D::CVect4D(const FLOAT fX,const FLOAT fY,const FLOAT fZ,const FLOAT fW)      // x=fX; y=fY; z=fZ; w=fW
  { m_fV[_X_] = fX; m_fV[_Y_] = fY; m_fV[_Z_] = fZ; m_fV[_W_] = fW; }

inline CVect4D::CVect4D(const CVect4D& v4V)                                               // copy
  { memcpy(m_fV,v4V.m_fV,sizeof(m_fV)); }

inline CVect4D::CVect4D(const CVect3D& v3V,const FLOAT fW)                                // cast v3 to v4; w=fW
  { memcpy(m_fV,v3V.m_fV,sizeof(v3V.m_fV)); m_fV[_W_] = fW; }

#ifdef _DANGEROUS_CAST_
inline CVect4D::CVect4D(const CVect3D& v3V)                                               // cast v3 to v4; w=1.f
  { memcpy(m_fV,v4V.m_fV,sizeof(m_fV)); m_fV[_W_] = 1.f; }
#endif

/************************************************ assignment operators ************************************************/
// assign or modify 4D vector
// 17/09/97, M
/**********************************************************************************************************************/

inline CVect4D& CVect4D::operator  = (const CVect4D& v4V)             // assignment
  { memcpy(m_fV,v4V.m_fV,sizeof(m_fV)); return *this; }

inline CVect4D& CVect4D::operator += (const CVect4D& v4V)             // incrementation
  { vAdd(this,v4V); return *this; }

inline CVect4D& CVect4D::operator -= (const CVect4D& v4V)             // decrementation
  { vSub(this,v4V); return *this; }

inline CVect4D& CVect4D::operator *= (const FLOAT fK)                 // multiplication by a constant
  { vMul(this,fK);  return *this; }

inline CVect4D& CVect4D::operator *= (const CVect4D& v4V)             // term by term *
  { vMul(this,v4V); return *this; }

inline CVect4D& CVect4D::operator /= (const FLOAT fK)                 // division by a constant
  { vDiv(this,fK);  return *this; }

inline CVect4D& CVect4D::operator /= (const CVect4D& v4V)             // term by term /
  { vDiv(this,v4V); return *this; }

inline FLOAT&   CVect4D::operator [] (const DWORD dwN)                // indexing
  {
  return m_fV[dwN]; 
  }

inline const FLOAT CVect4D::operator [] (const DWORD dwN) const       // indexing
  {
  return m_fV[dwN]; 
  }

/************************************************ friends operators ***************************************************/
// 4D vector maths, test, swap/min/max/prod
// 17/09/97, M
/**********************************************************************************************************************/

inline CVect4D  operator  - (const CVect4D& v4V)                      // -V
  { CVect4D v4Tmp; vNeg(&v4Tmp,v4V);    return(v4Tmp); }

inline CVect4D  operator  * (const FLOAT fK,const CVect4D& v4V)       // k*V
  { CVect4D v4Tmp; vMul(&v4Tmp,v4V,fK); return(v4Tmp); }

inline CVect4D  operator  * (const CVect4D& v4V,const FLOAT fK)       // V*k
  { CVect4D v4Tmp; vMul(&v4Tmp,v4V,fK); return(v4Tmp); }

inline CVect4D  operator  / (const CVect4D& v4V,const FLOAT fK)       // V/k
  { CVect4D v4Tmp; vDiv(&v4Tmp,v4V,fK); return(v4Tmp); }

inline CVect4D  operator  - (const CVect4D& v4V1,const CVect4D& v4V2) // V1-V2
  { CVect4D v4Tmp; vSub(&v4Tmp,v4V1,v4V2); return(v4Tmp); }

inline CVect4D  operator  + (const CVect4D& v4V1,const CVect4D& v4V2) // V1+V2
  { CVect4D v4Tmp; vAdd(&v4Tmp,v4V1,v4V2); return(v4Tmp); }

inline FLOAT    operator  * (const CVect4D& v4V1,const CVect4D& v4V2) // V1*V2 (dot product)
  { return (v4V1.m_fV[_X_]*v4V2.m_fV[_X_] + v4V1.m_fV[_Y_]*v4V2.m_fV[_Y_] + v4V1.m_fV[_Z_]*v4V2.m_fV[_Z_] + v4V1.m_fV[_W_]*v4V2.m_fV[_W_]); }

inline bool     operator == (const CVect4D& v4V1,const CVect4D& v4V2) // V1==V2
  { return ((v4V1.m_fV[_X_] == v4V2.m_fV[_X_]) && (v4V1.m_fV[_Y_] == v4V2.m_fV[_Y_]) && 
            (v4V1.m_fV[_Z_] == v4V2.m_fV[_Z_]) && (v4V1.m_fV[_W_] == v4V2.m_fV[_W_])); }

inline bool     operator != (const CVect4D& v4V1,const CVect4D& v4V2) // V1!=V2
  { return ((v4V1.m_fV[_X_] != v4V2.m_fV[_X_]) || (v4V1.m_fV[_Y_] != v4V2.m_fV[_Y_]) || 
            (v4V1.m_fV[_Z_] != v4V2.m_fV[_Z_]) || (v4V1.m_fV[_W_] != v4V2.m_fV[_W_])); }

inline CVect4D  operator  * (const CMat4x4& m4M, const CVect4D& v4V)  // M4*V4
  { CVect4D v4Tmp; vMul(&v4Tmp,m4M,v4V); return(v4Tmp); }

inline CVect4D  operator  * (const CVect4D& v4V, const CMat4x4& m4M)  // V4*M4
  { CVect4D v4Tmp; vMul(&v4Tmp,v4V,m4M); return(v4Tmp); }

inline ostream& operator << (ostream& s,CVect4D& v4V)                 // output
  { return s << '(' << v4V[_X_] << ';' << v4V[_Y_] << ';' << v4V[_Z_] << ';' << v4V[_W_] << ')'; }

inline ostream& operator << (ostream& s,const CVect4D& v4V)           // output
  { CVect4D v4Tmp(v4V);
    return s << '(' << v4Tmp[_X_] << ';' << v4Tmp[_Y_] << ';' << v4Tmp[_Z_] << ';' << v4Tmp[_W_] << ')'; }

/************************************************ friends functions ***************************************************/
// 4D vector maths, utilities
// 17/09/97, M
/**********************************************************************************************************************/

inline void    vSwap (CVect4D& v4V1,CVect4D& v4V2)                    // swap
  { CVect4D v4Tmp(v4V1); v4V1 = v4V2; v4V2 = v4Tmp; }

inline CVect4D v4Mini(const CVect4D& v4V1,const CVect4D& v4V2)        // mini(V1,V2) (smallest vector) 
  { 
  FLOAT f1 = v4V1*v4V1; FLOAT f2 = v4V2*v4V2;
  if(f1 < f2) return v4V1; else return v4V2; 
  }

inline CVect4D v4Maxi(const CVect4D& v4V1,const CVect4D& v4V2)        // mini(V1,V2) (longest  vector) 
  { 
  FLOAT f1 = v4V1*v4V1; FLOAT f2 = v4V2*v4V2;
  if(f1 > f2) return v4V1; else return v4V2; 
  }

inline CVect4D v4Prod(const CVect4D& v4V1,const CVect4D& v4V2)        // term by term *
  { CVect4D v4Tmp; vMul(&v4Tmp,v4V1,v4V2); return(v4Tmp); }

inline CVect4D v4Div (const CVect4D& v4V1,const CVect4D& v4V2)        // term by term /
  { CVect4D v4Tmp; vDiv(&v4Tmp,v4V1,v4V2); return(v4Tmp); }

/************************************************ special functions ***************************************************/
// 4D vector normalize
// 17/09/97, M
/**********************************************************************************************************************/

inline FLOAT CVect4D::fLength (void) const                            // vector length d
  { return(FLOAT)sqrt(fLength2()); }

inline FLOAT CVect4D::fLength2(void) const                            // vector squared length d^2
  { return(m_fV[_X_]*m_fV[_X_] + m_fV[_Y_]*m_fV[_Y_] + m_fV[_Z_]*m_fV[_Z_] + m_fV[_W_]*m_fV[_W_]); }

inline bool  CVect4D::boNormalize(FLOAT fNorm)                        // in:desired norm ; out:OK?
  {
  if(fNorm == 1.0f) return(boNormalize1());

  FLOAT fL = fLength();
  if(fL == 0) 
    {
    return(false);
    }
  *this *= fNorm/fL;
  return(true); 
  }

inline bool  CVect4D::boNormalize1(void)                              // norm to 1.0f (dedicated version)
  {
  FLOAT fL = fLength2();
  if((fL >= 1.0f-_EPS_NORMALIZE_) && (fL <= 1.0f+_EPS_NORMALIZE_))
    return(true);                                                     // already normalized

  if(fL == 0.f) 
    {
    return(false);
    }

  fL = (FLOAT)sqrt(fL);
  *this /= fL;
  return(true); 
  }

/************************************************ set  ****************************************************************/
// 4D vector assignment
// 14/10/99, P
/**********************************************************************************************************************/

inline void CVect4D::vSet(const FLOAT fD)
  { m_fV[_X_] = m_fV[_Y_] = m_fV[_Z_] = m_fV[_W_] = fD; }

inline void CVect4D::vSet(const FLOAT fX,const FLOAT fY,const FLOAT fZ,const FLOAT fW)
  { m_fV[_X_] = fX; m_fV[_Y_] = fY; m_fV[_Z_] = fZ; m_fV[_W_] = fW; }

inline void CVect4D::vSet(const CVect4D& v4V)
  { memcpy(m_fV,v4V.m_fV,sizeof(m_fV)); }

inline void CVect4D::vSet(const CVect3D& v3V,const FLOAT fW)
  { memcpy(m_fV,v3V.m_fV,sizeof(v3V.m_fV)); m_fV[_W_] = fW; }

/**********************************************************************************************************************/
/*                                                3*3 MATRIX                                                          */
/**********************************************************************************************************************/

/************************************************ constructors ********************************************************/
// initialize 3*3 matrix
// 17/09/97, M
/**********************************************************************************************************************/

inline CMat3x3::CMat3x3()                                                                           // do nothing
  {}

inline CMat3x3::CMat3x3(const FLOAT fD)                                                             // all values=fD
  { m_v3V[_X_] = m_v3V[_Y_] = m_v3V[_Z_] = CVect3D(fD); }

inline CMat3x3::CMat3x3(const CVect3D& m_v3V0,const CVect3D& m_v3V1,const CVect3D& m_v3V2)          // 3 lines of the matrix
  { m_v3V[_X_] = m_v3V0; m_v3V[_Y_] = m_v3V1; m_v3V[_Z_] = m_v3V2; }

inline CMat3x3::CMat3x3(const CMat3x3& m3M)                                                         // copy
  { memcpy(m_v3V,m3M.m_v3V,sizeof(m_v3V)); }

#ifdef _DANGEROUS_CAST_
inline CMat3x3::CMat3x3(const CMat4x4& m4M)                                                         // cast m4 to m3 (rotation part)
  { m_v3V[_X_].vSet(m4M.m_v4V[_X_],_W_); m_v3V[_Y_].vSet(m4M.m_v4V[_Y_],_W_); m_v3V[_Z_].vSet(m4M.m_v4V[_Z_],_W_); }
#endif

/************************************************ assignment operators ************************************************/
// assign or modify 3*3 matrix
// 17/09/97, M
/**********************************************************************************************************************/

inline CMat3x3& CMat3x3::operator  = (const CMat3x3& m3M)             // assignment
  { memcpy(m_v3V,m3M.m_v3V,sizeof(m_v3V)); return *this; }

inline CMat3x3& CMat3x3::operator += (const CMat3x3& m3M)             // incrementation
  { vAdd(this,m3M); return *this; }

inline CMat3x3& CMat3x3::operator -= (const CMat3x3& m3M)             // decrementation
  { vSub(this,m3M); return *this; }

inline CMat3x3& CMat3x3::operator *= (const FLOAT fK)                 // multiplication by a constant
  { vMul(this,fK);  return *this; }

inline CMat3x3& CMat3x3::operator *= (const CMat3x3& m3M)             // multiplication
  { vMul(this,m3M); return *this; }

inline CMat3x3& CMat3x3::operator /= (const FLOAT fK)                 // division by a constant
  { vDiv(this,fK);  return *this; }

inline CMat3x3& CMat3x3::operator /= (const CMat3x3& m3M)             // multiplication by inverse
  { vDiv(this,m3M); return *this; }

inline CVect3D& CMat3x3::operator [] (const DWORD dwN)                // indexing line
  {
  return m_v3V[dwN]; 
  }

inline const CVect3D CMat3x3::operator [] (const DWORD dwN) const     // indexing line
  {
  return m_v3V[dwN]; 
  }

/************************************************ friends operators ***************************************************/
// 3*3 matrix maths, test, swap
// 17/09/97, M
/**********************************************************************************************************************/

inline CMat3x3  operator  - (const CMat3x3& m3M)                      // -V
  { CMat3x3 m3Tmp; vNeg(&m3Tmp,m3M);    return(m3Tmp); }

inline CMat3x3  operator  * (const FLOAT fK,const CMat3x3& m3M)       // k*M
  { CMat3x3 m3Tmp; vMul(&m3Tmp,m3M,fK); return(m3Tmp); }

inline CMat3x3  operator  * (const CMat3x3& m3M,const FLOAT fK)       // M*k
  { CMat3x3 m3Tmp; vMul(&m3Tmp,m3M,fK); return(m3Tmp); }

inline CMat3x3  operator  / (const CMat3x3& m3M,const FLOAT fK)       // M/k
  { CMat3x3 m3Tmp; vDiv(&m3Tmp,m3M,fK); return(m3Tmp); }

inline CMat3x3  operator  - (const CMat3x3& m3M1,const CMat3x3& m3M2) // M1-M2
  { CMat3x3 m3Tmp; vSub(&m3Tmp,m3M1,m3M2); return(m3Tmp); }

inline CMat3x3  operator  + (const CMat3x3& m3M1,const CMat3x3& m3M2) // M1+M2
  { CMat3x3 m3Tmp; vAdd(&m3Tmp,m3M1,m3M2); return(m3Tmp); }

inline CMat3x3  operator  * (const CMat3x3& m3M1,const CMat3x3& m3M2) // M1*M2
  { CMat3x3 m3Tmp; vMul(&m3Tmp,m3M1,m3M2); return(m3Tmp); }

inline bool     operator == (const CMat3x3& m3M1,const CMat3x3& m3M2) // M1==M2
  { return ((m3M1.m_v3V[_X_] == m3M2.m_v3V[_X_]) && (m3M1.m_v3V[_Y_] == m3M2.m_v3V[_Y_]) && 
            (m3M1.m_v3V[_Z_] == m3M2.m_v3V[_Z_])); }

inline bool     operator != (const CMat3x3& m3M1,const CMat3x3& m3M2) // M1!=M2
  { return ((m3M1.m_v3V[_X_] != m3M2.m_v3V[_X_]) || (m3M1.m_v3V[_Y_] != m3M2.m_v3V[_Y_]) || 
            (m3M1.m_v3V[_Z_] != m3M2.m_v3V[_Z_])); }

inline ostream& operator << (ostream& s,CMat3x3& m3M)                 // output
  { return s << '\n' << m3M[_X_] << '\n' << m3M[_Y_] << '\n' << m3M[_Z_] << '\n'; }

inline ostream& operator << (ostream& s,const CMat3x3& m3M)           // output
  { CMat3x3 m3Tmp(m3M);
    return s << '\n' << m3Tmp[_X_] << '\n' << m3Tmp[_Y_] << '\n' << m3Tmp[_Z_] << '\n'; }

/************************************************ friends functions ***************************************************/
// 3*3 matrix maths, utilities
// 17/09/97, M
/**********************************************************************************************************************/

inline void vSwap(CMat3x3& m3M1,CMat3x3& m3M2)                        // swap
  { CMat3x3 m3Tmp(m3M1); m3M1 = m3M2; m3M2 = m3Tmp; }

inline CMat3x3 m3Prod(const CMat3x3& m3M1,const CMat3x3& m3M2)        // term by term *
  { CMat3x3 m3Tmp; vMul(&m3Tmp,m3M1,m3M2); return(m3Tmp); }

inline CMat3x3 m3Div (const CMat3x3& m3M1,const CMat3x3& m3M2)        // term by term /
  { CMat3x3 m3Tmp; vDiv(&m3Tmp,m3M1,m3M2); return(m3Tmp); }

inline CMat3x3 m3MxtM(const CMat3x3& m3M1,const CMat3x3& m3M2)        // M1*tM2
  { CMat3x3 m3Tmp; vMulMxtM(&m3Tmp,m3M1,m3M2); return(m3Tmp); }

inline CMat3x3 m3tMxM(const CMat3x3& m3M1,const CMat3x3& m3M2)        // tM1*M2
  { CMat3x3 m3Tmp; vMultMxM(&m3Tmp,m3M1,m3M2); return(m3Tmp); }

/************************************************ special functions ***************************************************/
// 3*3 matrix transpose,inverse
// 17/09/97, M
/**********************************************************************************************************************/

inline CMat3x3 CMat3x3::m3Transpose() const                           // transpose
  { CMat3x3 m3Tmp; vTranspose(&m3Tmp,*this); return(m3Tmp); }

inline void CMat3x3::vIdentity3D()                                    // 3D identity
  { 
  memset(&((*this)[0][0]),0,3*3*sizeof(FLOAT));
  (*this)[0][0] = (*this)[1][1] = (*this)[2][2] = 1.f; 
  }

/**********************************************************************************************************************/
/*                                                4*4 MATRIX                                                          */
/**********************************************************************************************************************/

/************************************************ constructors ********************************************************/
// initialize 4*4 matrix
// 17/09/97, M
/**********************************************************************************************************************/

inline CMat4x4::CMat4x4()                                                                                               // do nothing
  {}

inline CMat4x4::CMat4x4(const FLOAT fD)                                                                                 // all values=fD
  { m_v4V[_X_] = m_v4V[_Y_] = m_v4V[_Z_] = m_v4V[_W_] = CVect4D(fD); }

inline CMat4x4::CMat4x4(const CVect4D& m_v4V0,const CVect4D& m_v4V1,const CVect4D& m_v4V2,const CVect4D& m_v4V3)        // 4 lines of the matrix
  { m_v4V[_X_] = m_v4V0; m_v4V[_Y_] = m_v4V1; m_v4V[_Z_] = m_v4V2; m_v4V[_W_] = m_v4V3; }

inline CMat4x4::CMat4x4(const CMat4x4& m4M)                                                                             // copy
  { memcpy(m_v4V,m4M.m_v4V,sizeof(m_v4V)); }

#ifdef _DANGEROUS_CAST_
inline CMat4x4::CMat4x4(const CMat3x3& m3M)                                                                             // cast m3 to m4
  { 
  m_v4V[_X_].vSet(m3M.m_v3V[_X_],0.f); m_v4V[_Y_].vSet(m3M.m_v3V[_Y_],0.f); 
  m_v4V[_Z_].vSet(m3M.m_v3V[_Z_],0.f); m_v4V[_W_].vSet(0.f,0.f,0.f,1.f); 
  }
#endif

/************************************************ assignment operators ************************************************/
// assign or modify 4*4 matrix
// 17/09/97, M
/**********************************************************************************************************************/

inline CMat4x4& CMat4x4::operator  = (const CMat4x4& m4M)             // assignment
  { memcpy(m_v4V,m4M.m_v4V,sizeof(m_v4V)); return *this; }

inline CMat4x4& CMat4x4::operator += (const CMat4x4& m4M)             // incrementation
  { vAdd(this,m4M); return *this; }

inline CMat4x4& CMat4x4::operator -= (const CMat4x4& m4M)             // decrementation
  { vSub(this,m4M); return *this; }

inline CMat4x4& CMat4x4::operator *= (const FLOAT fK)                 // multiplication by a constant
  { vMul(this,fK);  return *this; }

inline CMat4x4& CMat4x4::operator *= (const CMat4x4& m4M)             // multiplication
  { vMul(this,m4M); return *this; }

inline CMat4x4& CMat4x4::operator /= (const FLOAT fK)                 // division by a constant
  { vDiv(this,fK);  return *this; }

inline CMat4x4& CMat4x4::operator /= (const CMat4x4& m4M)             // multiplication by inverse
  { vDiv(this,m4M); return *this; }

inline CVect4D& CMat4x4::operator [] (const DWORD dwN)                // indexing line
  {
  return m_v4V[dwN]; 
  }

inline const CVect4D CMat4x4::operator [] (const DWORD dwN) const     // indexing line
  {
  return m_v4V[dwN]; 
  }

/************************************************ friends operators ***************************************************/
// 4*4 matrix maths, test, swap
// 17/09/97, M
/**********************************************************************************************************************/

inline CMat4x4 operator  - (const CMat4x4& m4M)                       // -V
  { CMat4x4 m4Tmp; vNeg(&m4Tmp,m4M);    return(m4Tmp); }

inline CMat4x4 operator  * (const FLOAT fK,const CMat4x4& m4M)        // k*M
  { CMat4x4 m4Tmp; vMul(&m4Tmp,m4M,fK); return(m4Tmp); }

inline CMat4x4 operator  * (const CMat4x4& m4M,const FLOAT fK)        // M*k
  { CMat4x4 m4Tmp; vMul(&m4Tmp,m4M,fK); return(m4Tmp); }

inline CMat4x4 operator  / (const CMat4x4& m4M,const FLOAT fK)        // M/k
  { CMat4x4 m4Tmp; vDiv(&m4Tmp,m4M,fK); return(m4Tmp); }

inline CMat4x4 operator  - (const CMat4x4& m4M1,const CMat4x4& m4M2)  // M1-M2
  { CMat4x4 m4Tmp; vSub(&m4Tmp,m4M1,m4M2); return(m4Tmp); }

inline CMat4x4 operator  + (const CMat4x4& m4M1,const CMat4x4& m4M2)  // M1+M2
  { CMat4x4 m4Tmp; vAdd(&m4Tmp,m4M1,m4M2); return(m4Tmp); }

inline CMat4x4 operator  * (const CMat4x4& m4M1,const CMat4x4& m4M2)  // M1*M2
  { CMat4x4 m4Tmp; vMul(&m4Tmp,m4M1,m4M2); return(m4Tmp); }

inline bool     operator == (const CMat4x4& m4M1,const CMat4x4& m4M2) // M1==M2
  { return ((m4M1.m_v4V[_X_] == m4M2.m_v4V[_X_]) && (m4M1.m_v4V[_Y_] == m4M2.m_v4V[_Y_]) && 
            (m4M1.m_v4V[_Z_] == m4M2.m_v4V[_Z_]) && (m4M1.m_v4V[_W_] == m4M2.m_v4V[_W_])); }

inline bool     operator != (const CMat4x4& m4M1,const CMat4x4& m4M2) // M1!=M2
  { return ((m4M1.m_v4V[_X_] != m4M2.m_v4V[_X_]) || (m4M1.m_v4V[_Y_] != m4M2.m_v4V[_Y_]) || 
            (m4M1.m_v4V[_Z_] != m4M2.m_v4V[_Z_]) || (m4M1.m_v4V[_W_] != m4M2.m_v4V[_W_])); }

inline ostream& operator << (ostream& s,CMat4x4& m4M)                 // output
  { return s << '\n' << m4M[_X_] << '\n' << m4M[_Y_] << '\n' << m4M[_Z_] << '\n' << m4M[_W_] << '\n'; }

inline ostream& operator << (ostream& s,const CMat4x4& m4M)           // output
  { 
  CMat4x4 m4Tmp(m4M);
  return s << '\n' << m4Tmp[_X_] << '\n' << m4Tmp[_Y_] << '\n' << m4Tmp[_Z_] << '\n' << m4Tmp[_W_] << '\n'; 
  }

/************************************************ friends functions ***************************************************/
// 4*4 matrix maths, utilities
// 17/09/97, M
/**********************************************************************************************************************/

inline void vSwap(CMat4x4& m4M1,CMat4x4& m4M2)                        // swap
  { CMat4x4 m4Tmp(m4M1); m4M1 = m4M2; m4M2 = m4Tmp; }

inline CMat4x4 m4Prod(const CMat4x4& m4M1,const CMat4x4& m4M2)        // term by term *
  { CMat4x4 m4Tmp; vMul(&m4Tmp,m4M1,m4M2); return(m4Tmp); }

inline CMat4x4 m4Div (const CMat4x4& m4M1,const CMat4x4& m4M2)        // term by term /
  { CMat4x4 m4Tmp; vDiv(&m4Tmp,m4M1,m4M2); return(m4Tmp); }

inline CMat4x4 m4MxtM(const CMat4x4& m4M1,const CMat4x4& m4M2)        // M1*tM2
  { CMat4x4 m4Tmp; vMulMxtM(&m4Tmp,m4M1,m4M2); return(m4Tmp); }

inline CMat4x4 m4tMxM(const CMat4x4& m4M1,const CMat4x4& m4M2)        // tM1*M2
  { CMat4x4 m4Tmp; vMultMxM(&m4Tmp,m4M1,m4M2); return(m4Tmp); }

/************************************************ special functions ***************************************************/
// 4*4 matrix transpose,inverse
// 17/09/97, M
/**********************************************************************************************************************/

inline CMat4x4 CMat4x4::m4Transpose() const                           // transpose
  { CMat4x4 m4Tmp; vTranspose(&m4Tmp,*this); return(m4Tmp); }

inline void CMat4x4::vIdentity4D()                                    // 4D identity
  { 
  memset(&((*this)[0][0]),0,4*4*sizeof(FLOAT));
  (*this)[0][0] = (*this)[1][1] = (*this)[2][2] = (*this)[3][3] = 1.f; 
  }

/**********************************************************************************************************************/
/*                                                TRANSFORMATIONS                                                     */
/**********************************************************************************************************************/

/************************************************ basic transformations ***********************************************/
// identity
// 18/09/97, M
/**********************************************************************************************************************/

inline CMat3x3 m3Identity3D()                                      // 3D identity 
  { CMat3x3 m3Tmp; m3Tmp.vIdentity3D(); return(m3Tmp); }

inline CMat4x4 m4Identity4D()                                      // 4D identity
  { CMat4x4 m4Tmp; m4Tmp.vIdentity4D(); return(m4Tmp); }

/**********************************************************************************************************************/

#endif
