/**********************************************************************************************************************/
/*                                                QUATERNION CLASS                                                    */
/* 26/08/00, M : creation of the file                                                                                 */
/**********************************************************************************************************************/

/************************************************ INCLUDE *************************************************************/

#include  "stdafx.h"
#include  "Quaternion.inl"

#include  <math.h>
#include  "CVectMat.inl"

/**********************************************************************************************************************/
/*                                                FRIEND FUNCTIONS                                                    */
/**********************************************************************************************************************/

void vAdd(CQuaternion* lpquatQ, const CQuaternion& quatQ1)            // Q+=Q1
  {
  lpquatQ->m_v3Vector += quatQ1.m_v3Vector;
  lpquatQ->m_fScalar  += quatQ1.m_fScalar;
  }

void vSub(CQuaternion* lpquatQ, const CQuaternion& quatQ1)            // Q-=Q1
  {
  lpquatQ->m_v3Vector -= quatQ1.m_v3Vector;
  lpquatQ->m_fScalar  -= quatQ1.m_fScalar;
  }

void vMul(CQuaternion* lpquatQ, const CQuaternion& quatQ1)            // Q*=Q1
  {
  CVect3D v3Vector   ((lpquatQ->m_v3Vector^quatQ1.m_v3Vector) + 
                       lpquatQ->m_v3Vector*quatQ1.m_fScalar   + lpquatQ->m_fScalar *quatQ1.m_v3Vector);
  lpquatQ->m_fScalar = lpquatQ->m_fScalar *quatQ1.m_fScalar   - lpquatQ->m_v3Vector*quatQ1.m_v3Vector;
  lpquatQ->vSetVector(v3Vector);
  }

void vMul(CQuaternion* lpquatQ, const FLOAT fK)                       // Q*=k
  {
  lpquatQ->m_v3Vector *= fK;
  lpquatQ->m_fScalar  *= fK;
  }

void vDiv(CQuaternion* lpquatQ, const FLOAT fK)                       // Q/=k
  {
  FLOAT fKinv = 1.f/fK;
  lpquatQ->m_v3Vector *= fKinv;
  lpquatQ->m_fScalar  *= fKinv;
  }

void vNeg(CQuaternion* lpquatQ, const CQuaternion& quatQ1)            // Q=-Q1
  {
  lpquatQ->m_v3Vector = -quatQ1.m_v3Vector;
  lpquatQ->m_fScalar  = -quatQ1.m_fScalar;
  }

void vMul(CQuaternion* lpquatQ, const CQuaternion& quatQ1,const FLOAT fK)
  {                                                                   // Q=Q1*k
  lpquatQ->m_v3Vector = quatQ1.m_v3Vector*fK;
  lpquatQ->m_fScalar  = quatQ1.m_fScalar *fK;
  }

void vDiv(CQuaternion* lpquatQ, const CQuaternion& quatQ1,const FLOAT fK)
  {                                                                   // Q=Q1/k
  FLOAT fKinv = 1.f/fK;
  lpquatQ->m_v3Vector = quatQ1.m_v3Vector*fKinv;
  lpquatQ->m_fScalar  = quatQ1.m_fScalar *fKinv;
  }

void vAdd(CQuaternion* lpquatQ, const CQuaternion& quatQ1,const CQuaternion& quatQ2)
  {                                                                   // Q=Q1+Q2
  lpquatQ->m_v3Vector = quatQ1.m_v3Vector+quatQ2.m_v3Vector;
  lpquatQ->m_fScalar  = quatQ1.m_fScalar +quatQ2.m_fScalar;
  }

void vSub(CQuaternion* lpquatQ, const CQuaternion& quatQ1,const CQuaternion& quatQ2)
  {                                                                   // Q=Q1-Q2
  lpquatQ->m_v3Vector = quatQ1.m_v3Vector-quatQ2.m_v3Vector;
  lpquatQ->m_fScalar  = quatQ1.m_fScalar -quatQ2.m_fScalar;
  }

void vMul(CQuaternion* lpquatQ, const CQuaternion& quatQ1,const CQuaternion& quatQ2)
  {                                                                   // Q=Q1*Q2
  lpquatQ->m_v3Vector = (quatQ1.m_v3Vector^quatQ2.m_v3Vector) + quatQ1.m_v3Vector*quatQ2.m_fScalar + quatQ1.m_fScalar*quatQ2.m_v3Vector;
  lpquatQ->m_fScalar  =  quatQ1.m_fScalar *quatQ2.m_fScalar   - quatQ1.m_v3Vector*quatQ2.m_v3Vector;
  }

/**********************************************************************************************************************/
/*                                                SPECIAL FUNCTIONS                                                   */
/**********************************************************************************************************************/

/************************************************ quatSlerp ***********************************************************/
// spherical linear interpolation [AART]
// 27/08/00, M
// in : quaternions to interpolate,parameter
// out: intermediate quaternion
/**********************************************************************************************************************/

CQuaternion quatSlerp(const CQuaternion& quatQ1,const CQuaternion& quatQ2,const FLOAT fT)
  {
  FLOAT fOmega,fCosom,fSinom,fSclp,fSclq;
  fCosom = quatQ1.m_v3Vector*quatQ2.m_v3Vector+quatQ1.m_fScalar*quatQ2.m_fScalar;

  if((1.f+fCosom) > 1e-5)
    {
    if((1.f-fCosom) > 1e-5)
      {
      fOmega = (FLOAT)acos(fCosom);
      fSinom = (FLOAT)sin (fOmega);
      fSclp  = (FLOAT)sin((1.f-fT)*fOmega)/fSinom;
      fSclq  = (FLOAT)sin(     fT *fOmega)/fSinom;
      }

    else
      {
      fSclp = 1.f-fT;
      fSclq = fT;
      }

    return(fSclp*quatQ1+fSclq*quatQ2);
    }

  else
    {
    CQuaternion quatR(-quatQ1.m_v3Vector[_Y_],+quatQ1.m_v3Vector[_X_],-quatQ1.m_fScalar,+quatQ1.m_v3Vector[_Z_]);
    fSclp = (FLOAT)sin((1.f-fT)*_PI_*0.5f);
    fSclq = (FLOAT)sin(     fT *_PI_*0.5f);
    
    quatR.m_v3Vector = fSclp*quatQ1.m_v3Vector + fSclq*quatR.m_v3Vector;
    return quatR;
    }
  }

/************************************************ quatConjugate *******************************************************/
// get quaternion conjugate
// 03/10/00, M
// in :
// out: conjugate of 'this'
/**********************************************************************************************************************/

CQuaternion CQuaternion::quatConjugate() const
  {
  return CQuaternion(-m_v3Vector,m_fScalar);
  }

/**********************************************************************************************************************/
/*                                                CONVERSIONS                                                         */
/**********************************************************************************************************************/

/************************************************ vFromAxisAngle ******************************************************/
// axis+angle -> quaternion
// 27/08/00, M
// in : axis,angle
// out:
/**********************************************************************************************************************/

void CQuaternion::vFromAxisAngle(const CVect3D& v3Axis,const FLOAT fAngle)
  {
  m_v3Vector = v3Axis*(FLOAT)sin(fAngle*0.5f);
  m_fScalar  =        (FLOAT)cos(fAngle*0.5f);
  }

/************************************************ vToAxisAngle ********************************************************/
// quaternion -> axis+angle [Ni]
// 27/08/00, M
// in : axis ptr,angle ptr
// out:
/**********************************************************************************************************************/

void CQuaternion::vToAxisAngle(CVect3D* lpv3Axis,FLOAT* lpfAngle)
  {
  FLOAT fLength2 = m_v3Vector*m_v3Vector;

  if(fLength2 < 1e-6)
    {
    lpv3Axis->vSet(0.f);
    *lpfAngle = 0.f;
    }

  else
    {
    *lpfAngle = 2.f*(FLOAT)acos(m_fScalar);
    FLOAT fInvLength2 = 1.f/fLength2;
    lpv3Axis->vSet(m_v3Vector*fInvLength2);
    }
  }

/************************************************ vFromRotation *******************************************************/
// m3x3 -> quaternion [AART]
// 27/08/00, M
// in : 3*3 matrix
// out:
/**********************************************************************************************************************/

void CQuaternion::vFromRotation(const CMat3x3& m3M)
  {
  FLOAT fTr = m3M[0][0]+m3M[1][1]+m3M[2][2];
  FLOAT fS;

  if(fTr > 0.f)
    {
    fS = (FLOAT)sqrt(fTr+1.f);
    m_fScalar = fS*0.5f;
    fS = 0.5f/fS;
    m_v3Vector.vSet((m3M[2][1]-m3M[1][2])*fS,(m3M[0][2]-m3M[2][0])*fS,(m3M[1][0]-m3M[0][1])*fS);
    }

  else
    {
    DWORD dwNext[3] = { _Y_,_Z_,_X_ };
    DWORD dwI = _X_;
    if(m3M[_Y_][_Y_] > m3M[_X_][_X_]) dwI = _Y_;
    if(m3M[_Z_][_Z_] > m3M[dwI][dwI]) dwI = _Z_;
    DWORD dwJ = dwNext[dwI];
    DWORD dwK = dwNext[dwJ];

    fS = (FLOAT)sqrt((m3M[dwI][dwI] - (m3M[dwJ][dwJ]+m3M[dwK][dwK])) + 1.f);
    m_v3Vector[dwI] = fS*0.5f;
    fS = 0.5f/fS;
    m_fScalar       = (m3M[dwK][dwJ]-m3M[dwJ][dwK])*fS;
    m_v3Vector[dwJ] = (m3M[dwJ][dwI]-m3M[dwI][dwJ])*fS;
    m_v3Vector[dwK] = (m3M[dwK][dwI]-m3M[dwI][dwK])*fS;
    }
  }

/************************************************ vFromRotation *******************************************************/
// m4x4 -> quaternion [AART]
// 27/08/00, M
// in : 4*4 matrix
// out:
// rem: save as above (m3M->m4M)
/**********************************************************************************************************************/

void CQuaternion::vFromRotation(const CMat4x4& m4M)
  {
  FLOAT fTr = m4M[0][0]+m4M[1][1]+m4M[2][2];
  FLOAT fS;

  if(fTr > 0.f)
    {
    fS = (FLOAT)sqrt(fTr+1.f);
    m_fScalar = fS*0.5f;
    fS = 0.5f/fS;
    m_v3Vector.vSet((m4M[2][1]-m4M[1][2])*fS,(m4M[0][2]-m4M[2][0])*fS,(m4M[1][0]-m4M[0][1])*fS);
    }

  else
    {
    DWORD dwNext[3] = { _Y_,_Z_,_X_ };
    DWORD dwI = _X_;
    if(m4M[_Y_][_Y_] > m4M[_X_][_X_]) dwI = _Y_;
    if(m4M[_Z_][_Z_] > m4M[dwI][dwI]) dwI = _Z_;
    DWORD dwJ = dwNext[dwI];
    DWORD dwK = dwNext[dwJ];

    fS = (FLOAT)sqrt((m4M[dwI][dwI] - (m4M[dwJ][dwJ]+m4M[dwK][dwK])) + 1.f);
    m_v3Vector[dwI] = fS*0.5f;
    fS = 0.5f/fS;
    m_fScalar       = (m4M[dwK][dwJ]-m4M[dwJ][dwK])*fS;
    m_v3Vector[dwJ] = (m4M[dwJ][dwI]-m4M[dwI][dwJ])*fS;
    m_v3Vector[dwK] = (m4M[dwK][dwI]-m4M[dwI][dwK])*fS;
    }
  }

/************************************************ vToRotation *********************************************************/
// quaternion -> m3x3
// 27/08/00, M
// in : 3*3 matrix ptr
// out:
/**********************************************************************************************************************/

void CQuaternion::vToRotation(CMat3x3* lpm3M) const
  {
  FLOAT s,xs,ys,zs,wx,wy,wz,xx,xy,xz,yy,yz,zz;
  s = 2.f/(m_v3Vector*m_v3Vector+m_fScalar*m_fScalar);      // NB: unit quaternion => s=2.f;

  xs = m_v3Vector[_X_]*s;  ys = m_v3Vector[_Y_]*s;  zs = m_v3Vector[_Z_]*s;
  wx = m_fScalar      *xs; wy = m_fScalar      *ys; wz = m_fScalar      *zs;
  xx = m_v3Vector[_X_]*xs; xy = m_v3Vector[_X_]*ys; xz = m_v3Vector[_X_]*zs;
  yy = m_v3Vector[_Y_]*ys; yz = m_v3Vector[_Y_]*zs; zz = m_v3Vector[_Z_]*zs;

  (*lpm3M)[0].vSet(1.f-(yy+zz),xy-wz,xz+wy);
  (*lpm3M)[1].vSet(xy+wz,1.f-(xx+zz),yz-wx);
  (*lpm3M)[2].vSet(xz-wy,yz+wx,1.f-(xx+yy));
  }

/************************************************ vToRotation *********************************************************/
// quaternion -> m4x4
// 27/08/00, M
// in : 4*4 matrix ptr
// out:
/**********************************************************************************************************************/

void CQuaternion::vToRotation(CMat4x4* lpm4M) const
  {
  FLOAT s,xs,ys,zs,wx,wy,wz,xx,xy,xz,yy,yz,zz;
  s = 2.f/(m_v3Vector*m_v3Vector+m_fScalar*m_fScalar);      // NB: unit quaternion => s=2.f;

  xs = m_v3Vector[_X_]*s;  ys = m_v3Vector[_Y_]*s;  zs = m_v3Vector[_Z_]*s;
  wx = m_fScalar      *xs; wy = m_fScalar      *ys; wz = m_fScalar      *zs;
  xx = m_v3Vector[_X_]*xs; xy = m_v3Vector[_X_]*ys; xz = m_v3Vector[_X_]*zs;
  yy = m_v3Vector[_Y_]*ys; yz = m_v3Vector[_Y_]*zs; zz = m_v3Vector[_Z_]*zs;

  (*lpm4M)[0].vSet(1.f-(yy+zz),xy-wz,xz+wy,0.f);
  (*lpm4M)[1].vSet(xy+wz,1.f-(xx+zz),yz-wx,0.f);
  (*lpm4M)[2].vSet(xz-wy,yz+wx,1.f-(xx+yy),0.f);
  (*lpm4M)[3].vSet(0.f,0.f,0.f,            1.f);
  }
